@@ -4,22 +4,24 @@
 use std::{collections::HashMap, path::Path};
 
 use gettextrs::gettext;
+use matrix_sdk::authentication::oauth::ClientId;
 use oo7::{Item, Keyring};
 use ruma::UserId;
+use serde::Deserialize;
 use thiserror::Error;
 use tokio::fs;
 use tracing::{debug, error, info};
 use url::Url;
 
-use super::{Secret, SecretError, StoredSession, SESSION_ID_LENGTH};
+use super::{SecretError, SecretExt, SessionTokens, StoredSession, SESSION_ID_LENGTH};
 use crate::{gettext_f, prelude::*, spawn_tokio, utils::matrix, APP_ID, PROFILE};
 
 /// The current version of the stored session.
-pub const CURRENT_VERSION: u8 = 6;
+const CURRENT_VERSION: u8 = 7;
 /// The minimum supported version for the stored sessions.
 ///
 /// Currently, this matches the version when Fractal 5 was released.
-pub const MIN_SUPPORTED_VERSION: u8 = 4;
+const MIN_SUPPORTED_VERSION: u8 = 4;
 
 /// Keys used in the Linux secret backend.
 mod keys {
@@ -39,17 +41,47 @@ mod keys {
     pub(super) const DB_PATH: &str = "db-path";
     /// The attribute for the session ID.
     pub(super) const ID: &str = "id";
+    /// The attribute for the client ID.
+    pub(super) const CLIENT_ID: &str = "client-id";
 }
 
-/// Retrieves all sessions stored in the secret backend.
-pub async fn restore_sessions() -> Result<Vec<StoredSession>, SecretError> {
-    match restore_sessions_inner().await {
-        Ok(sessions) => Ok(sessions),
-        Err(error) => {
-            error!("Could not restore previous sessions: {error}");
-            Err(error.into())
+/// Secret API under Linux.
+pub(crate) struct LinuxSecret;
+
+impl SecretExt for LinuxSecret {
+    async fn restore_sessions() -> Result<Vec<StoredSession>, SecretError> {
+        let handle = spawn_tokio!(async move { restore_sessions_inner().await });
+        match handle.await.expect("task was not aborted") {
+            Ok(sessions) => Ok(sessions),
+            Err(error) => {
+                error!("Could not restore previous sessions: {error}");
+                Err(error.into())
+            }
         }
     }
+
+    async fn store_session(session: StoredSession) -> Result<(), SecretError> {
+        let handle = spawn_tokio!(async move { store_session_inner(session).await });
+        match handle.await.expect("task was not aborted") {
+            Ok(()) => Ok(()),
+            Err(error) => {
+                error!("Could not store session: {error}");
+                Err(error.into())
+            }
+        }
+    }
+
+    async fn delete_session(session: &StoredSession) {
+        let attributes = session.attributes();
+
+        spawn_tokio!(async move {
+            if let Err(error) = delete_item_with_attributes(&attributes).await {
+                error!("Could not delete session data from secret backend: {error}");
+            }
+        })
+        .await
+        .expect("task was not aborted");
+    }
 }
 
 async fn restore_sessions_inner() -> Result<Vec<StoredSession>, oo7::Error> {
@@ -74,7 +106,8 @@ async fn restore_sessions_inner() -> Result<Vec<StoredSession>, oo7::Error> {
             Err(LinuxSecretError::OldVersion {
                 version,
                 mut session,
-                attributes,
+                item,
+                access_token,
             }) => {
                 if version < MIN_SUPPORTED_VERSION {
                     info!(
@@ -83,10 +116,12 @@ async fn restore_sessions_inner() -> Result<Vec<StoredSession>, oo7::Error> {
                     );
 
                     // Try to log it out.
-                    log_out_session(session.clone()).await;
+                    if let Some(access_token) = access_token {
+                        log_out_session(session.clone(), access_token).await;
+                    }
 
                     // Delete the session from the secret backend.
-                    delete_session(&session).await;
+                    LinuxSecret::delete_session(&session).await;
 
                     // Delete the session data folders.
                     spawn_tokio!(async move {
@@ -101,7 +136,7 @@ async fn restore_sessions_inner() -> Result<Vec<StoredSession>, oo7::Error> {
                         }
                     })
                     .await
-                    .unwrap();
+                    .expect("task was not aborted");
 
                     continue;
                 }
@@ -110,7 +145,7 @@ async fn restore_sessions_inner() -> Result<Vec<StoredSession>, oo7::Error> {
                     "Found session {} for user {} with old version {}, applying migrations…",
                     session.id, session.user_id, version,
                 );
-                session.apply_migrations(version, attributes).await;
+                session.apply_migrations(version, item, access_token).await;
 
                 sessions.push(session);
             }
@@ -126,25 +161,11 @@ async fn restore_sessions_inner() -> Result<Vec<StoredSession>, oo7::Error> {
     Ok(sessions)
 }
 
-/// Write the given session to the secret backend.
-///
-/// Note that this overwrites any previously stored session with the same
-/// attributes.
-pub async fn store_session(session: StoredSession) -> Result<(), SecretError> {
-    match store_session_inner(session).await {
-        Ok(()) => Ok(()),
-        Err(error) => {
-            error!("Could not store session: {error}");
-            Err(error.into())
-        }
-    }
-}
-
 async fn store_session_inner(session: StoredSession) -> Result<(), oo7::Error> {
     let keyring = Keyring::new().await?;
 
     let attributes = session.attributes();
-    let secret = serde_json::to_string(&session.secret).unwrap();
+    let secret = oo7::Secret::text(session.passphrase);
 
     keyring
         .create_item(
@@ -163,26 +184,19 @@ async fn store_session_inner(session: StoredSession) -> Result<(), oo7::Error> {
     Ok(())
 }
 
-/// Delete the given session from the secret backend.
-pub async fn delete_session(session: &StoredSession) {
-    let attributes = session.attributes();
-
-    spawn_tokio!(async move {
-        if let Err(error) = delete_item_with_attributes(&attributes).await {
-            error!("Could not delete session data from secret backend: {error}");
-        }
-    })
-    .await
-    .unwrap();
-}
-
 /// Create a client and log out the given session.
-async fn log_out_session(session: StoredSession) {
+async fn log_out_session(session: StoredSession, access_token: String) {
     debug!("Logging out session");
+
+    let tokens = SessionTokens {
+        access_token,
+        refresh_token: None,
+    };
+
     spawn_tokio!(async move {
-        match matrix::client_with_stored_session(session).await {
+        match matrix::client_with_stored_session(session, tokens).await {
             Ok(client) => {
-                if let Err(error) = client.matrix_auth().logout().await {
+                if let Err(error) = client.logout().await {
                     error!("Could not log out session: {error}");
                 }
             }
@@ -192,11 +206,11 @@ async fn log_out_session(session: StoredSession) {
         }
     })
     .await
-    .unwrap();
+    .expect("task was not aborted");
 }
 
 impl StoredSession {
-    /// Build self from a secret.
+    /// Build self from an item.
     async fn try_from_secret_item(item: Item) -> Result<Self, LinuxSecretError> {
         let attributes = item.attributes().await?;
 
@@ -208,6 +222,7 @@ impl StoredSession {
         let homeserver = parse_attribute(&attributes, keys::HOMESERVER, Url::parse)?;
         let user_id = parse_attribute(&attributes, keys::USER, |s| UserId::parse(s))?;
         let device_id = get_attribute(&attributes, keys::DEVICE_ID)?.as_str().into();
+
         let id = if version <= 5 {
             let string = get_attribute(&attributes, keys::DB_PATH)?;
             Path::new(string)
@@ -219,21 +234,38 @@ impl StoredSession {
         } else {
             get_attribute(&attributes, keys::ID)?.clone()
         };
-        let secret = match item.secret().await {
+
+        let client_id = attributes.get(keys::CLIENT_ID).cloned().map(ClientId::new);
+
+        let (passphrase, access_token) = match item.secret().await {
             Ok(secret) => {
-                if version <= 4 {
-                    match rmp_serde::from_slice::<Secret>(&secret) {
-                        Ok(secret) => secret,
-                        Err(error) => {
-                            error!("Could not parse secret in stored session: {error}");
-                            return Err(LinuxSecretFieldError::Invalid.into());
+                if version <= 6 {
+                    let secret_data = if version <= 4 {
+                        match rmp_serde::from_slice::<V4SecretData>(&secret) {
+                            Ok(secret) => secret,
+                            Err(error) => {
+                                error!("Could not parse secret in stored session: {error}");
+                                return Err(LinuxSecretFieldError::Invalid.into());
+                            }
                         }
-                    }
+                    } else {
+                        match serde_json::from_slice(&secret) {
+                            Ok(secret) => secret,
+                            Err(error) => {
+                                error!("Could not parse secret in stored session: {error:?}");
+                                return Err(LinuxSecretFieldError::Invalid.into());
+                            }
+                        }
+                    };
+
+                    (secret_data.passphrase, Some(secret_data.access_token))
                 } else {
-                    match serde_json::from_slice(&secret) {
-                        Ok(secret) => secret,
+                    // Even if we store the secret as plain text, the file backend always returns a
+                    // blob so let's always treat it as a byte slice.
+                    match String::from_utf8(secret.as_bytes().to_owned()) {
+                        Ok(passphrase) => (passphrase.clone(), None),
                         Err(error) => {
-                            error!("Could not parse secret in stored session: {error:?}");
+                            error!("Could not get secret in stored session: {error}");
                             return Err(LinuxSecretFieldError::Invalid.into());
                         }
                     }
@@ -250,14 +282,16 @@ impl StoredSession {
             user_id,
             device_id,
             id,
-            secret,
+            client_id,
+            passphrase: passphrase.into(),
         };
 
         if version < CURRENT_VERSION {
             Err(LinuxSecretError::OldVersion {
                 version,
                 session,
-                attributes,
+                item,
+                access_token,
             })
         } else {
             Ok(session)
@@ -266,7 +300,7 @@ impl StoredSession {
 
     /// Get the attributes from `self`.
     fn attributes(&self) -> HashMap<&'static str, String> {
-        HashMap::from([
+        let mut attributes = HashMap::from([
             (keys::HOMESERVER, self.homeserver.to_string()),
             (keys::USER, self.user_id.to_string()),
             (keys::DEVICE_ID, self.device_id.to_string()),
@@ -274,52 +308,94 @@ impl StoredSession {
             (keys::VERSION, CURRENT_VERSION.to_string()),
             (keys::PROFILE, PROFILE.to_string()),
             (keys::XDG_SCHEMA, APP_ID.to_owned()),
-        ])
+        ]);
+
+        if let Some(client_id) = &self.client_id {
+            attributes.insert(keys::CLIENT_ID, client_id.as_str().to_owned());
+        }
+
+        attributes
     }
 
     /// Migrate this session to the current version.
-    async fn apply_migrations(&mut self, from_version: u8, attributes: HashMap<String, String>) {
+    async fn apply_migrations(
+        &mut self,
+        from_version: u8,
+        item: Item,
+        access_token: Option<String>,
+    ) {
+        // Version 5 changes the serialization of the secret from MessagePack to JSON.
+        // We can ignore the migration because we changed the format of the secret again
+        // in version 7.
+
         if from_version < 6 {
-            // Version 5 changes the serialization of the secret from MessagePack to JSON.
             // Version 6 truncates sessions IDs, changing the path of the databases, and
             // removes the `db-path` attribute to replace it with the `id` attribute.
-            // They both remove and add again the item in the secret backend so we merged
-            // the migrations.
+            // Because we need to update the `version` in the attributes for version 7, we
+            // only migrate the path here.
             info!("Migrating to version 6…");
 
-            // Keep the old state of the session.
+            // Get the old path of the session.
             let old_path = self.data_path();
 
             // Truncate the session ID.
             self.id.truncate(SESSION_ID_LENGTH);
             let new_path = self.data_path();
 
-            let clone = self.clone();
             spawn_tokio!(async move {
                 debug!(
                     "Renaming databases directory to: {}",
                     new_path.to_string_lossy()
                 );
+
                 if let Err(error) = fs::rename(old_path, new_path).await {
                     error!("Could not rename databases directory: {error}");
                 }
+            })
+            .await
+            .expect("task was not aborted");
+        }
 
-                // Changing an attribute in an item creates a new item in oo7 because of a bug,
-                // so we need to delete it and create a new one.
-                if let Err(error) = delete_item_with_attributes(&attributes).await {
-                    error!("Could not remove outdated session: {error}");
+        if from_version < 7 {
+            // Version 7 moves the access token to a separate file. Only the passphrase is
+            // stored as the secret now.
+            info!("Migrating to version 7…");
+
+            let new_attributes = self.attributes();
+            let new_secret = oo7::Secret::text(&self.passphrase);
+
+            spawn_tokio!(async move {
+                if let Err(error) = item.set_secret(new_secret).await {
+                    error!("Could not store updated session secret: {error}");
                 }
 
-                if let Err(error) = store_session_inner(clone).await {
-                    error!("Could not store updated session: {error}");
+                if let Err(error) = item.set_attributes(&new_attributes).await {
+                    error!("Could not store updated session attributes: {error}");
                 }
             })
             .await
-            .unwrap();
+            .expect("task was not aborted");
+
+            if let Some(access_token) = access_token {
+                let session_tokens = SessionTokens {
+                    access_token,
+                    refresh_token: None,
+                };
+                self.store_tokens(session_tokens).await;
+            }
         }
     }
 }
 
+/// Secret data that was stored in the secret backend from versions 4 through 6.
+#[derive(Clone, Deserialize)]
+struct V4SecretData {
+    /// The access token to provide to the homeserver for authentication.
+    access_token: String,
+    /// The passphrase used to encrypt the local databases.
+    passphrase: String,
+}
+
 /// Get the attribute with the given key in the given map.
 fn get_attribute<'a>(
     attributes: &'a HashMap<String, String>,
@@ -354,7 +430,7 @@ where
 /// Any error that can happen when retrieving an attribute from the secret
 /// backends on Linux.
 #[derive(Debug, Error)]
-pub enum LinuxSecretFieldError {
+enum LinuxSecretFieldError {
     /// An attribute is missing.
     ///
     /// This should only happen if for some reason we get an item from a
@@ -385,7 +461,7 @@ async fn delete_item_with_attributes(
 #[derive(Debug, Error)]
 // Complains about StoredSession in OldVersion, but we need it.
 #[allow(clippy::large_enum_variant)]
-pub enum LinuxSecretError {
+enum LinuxSecretError {
     /// A session with an unsupported version was found.
     #[error("Session found with unsupported version {0}")]
     UnsupportedVersion(u8),
@@ -397,12 +473,12 @@ pub enum LinuxSecretError {
         version: u8,
         /// The session that was found.
         session: StoredSession,
-        /// The attributes of the secret item for the session.
+        /// The item for the session.
+        item: Item,
+        /// The access token that was found.
         ///
-        /// We use it to update the secret item because, if we use the `Item`
-        /// directly, the Secret portal API returns errors saying that the file
-        /// has changed after the first item was modified.
-        attributes: HashMap<String, String>,
+        /// It needs to be stored outside of the secret backend now.
+        access_token: Option<String>,
     },
 
     /// An error occurred while retrieving a field of the session.
@@ -426,15 +502,15 @@ impl From<oo7::Error> for SecretError {
 impl UserFacingError for oo7::Error {
     fn to_user_facing(&self) -> String {
         match self {
-            oo7::Error::Portal(error) => error.to_user_facing(),
+            oo7::Error::File(error) => error.to_user_facing(),
             oo7::Error::DBus(error) => error.to_user_facing(),
         }
     }
 }
 
-impl UserFacingError for oo7::portal::Error {
+impl UserFacingError for oo7::file::Error {
     fn to_user_facing(&self) -> String {
-        use oo7::portal::Error;
+        use oo7::file::Error;
 
         match self {
             Error::FileHeaderMismatch(_) |
@@ -446,6 +522,8 @@ impl UserFacingError for oo7::portal::Error {
             Error::SaltSizeMismatch(_, _) |
             Error::ChecksumMismatch |
             Error::AlgorithmMismatch(_) |
+            Error::IncorrectSecret |
+            Error::Crypto(_) |
             Error::Utf8(_) => gettext(
                 "The secret storage file is corrupted.",
             ),
@@ -459,15 +537,15 @@ impl UserFacingError for oo7::portal::Error {
             Error::TargetFileChanged(_) => gettext(
                 "The secret storage file has been changed by another process.",
             ),
-            Error::PortalBus(_) => gettext(
-                "An unexpected error occurred when interacting with the D-Bus Secret Portal backend.",
-            ),
-            Error::CancelledPortalRequest => gettext(
+            Error::Portal(ashpd::Error::Portal(ashpd::PortalError::Cancelled(_))) => gettext(
                 "The request to the Flatpak Secret Portal was cancelled. Make sure to accept any prompt asking to access it.",
             ),
-            Error::PortalNotAvailable => gettext(
+            Error::Portal(ashpd::Error::PortalNotFound(_)) => gettext(
                 "The Flatpak Secret Portal is not available. Make sure xdg-desktop-portal is installed, and it is at least at version 1.5.0.",
             ),
+            Error::Portal(_) => gettext(
+                "An unexpected error occurred when interacting with the D-Bus Secret Portal backend.",
+            ),
             Error::WeakKey(_) => gettext(
                 "The Flatpak Secret Portal provided a key that is too weak to be secure.",
             ),
@@ -489,13 +567,13 @@ impl UserFacingError for oo7::dbus::Error {
                 ServiceError::ZBus(_) => gettext(
                     "An unexpected error occurred when interacting with the D-Bus Secret Service.",
                 ),
-                ServiceError::IsLocked => gettext(
+                ServiceError::IsLocked(_) => gettext(
                     "The collection or item is locked.",
                 ),
-                ServiceError::NoSession => gettext(
+                ServiceError::NoSession(_) => gettext(
                     "The D-Bus Secret Service session does not exist.",
                 ),
-                ServiceError::NoSuchObject => gettext(
+                ServiceError::NoSuchObject(_) => gettext(
                     "The collection or item does not exist.",
                 ),
             },
@@ -505,7 +583,8 @@ impl UserFacingError for oo7::dbus::Error {
             Error::NotFound(_) => gettext(
                 "Could not access the default collection. Make sure a keyring was created and set as default.",
             ),
-            Error::Zbus(_) |
+            Error::ZBus(_) |
+            Error::Crypto(_) |
             Error::IO(_) => gettext(
                 "An unexpected error occurred when interacting with the D-Bus Secret Service.",
             ),
