@@ -1,7 +1,6 @@
 mod ignored_users;
 mod notifications;
-mod remote_room;
-mod remote_user;
+mod remote;
 mod room;
 mod room_list;
 mod security;
@@ -12,18 +11,17 @@ mod user;
 mod user_sessions_list;
 mod verification;
 
-pub use self::{
+pub(crate) use self::{
     ignored_users::IgnoredUsers,
     notifications::{
         Notifications, NotificationsGlobalSetting, NotificationsRoomSetting, NotificationsSettings,
     },
-    remote_room::RemoteRoom,
-    remote_user::RemoteUser,
+    remote::*,
     room::*,
-    room_list::RoomList,
+    room_list::*,
     security::*,
     session::*,
-    session_settings::{SessionSettings, StoredSessionSettings},
+    session_settings::*,
     sidebar_data::{
         Selection, SidebarIconItem, SidebarIconItemType, SidebarItemList, SidebarListModel,
         SidebarSection, SidebarSectionName,
