@@ -32,7 +32,6 @@ use tracing::{debug, error, warn};
 
 mod aliases;
 mod category;
-mod event;
 mod highlight_flags;
 mod join_rule;
 mod member;
@@ -44,7 +43,6 @@ mod typing_list;
 pub(crate) use self::{
     aliases::{AddAltAliasError, RegisterLocalAliasError, RoomAliases},
     category::{RoomCategory, TargetRoomCategory},
-    event::*,
     highlight_flags::HighlightFlags,
     join_rule::{JoinRule, JoinRuleValue},
     member::{Member, Membership},
@@ -161,9 +159,9 @@ mod imp {
         /// there is only one other member.
         #[property(get)]
         direct_member: RefCell<Option<Member>>,
-        /// The timeline of this room.
+        /// The live timeline of this room.
         #[property(get)]
-        timeline: OnceCell<Timeline>,
+        live_timeline: OnceCell<Timeline>,
         /// The timestamp of the room's latest activity.
         ///
         /// This is the timestamp of the latest event that counts as possibly
@@ -254,7 +252,7 @@ mod imp {
                 .set(matrix_room)
                 .expect("matrix room is uninitialized");
 
-            self.init_timeline();
+            self.init_live_timeline();
             self.aliases.init(&obj);
             self.load_predecessor();
             self.watch_members();
@@ -284,7 +282,7 @@ mod imp {
                                 | RoomCategory::Normal
                                 | RoomCategory::LowPriority
                         );
-                        imp.timeline().set_preload(preload);
+                        imp.live_timeline().set_preload(preload);
 
                         imp.permissions.init(&imp.obj()).await;
                     }
@@ -449,7 +447,7 @@ mod imp {
 
                 self.set_has_avatar(true);
                 return;
-            };
+            }
 
             self.set_has_avatar(false);
 
@@ -623,7 +621,7 @@ mod imp {
                     .set(room_tombstone.replacement_room)
                     .expect("successor ID is uninitialized");
                 obj.notify_successor_id_string();
-            };
+            }
 
             // Try to get the successor.
             self.update_successor();
@@ -784,7 +782,7 @@ mod imp {
         }
 
         /// The member corresponding to our own user.
-        fn own_member(&self) -> &Member {
+        pub(super) fn own_member(&self) -> &Member {
             self.own_member.get().expect("Own member was initialized")
         }
 
@@ -983,9 +981,11 @@ mod imp {
             self.set_direct_member(Some(direct_member));
         }
 
-        /// Initialize the timeline of this room.
-        fn init_timeline(&self) {
-            let timeline = self.timeline.get_or_init(|| Timeline::new(&self.obj()));
+        /// Initialize the live timeline of this room.
+        fn init_live_timeline(&self) {
+            let timeline = self
+                .live_timeline
+                .get_or_init(|| Timeline::new(&self.obj()));
 
             timeline.connect_read_change_trigger(clone!(
                 #[weak(rename_to = imp)]
@@ -998,9 +998,11 @@ mod imp {
             ));
         }
 
-        /// The timeline of this room.
-        fn timeline(&self) -> &Timeline {
-            self.timeline.get().expect("timeline is initialized")
+        /// The live timeline of this room.
+        fn live_timeline(&self) -> &Timeline {
+            self.live_timeline
+                .get()
+                .expect("live timeline is initialized")
         }
 
         /// Set the timestamp of the room's latest possibly unread event.
@@ -1025,7 +1027,7 @@ mod imp {
 
         /// Handle the trigger emitted when a read change might have occurred.
         async fn handle_read_change_trigger(&self) {
-            let timeline = self.timeline();
+            let timeline = self.live_timeline();
 
             if let Some(has_unread) = timeline.has_unread_messages().await {
                 self.set_is_read(!has_unread);
@@ -1096,15 +1098,15 @@ mod imp {
         async fn update_is_encrypted(&self) {
             let matrix_room = self.matrix_room();
             let matrix_room_clone = matrix_room.clone();
-            let handle = spawn_tokio!(async move { matrix_room_clone.is_encrypted().await });
+            let handle =
+                spawn_tokio!(async move { matrix_room_clone.latest_encryption_state().await });
 
             match handle.await.expect("task was not aborted") {
-                Ok(true) => {
-                    self.is_encrypted.set(true);
-                    self.obj().notify_is_encrypted();
-                }
-                Ok(false) => {
-                    // Ignore as the room encryption cannot be disabled.
+                Ok(state) => {
+                    if state.is_encrypted() {
+                        self.is_encrypted.set(true);
+                        self.obj().notify_is_encrypted();
+                    }
                 }
                 Err(error) => {
                     // It can be expected to not be allowed to access the encryption state if the
@@ -1173,7 +1175,7 @@ mod imp {
             let matrix_room = self.matrix_room();
             if matrix_room.state() != RoomState::Joined {
                 return;
-            };
+            }
 
             let (typing_drop_guard, receiver) = matrix_room.subscribe_to_typing_notifications();
             let stream = BroadcastStream::new(receiver);
@@ -1627,10 +1629,11 @@ impl Room {
 
     /// Toggle the `key` reaction on the given related event in this room.
     pub(crate) async fn toggle_reaction(&self, key: String, event: &Event) -> Result<(), ()> {
-        let timeline = self.timeline().matrix_timeline();
+        let matrix_timeline = self.live_timeline().matrix_timeline();
         let identifier = event.identifier();
 
-        let handle = spawn_tokio!(async move { timeline.toggle_reaction(&identifier, &key).await });
+        let handle =
+            spawn_tokio!(async move { matrix_timeline.toggle_reaction(&identifier, &key).await });
 
         if let Err(error) = handle.await.expect("task was not aborted") {
             error!("Could not toggle reaction: {error}");
@@ -1656,7 +1659,7 @@ impl Room {
             t => t,
         };
 
-        let matrix_timeline = self.timeline().matrix_timeline();
+        let matrix_timeline = self.live_timeline().matrix_timeline();
         let handle = spawn_tokio!(async move {
             match position {
                 ReceiptPosition::End => matrix_timeline.mark_as_read(receipt_type).await,
@@ -1678,7 +1681,7 @@ impl Room {
         let matrix_room = self.matrix_room();
         if matrix_room.state() != RoomState::Joined {
             return;
-        };
+        }
 
         let matrix_room = matrix_room.clone();
         let handle = spawn_tokio!(async move { matrix_room.typing_notice(is_typing).await });
@@ -1687,7 +1690,7 @@ impl Room {
             match handle.await.expect("task was not aborted") {
                 Ok(()) => {}
                 Err(error) => error!("Could not send typing notification: {error}"),
-            };
+            }
         });
     }
 
@@ -1703,7 +1706,7 @@ impl Room {
         let matrix_room = self.matrix_room();
         if matrix_room.state() != RoomState::Joined {
             return Ok(());
-        };
+        }
 
         let events_clone = events.to_owned();
         let matrix_room = matrix_room.clone();
@@ -2006,10 +2009,11 @@ impl Room {
     ///
     /// The events must be in reverse chronological order.
     fn update_latest_activity<'a>(&self, events: impl Iterator<Item = &'a Event>) {
+        let own_user_id = self.imp().own_member().user_id();
         let mut latest_activity = self.latest_activity();
 
         for event in events {
-            if event.counts_as_unread() {
+            if event.counts_as_activity(own_user_id) {
                 latest_activity = latest_activity.max(event.origin_server_ts().get().into());
                 break;
             }
