@@ -24,38 +24,31 @@ use ruma::{
 use tokio::task::AbortHandle;
 use tracing::error;
 
+mod event;
+mod timeline_diff_minimizer;
 mod timeline_item;
-mod timeline_item_diff_minimizer;
 mod virtual_item;
 
-use self::timeline_item_diff_minimizer::{TimelineItemDiff, TimelineItemStore};
+use self::timeline_diff_minimizer::{TimelineDiff, TimelineDiffItemStore};
 pub(crate) use self::{
+    event::*,
     timeline_item::{TimelineItem, TimelineItemExt, TimelineItemImpl},
     virtual_item::{VirtualItem, VirtualItemKind},
 };
-use super::{Event, Room};
-use crate::{prelude::*, spawn, spawn_tokio};
-
-/// The possible states of the timeline.
-#[derive(Debug, Default, Hash, Eq, PartialEq, Clone, Copy, glib::Enum)]
-#[repr(u32)]
-#[enum_type(name = "TimelineState")]
-pub enum TimelineState {
-    /// The timeline is not initialized yet.
-    #[default]
-    Initial,
-    /// The timeline is currently loading.
-    Loading,
-    /// The timeline has been initialized and there is no ongoing action.
-    Ready,
-    /// An error occurred with the timeline.
-    Error,
-    /// We have reached the beginning of the timeline.
-    Complete,
-}
+use super::Room;
+use crate::{
+    prelude::*,
+    spawn, spawn_tokio,
+    utils::{LoadingState, SingleItemListModel},
+};
 
 /// The number of events to request when loading more history.
 const MAX_BATCH_SIZE: u16 = 20;
+/// The maximum time between contiguous events before we show their header, in
+/// milliseconds.
+///
+/// This matches 20 minutes.
+const MAX_TIME_BETWEEN_HEADERS: u64 = 20 * 60 * 1000;
 
 mod imp {
     use std::{
@@ -69,7 +62,7 @@ mod imp {
 
     use super::*;
 
-    #[derive(Debug, glib::Properties)]
+    #[derive(Debug, Default, glib::Properties)]
     #[properties(wrapper_type = super::Timeline)]
     pub struct Timeline {
         /// The room containing this timeline.
@@ -78,27 +71,41 @@ mod imp {
         /// The underlying SDK timeline.
         matrix_timeline: OnceCell<Arc<SdkTimeline>>,
         /// Items added at the start of the timeline.
-        start_items: gio::ListStore,
+        ///
+        /// Currently this can only contain one item at a time.
+        start_items: OnceCell<SingleItemListModel>,
         /// Items provided by the SDK timeline.
-        pub(super) sdk_items: gio::ListStore,
+        sdk_items: OnceCell<gio::ListStore>,
+        /// Filter for the list of items provided by the SDK timeline.
+        filter: gtk::CustomFilter,
+        /// Filtered list of items provided by the SDK timeline.
+        filtered_sdk_items: gtk::FilterListModel,
         /// Items added at the end of the timeline.
-        end_items: gio::ListStore,
+        ///
+        /// Currently this can only contain one item at a time.
+        end_items: OnceCell<SingleItemListModel>,
         /// The `GListModel` containing all the timeline items.
-        #[property(get)]
-        items: gtk::FlattenListModel,
+        #[property(get = Self::items)]
+        items: OnceCell<gtk::FlattenListModel>,
         /// A Hashmap linking a `TimelineEventItemId` to the corresponding
         /// `Event`.
         pub(super) event_map: RefCell<HashMap<TimelineEventItemId, Event>>,
-        /// The state of the timeline.
-        #[property(get, builder(TimelineState::default()))]
-        state: Cell<TimelineState>,
+        /// The loading state of the timeline.
+        #[property(get, builder(LoadingState::default()))]
+        state: Cell<LoadingState>,
+        /// Whether we are loading events at the start of the timeline.
+        #[property(get)]
+        is_loading_start: Cell<bool>,
         /// Whether the timeline is empty.
         #[property(get = Self::is_empty)]
         is_empty: PhantomData<bool>,
         /// Whether the timeline should be pre-loaded when it is ready.
         #[property(get, set = Self::set_preload, explicit_notify)]
         preload: Cell<bool>,
-        /// Whether the timeline has the `m.room.create` event of the room.
+        /// Whether we have reached the start of the timeline.
+        #[property(get)]
+        has_reached_start: Cell<bool>,
+        /// Whether we have the `m.room.create` event in the timeline.
         #[property(get)]
         has_room_create: Cell<bool>,
         diff_handle: OnceCell<AbortHandle>,
@@ -106,36 +113,6 @@ mod imp {
         read_receipts_changed_handle: OnceCell<AbortHandle>,
     }
 
-    impl Default for Timeline {
-        fn default() -> Self {
-            let start_items = gio::ListStore::new::<TimelineItem>();
-            let sdk_items = gio::ListStore::new::<TimelineItem>();
-            let end_items = gio::ListStore::new::<TimelineItem>();
-
-            let model_list = gio::ListStore::new::<gio::ListModel>();
-            model_list.append(&start_items);
-            model_list.append(&sdk_items);
-            model_list.append(&end_items);
-
-            Self {
-                room: Default::default(),
-                matrix_timeline: Default::default(),
-                start_items,
-                sdk_items,
-                end_items,
-                items: gtk::FlattenListModel::new(Some(model_list)),
-                event_map: Default::default(),
-                state: Default::default(),
-                is_empty: Default::default(),
-                has_room_create: Default::default(),
-                preload: Default::default(),
-                diff_handle: Default::default(),
-                back_pagination_status_handle: Default::default(),
-                read_receipts_changed_handle: Default::default(),
-            }
-        }
-    }
-
     #[glib::object_subclass]
     impl ObjectSubclass for Timeline {
         const NAME: &'static str = "Timeline";
@@ -150,6 +127,25 @@ mod imp {
             SIGNALS.as_ref()
         }
 
+        fn constructed(&self) {
+            self.parent_constructed();
+
+            self.filter.set_filter_func(clone!(
+                #[weak(rename_to = imp)]
+                self,
+                #[upgrade_or]
+                true,
+                move |obj| {
+                    // Hide the timeline start item if we have the `m.room.create` event too.
+                    obj.downcast_ref::<VirtualItem>().is_none_or(|item| {
+                        !(imp.has_room_create.get()
+                            && item.kind() == VirtualItemKind::TimelineStart)
+                    })
+                }
+            ));
+            self.filtered_sdk_items.set_filter(Some(&self.filter));
+        }
+
         fn dispose(&self) {
             if let Some(handle) = self.diff_handle.get() {
                 handle.abort();
@@ -214,6 +210,14 @@ mod imp {
 
             let (values, timeline_stream) = matrix_timeline.subscribe().await;
 
+            if *IS_AT_TRACE_LEVEL {
+                tracing::trace!(
+                    room = self.room().human_readable_id(),
+                    items = ?sdk_items_to_log(&values),
+                    "Initial timeline items",
+                );
+            }
+
             if !values.is_empty() {
                 self.update_with_single_diff(VectorDiff::Append { values });
             }
@@ -242,7 +246,7 @@ mod imp {
             let diff_handle = spawn_tokio!(fut);
             self.diff_handle
                 .set(diff_handle.abort_handle())
-                .expect("handle is uninitialized");
+                .expect("handle should be uninitialized");
 
             self.watch_read_receipts().await;
 
@@ -250,31 +254,136 @@ mod imp {
                 self.preload().await;
             }
 
-            self.set_state(TimelineState::Ready);
+            self.set_state(LoadingState::Ready);
         }
 
         /// The underlying SDK timeline.
         pub(super) fn matrix_timeline(&self) -> &Arc<SdkTimeline> {
             self.matrix_timeline
                 .get()
-                .expect("matrix timeline is initialized")
+                .expect("matrix timeline should be initialized")
+        }
+
+        /// Items added at the start of the timeline.
+        fn start_items(&self) -> &SingleItemListModel {
+            self.start_items.get_or_init(|| {
+                let model = SingleItemListModel::new(&VirtualItem::spinner(&self.obj()));
+                model.set_is_hidden(true);
+                model
+            })
+        }
+
+        /// Items provided by the SDK timeline.
+        pub(super) fn sdk_items(&self) -> &gio::ListStore {
+            self.sdk_items.get_or_init(|| {
+                let sdk_items = gio::ListStore::new::<TimelineItem>();
+                self.filtered_sdk_items.set_model(Some(&sdk_items));
+                sdk_items
+            })
+        }
+
+        /// Items added at the end of the timeline.
+        fn end_items(&self) -> &SingleItemListModel {
+            self.end_items.get_or_init(|| {
+                let model = SingleItemListModel::new(&VirtualItem::typing(&self.obj()));
+                model.set_is_hidden(true);
+                model
+            })
+        }
+
+        /// The `GListModel` containing all the timeline items.
+        fn items(&self) -> gtk::FlattenListModel {
+            self.items
+                .get_or_init(|| {
+                    let model_list = gio::ListStore::new::<gio::ListModel>();
+                    model_list.append(self.start_items());
+                    model_list.append(&self.filtered_sdk_items);
+                    model_list.append(self.end_items());
+                    gtk::FlattenListModel::new(Some(model_list))
+                })
+                .clone()
         }
 
         /// Whether the timeline is empty.
         fn is_empty(&self) -> bool {
-            self.sdk_items.n_items() == 0
+            self.filtered_sdk_items.n_items() == 0
+        }
+
+        /// Set the loading state of the timeline.
+        fn set_state(&self, state: LoadingState) {
+            if self.state.get() == state {
+                return;
+            }
+
+            self.state.set(state);
+
+            self.obj().notify_state();
+        }
+
+        /// Update the loading state of the timeline.
+        fn update_loading_state(&self) {
+            let is_loading = self.is_loading_start.get();
+
+            if is_loading {
+                self.set_state(LoadingState::Loading);
+            } else if self.state.get() != LoadingState::Error {
+                self.set_state(LoadingState::Ready);
+            }
+        }
+
+        /// Set whether we are loading events at the start of the timeline.
+        fn set_loading_start(&self, is_loading_start: bool) {
+            if self.is_loading_start.get() == is_loading_start {
+                return;
+            }
+
+            self.is_loading_start.set(is_loading_start);
+
+            self.update_loading_state();
+            self.start_items().set_is_hidden(!is_loading_start);
+            self.obj().notify_is_loading_start();
+        }
+
+        /// Set whether we have reached the start of the timeline.
+        fn set_has_reached_start(&self, has_reached_start: bool) {
+            if self.has_reached_start.get() == has_reached_start {
+                // Nothing to do.
+                return;
+            }
+
+            self.has_reached_start.set(has_reached_start);
+
+            self.obj().notify_has_reached_start();
         }
 
         /// Set whether the timeline has the `m.room.create` event of the room.
-        pub(super) fn set_has_room_create(&self, has_room_create: bool) {
+        fn set_has_room_create(&self, has_room_create: bool) {
             if self.has_room_create.get() == has_room_create {
                 return;
             }
 
             self.has_room_create.set(has_room_create);
+
+            let change = if has_room_create {
+                gtk::FilterChange::MoreStrict
+            } else {
+                gtk::FilterChange::LessStrict
+            };
+            self.filter.changed(change);
+
             self.obj().notify_has_room_create();
         }
 
+        /// Clear the state of the timeline.
+        ///
+        /// This doesn't handle removing items in `sdk_items` because it can be
+        /// optimized by the caller of the function.
+        fn clear(&self) {
+            self.event_map.borrow_mut().clear();
+            self.set_has_reached_start(false);
+            self.set_has_room_create(false);
+        }
+
         /// Set whether the timeline should be pre-loaded when it is ready.
         fn set_preload(&self, preload: bool) {
             if self.preload.get() == preload {
@@ -284,7 +393,7 @@ mod imp {
             self.preload.set(preload);
             self.obj().notify_preload();
 
-            if preload && self.can_load() {
+            if preload && self.can_paginate_backwards() {
                 spawn!(
                     glib::Priority::DEFAULT_IDLE,
                     clone!(
@@ -300,13 +409,17 @@ mod imp {
 
         /// Preload the timeline, if there are not enough items.
         async fn preload(&self) {
-            if self.sdk_items.n_items() < u32::from(MAX_BATCH_SIZE) {
-                self.load(|| ControlFlow::Break(())).await;
+            if self.filtered_sdk_items.n_items() < u32::from(MAX_BATCH_SIZE) {
+                self.paginate_backwards(|| ControlFlow::Break(())).await;
             }
         }
 
         /// Update this timeline with the given diff list.
         fn update_with_diff_list(&self, diff_list: Vec<VectorDiff<Arc<SdkTimelineItem>>>) {
+            if *IS_AT_TRACE_LEVEL {
+                self.log_diff_list(&diff_list);
+            }
+
             let was_empty = self.is_empty();
 
             if let Some(diff_list) = self.try_minimize_diff_list(diff_list) {
@@ -316,6 +429,10 @@ mod imp {
                 }
             }
 
+            if *IS_AT_TRACE_LEVEL {
+                self.log_items();
+            }
+
             let obj = self.obj();
             if self.is_empty() != was_empty {
                 obj.notify_is_empty();
@@ -353,12 +470,11 @@ mod imp {
                         .map(|item| self.create_item(&item))
                         .collect::<Vec<_>>();
 
-                    self.update_items(self.sdk_items.n_items(), 0, &new_list);
+                    self.update_items(self.sdk_items().n_items(), 0, &new_list);
                 }
                 VectorDiff::Clear => {
-                    self.sdk_items.remove_all();
-                    self.event_map.borrow_mut().clear();
-                    self.set_has_room_create(false);
+                    self.sdk_items().remove_all();
+                    self.clear();
                 }
                 VectorDiff::PushFront { value } => {
                     let item = self.create_item(&value);
@@ -366,13 +482,13 @@ mod imp {
                 }
                 VectorDiff::PushBack { value } => {
                     let item = self.create_item(&value);
-                    self.update_items(self.sdk_items.n_items(), 0, &[item]);
+                    self.update_items(self.sdk_items().n_items(), 0, &[item]);
                 }
                 VectorDiff::PopFront => {
                     self.update_items(0, 1, &[]);
                 }
                 VectorDiff::PopBack => {
-                    self.update_items(self.sdk_items.n_items().saturating_sub(1), 1, &[]);
+                    self.update_items(self.sdk_items().n_items().saturating_sub(1), 1, &[]);
                 }
                 VectorDiff::Insert { index, value } => {
                     let item = self.create_item(&value);
@@ -399,15 +515,14 @@ mod imp {
                 }
                 VectorDiff::Truncate { length } => {
                     let length = length as u32;
-                    let old_len = self.sdk_items.n_items();
+                    let old_len = self.sdk_items().n_items();
                     self.update_items(length, old_len.saturating_sub(length), &[]);
                 }
                 VectorDiff::Reset { values } => {
                     // Reset the state.
-                    self.event_map.borrow_mut().clear();
-                    self.set_has_room_create(false);
+                    self.clear();
 
-                    let removed = self.sdk_items.n_items();
+                    let removed = self.sdk_items().n_items();
                     let new_list = values
                         .into_iter()
                         .map(|item| self.create_item(&item))
@@ -420,7 +535,7 @@ mod imp {
 
         /// Get the item at the given position.
         fn item_at(&self, pos: u32) -> Option<TimelineItem> {
-            self.sdk_items.item(pos).and_downcast()
+            self.sdk_items().item(pos).and_downcast()
         }
 
         /// Update the items at the given position by removing the given number
@@ -436,7 +551,7 @@ mod imp {
                 self.remove_item(&item);
             }
 
-            self.sdk_items.splice(pos, n_removals, additions);
+            self.sdk_items().splice(pos, n_removals, additions);
 
             // Update the header visibility of all the new additions, and the first item
             // after this batch.
@@ -453,35 +568,58 @@ mod imp {
         /// Update the headers of the item at the given position and the given
         /// number of items after it.
         fn update_items_headers(&self, pos: u32, nb: u32) {
-            let sdk_items = &self.sdk_items;
+            let sdk_items = self.sdk_items();
 
-            let mut previous_sender = if pos > 0 {
+            let (mut previous_sender, mut previous_timestamp) = if pos > 0 {
                 sdk_items
                     .item(pos - 1)
-                    .and_downcast::<TimelineItem>()
-                    .filter(TimelineItem::can_hide_header)
-                    .and_then(|item| item.event_sender_id())
+                    .and_downcast::<Event>()
+                    .filter(Event::can_show_header)
+                    .map(|event| (event.sender_id(), event.origin_server_ts()))
             } else {
                 None
-            };
+            }
+            .unzip();
 
             // Update the headers of changed events plus the first event after them.
             for i in pos..=pos + nb {
                 let Some(current) = self.item_at(i) else {
                     break;
                 };
-
-                let current_sender = current.event_sender_id();
-
-                if !current.can_hide_header() {
-                    current.set_show_header(false);
+                let Ok(current) = current.downcast::<Event>() else {
                     previous_sender = None;
-                } else if current_sender != previous_sender {
-                    current.set_show_header(true);
-                    previous_sender = current_sender;
-                } else {
-                    current.set_show_header(false);
+                    continue;
+                };
+
+                let current_sender = current.sender_id();
+
+                if !current.can_show_header() {
+                    current.set_header_state(EventHeaderState::Hidden);
+                    previous_sender = None;
+                    previous_timestamp = None;
+                    continue;
                 }
+
+                let header_state = if previous_sender
+                    .as_ref()
+                    .is_none_or(|previous_sender| current_sender != *previous_sender)
+                {
+                    // The sender is different, show the full header.
+                    EventHeaderState::Full
+                } else if previous_timestamp
+                    .and_then(|ts| current.origin_server_ts().0.checked_sub(ts.0))
+                    .is_some_and(|elapsed| u64::from(elapsed) >= MAX_TIME_BETWEEN_HEADERS)
+                {
+                    // Too much time has passed, show the timestamp.
+                    EventHeaderState::TimestampOnly
+                } else {
+                    // Do not show header.
+                    EventHeaderState::Hidden
+                };
+
+                current.set_header_state(header_state);
+                previous_sender = Some(current_sender);
+                previous_timestamp = Some(current.origin_server_ts());
             }
         }
 
@@ -501,8 +639,8 @@ mod imp {
                 for id in identifiers {
                     // We check if we are removing the right event, in case we receive a diff that
                     // adds an existing event to another place, making us create a new event, before
-                    // another diff that removes it from its old place, making
-                    // us remove the old event.
+                    // another diff that removes it from its old place, making us remove the old
+                    // event.
                     let found = event_map.get(&id).is_some_and(|e| e == event);
 
                     if found {
@@ -511,48 +649,48 @@ mod imp {
                     }
                 }
 
-                if removed_from_map && event.is_room_create_event() {
+                if removed_from_map && event.is_room_create() {
                     self.set_has_room_create(false);
                 }
             }
         }
 
-        /// Whether we can load more events with the current state of the
-        /// timeline.
-        pub(super) fn can_load(&self) -> bool {
-            // We don't want to load twice at the same time, and it's useless to try to load
-            // more history before the timeline is ready or when we reached the
-            // start.
-            !matches!(
-                self.state.get(),
-                TimelineState::Initial | TimelineState::Loading | TimelineState::Complete
-            )
+        /// Whether we can load more events at the start of the timeline with
+        /// the current state.
+        pub(super) fn can_paginate_backwards(&self) -> bool {
+            // We do not want to load twice at the same time, and it's useless to try to
+            // load more history before the timeline is ready or if we have
+            // reached the start of the timeline.
+            self.state.get() != LoadingState::Initial
+                && !self.is_loading_start.get()
+                && !self.has_reached_start.get()
         }
 
         /// Load more events at the start of the timeline until the given
         /// function tells us to stop.
-        pub(super) async fn load<F>(&self, continue_fn: F)
+        pub(super) async fn paginate_backwards<F>(&self, continue_fn: F)
         where
             F: Fn() -> ControlFlow<()>,
         {
-            self.set_state(TimelineState::Loading);
+            self.set_loading_start(true);
 
             loop {
-                if !self.load_inner().await {
-                    return;
+                if !self.paginate_backwards_inner().await {
+                    break;
                 }
 
                 if continue_fn().is_break() {
-                    self.set_state(TimelineState::Ready);
-                    return;
+                    break;
                 }
             }
+
+            self.set_loading_start(false);
         }
 
         /// Load more events at the start of the timeline.
         ///
         /// Returns `true` if more events can be loaded.
-        async fn load_inner(&self) -> bool {
+        async fn paginate_backwards_inner(&self) -> bool {
             let matrix_timeline = self.matrix_timeline().clone();
             let handle =
                 spawn_tokio!(
@@ -562,62 +700,31 @@ mod imp {
             match handle.await.expect("task was not aborted") {
                 Ok(reached_start) => {
                     if reached_start {
-                        self.set_state(TimelineState::Complete);
+                        self.set_has_reached_start(true);
                     }
 
                     !reached_start
                 }
                 Err(error) => {
                     error!("Could not load timeline: {error}");
-                    self.set_state(TimelineState::Error);
+                    self.set_state(LoadingState::Error);
                     false
                 }
             }
         }
 
-        /// Set the state of the timeline.
-        pub(super) fn set_state(&self, state: TimelineState) {
-            if self.state.get() == state {
-                return;
-            }
-
-            self.state.set(state);
-
-            let start_items = &self.start_items;
-            let removed = start_items.n_items();
-
-            match state {
-                TimelineState::Loading => start_items.splice(0, removed, &[VirtualItem::spinner()]),
-                TimelineState::Complete => {
-                    start_items.splice(0, removed, &[VirtualItem::timeline_start()]);
-                }
-                _ => start_items.remove_all(),
-            }
-
-            self.obj().notify_state();
-        }
-
-        /// Whether the timeline has a typing row.
-        fn has_typing_row(&self) -> bool {
-            self.end_items.n_items() > 0
-        }
-
         /// Add the typing row to the timeline, if it isn't present already.
         fn add_typing_row(&self) {
-            if self.has_typing_row() {
-                return;
-            }
-
-            self.end_items.append(&VirtualItem::typing());
+            self.end_items().set_is_hidden(false);
         }
 
         /// Remove the typing row from the timeline.
         pub(super) fn remove_empty_typing_row(&self) {
-            if !self.has_typing_row() || !self.room().typing_list().is_empty() {
+            if !self.room().typing_list().is_empty() {
                 return;
             }
 
-            self.end_items.remove_all();
+            self.end_items().set_is_hidden(true);
         }
 
         /// Listen to read receipts changes.
@@ -657,12 +764,12 @@ mod imp {
         }
     }
 
-    impl TimelineItemStore for Timeline {
+    impl TimelineDiffItemStore for Timeline {
         type Item = TimelineItem;
         type Data = Arc<SdkTimelineItem>;
 
         fn items(&self) -> Vec<TimelineItem> {
-            self.sdk_items
+            self.sdk_items()
                 .snapshot()
                 .into_iter()
                 .map(|obj| {
@@ -673,8 +780,7 @@ mod imp {
         }
 
         fn create_item(&self, data: &Arc<SdkTimelineItem>) -> TimelineItem {
-            let room = self.room();
-            let item = TimelineItem::new(data, room);
+            let item = TimelineItem::new(data, &self.obj());
 
             if let Some(event) = item.downcast_ref::<Event>() {
                 self.event_map
@@ -683,13 +789,13 @@ mod imp {
 
                 // Keep track of the activity of the sender.
                 if event.counts_as_unread() {
-                    if let Some(members) = room.members() {
+                    if let Some(members) = self.room().members() {
                         let member = members.get_or_create(event.sender_id());
                         member.set_latest_activity(u64::from(event.origin_server_ts().get()));
                     }
                 }
 
-                if event.is_room_create_event() {
+                if event.is_room_create() {
                     self.set_has_room_create(true);
                 }
             }
@@ -712,19 +818,116 @@ mod imp {
             }
         }
 
-        fn apply_item_diff_list(&self, item_diff_list: Vec<TimelineItemDiff<TimelineItem>>) {
+        fn apply_item_diff_list(&self, item_diff_list: Vec<TimelineDiff<TimelineItem>>) {
             for item_diff in item_diff_list {
                 match item_diff {
-                    TimelineItemDiff::Splice(splice) => {
+                    TimelineDiff::Splice(splice) => {
                         self.update_items(splice.pos, splice.n_removals, &splice.additions);
                     }
-                    TimelineItemDiff::Update(update) => {
+                    TimelineDiff::Update(update) => {
                         self.update_items_headers(update.pos, update.n_items);
                     }
                 }
             }
         }
     }
+
+    /// The default log filter initialized with the `RUST_LOG` environment
+    /// variable.
+    ///
+    /// Used to know if we are likely to need to log the diff.
+    static IS_AT_TRACE_LEVEL: LazyLock<bool> = LazyLock::new(|| {
+        tracing_subscriber::EnvFilter::try_from_default_env()
+            // If the env variable is not set, we know that we are not at trace level.
+            .ok()
+            .and_then(|filter| filter.max_level_hint())
+            .is_some_and(|max| max == tracing::level_filters::LevelFilter::TRACE)
+    });
+
+    /// Temporary methods to debug items in the timeline.
+    impl Timeline {
+        /// Log the given diff list.
+        fn log_diff_list(&self, diff_list: &[VectorDiff<Arc<SdkTimelineItem>>]) {
+            let mut log_list = Vec::with_capacity(diff_list.len());
+
+            for diff in diff_list {
+                let log = match diff {
+                    VectorDiff::Append { values } => {
+                        format!("append: {:?}", sdk_items_to_log(values))
+                    }
+                    VectorDiff::Clear => "clear".to_owned(),
+                    VectorDiff::PushFront { value } => {
+                        format!("push_front: {}", sdk_item_to_log(value))
+                    }
+                    VectorDiff::PushBack { value } => {
+                        format!("push_back: {}", sdk_item_to_log(value))
+                    }
+                    VectorDiff::PopFront => "pop_front".to_owned(),
+                    VectorDiff::PopBack => "pop_back".to_owned(),
+                    VectorDiff::Insert { index, value } => {
+                        format!("insert at {index}: {}", sdk_item_to_log(value))
+                    }
+                    VectorDiff::Set { index, value } => {
+                        format!("set at {index}: {}", sdk_item_to_log(value))
+                    }
+                    VectorDiff::Remove { index } => format!("remove at {index}"),
+                    VectorDiff::Truncate { length } => format!("truncate at {length}"),
+                    VectorDiff::Reset { values } => {
+                        format!("reset: {:?}", sdk_items_to_log(values))
+                    }
+                };
+
+                log_list.push(log);
+            }
+
+            tracing::trace!(
+                room = self.room().human_readable_id(),
+                "Diff list: {log_list:#?}"
+            );
+        }
+
+        /// Log the items in this timeline.
+        fn log_items(&self) {
+            let items = self
+                .sdk_items()
+                .iter::<TimelineItem>()
+                .filter_map(|item| item.as_ref().map(item_to_log).ok())
+                .collect::<Vec<_>>();
+
+            tracing::trace!(
+                room = self.room().human_readable_id(),
+                "Timeline: {items:#?}"
+            );
+        }
+    }
+
+    // Helper methods for logging items.
+    fn sdk_items_to_log(
+        items: &matrix_sdk_ui::eyeball_im::Vector<Arc<SdkTimelineItem>>,
+    ) -> Vec<String> {
+        items.iter().map(|item| sdk_item_to_log(item)).collect()
+    }
+
+    fn sdk_item_to_log(item: &SdkTimelineItem) -> String {
+        match item.kind() {
+            matrix_sdk_ui::timeline::TimelineItemKind::Event(event) => {
+                format!("event::{:?}", event.identifier())
+            }
+            matrix_sdk_ui::timeline::TimelineItemKind::Virtual(virtual_item) => {
+                format!("virtual::{virtual_item:?}")
+            }
+        }
+    }
+
+    fn item_to_log(item: &TimelineItem) -> String {
+        if let Some(virtual_item) = item.downcast_ref::<VirtualItem>() {
+            format!("virtual::{:?}", virtual_item.kind())
+        } else if let Some(event) = item.downcast_ref::<Event>() {
+            format!("event::{:?}", event.identifier())
+        } else {
+            "Unknown item".to_owned()
+        }
+    }
 }
 
 glib::wrapper! {
@@ -762,17 +965,17 @@ impl Timeline {
 
     /// Load more events at the start of the timeline until the given function
     /// tells us to stop.
-    pub(crate) async fn load<F>(&self, continue_fn: F)
+    pub(crate) async fn paginate_backwards<F>(&self, continue_fn: F)
     where
         F: Fn() -> ControlFlow<()>,
     {
         let imp = self.imp();
 
-        if !imp.can_load() {
+        if !imp.can_paginate_backwards() {
             return;
         }
 
-        imp.load(continue_fn).await;
+        imp.paginate_backwards(continue_fn).await;
     }
 
     /// Get the event with the given identifier from this `Timeline`.
@@ -786,24 +989,15 @@ impl Timeline {
     /// Get the position of the event with the given identifier in this
     /// `Timeline`.
     pub(crate) fn find_event_position(&self, identifier: &TimelineEventItemId) -> Option<usize> {
-        for (pos, item) in self
-            .items()
+        self.items()
             .iter::<glib::Object>()
-            .map(|o| o.ok().and_downcast::<TimelineItem>())
             .enumerate()
-        {
-            let Some(item) = item else {
-                break;
-            };
-
-            if let Some(event) = item.downcast_ref::<Event>() {
-                if event.matches_identifier(identifier) {
-                    return Some(pos);
-                }
-            }
-        }
-
-        None
+            .find_map(|(index, item)| {
+                item.ok()
+                    .and_downcast::<Event>()
+                    .is_some_and(|event| event.matches_identifier(identifier))
+                    .then_some(index)
+            })
     }
 
     /// Remove the typing row from the timeline.
@@ -828,7 +1022,7 @@ impl Timeline {
         .await
         .expect("task was not aborted");
 
-        let sdk_items = &self.imp().sdk_items;
+        let sdk_items = self.imp().sdk_items();
         let count = sdk_items.n_items();
 
         for pos in (0..count).rev() {
@@ -856,7 +1050,7 @@ impl Timeline {
     pub(crate) fn redactable_events_for(&self, user_id: &UserId) -> Vec<OwnedEventId> {
         let mut events = vec![];
 
-        for item in self.imp().sdk_items.iter::<glib::Object>() {
+        for item in self.imp().sdk_items().iter::<glib::Object>() {
             let Ok(item) = item else {
                 // The iterator is broken.
                 break;
