@@ -1,31 +1,25 @@
 use std::time::Duration;
 
-use futures_util::StreamExt;
+use futures_util::{lock::Mutex, StreamExt};
 use gettextrs::gettext;
-use gtk::{
-    gio, glib,
-    glib::{clone, signal::SignalHandlerId},
-    prelude::*,
-    subclass::prelude::*,
-};
+use gtk::{gio, glib, glib::clone, prelude::*, subclass::prelude::*};
 use matrix_sdk::{
-    authentication::matrix::MatrixSession, config::SyncSettings, sync::SyncResponse, Client,
+    config::SyncSettings, media::MediaRetentionPolicy, sync::SyncResponse, Client, SessionChange,
 };
 use ruma::{
     api::client::{
-        error::ErrorKind,
         filter::{FilterDefinition, RoomFilter},
         search::search_events::v3::UserProfile,
     },
     assign,
 };
-use tokio::task::AbortHandle;
-use tracing::{debug, error, warn};
-use url::Url;
+use tokio::{task::AbortHandle, time::sleep};
+use tokio_stream::wrappers::BroadcastStream;
+use tracing::{debug, error, info};
 
 use super::{
-    IgnoredUsers, Notifications, RoomList, SessionSecurity, SessionSettings, SidebarItemList,
-    SidebarListModel, User, UserSessionsList, VerificationList,
+    IgnoredUsers, Notifications, RemoteCache, RoomList, SessionSecurity, SessionSettings,
+    SidebarItemList, SidebarListModel, User, UserSessionsList, VerificationList,
 };
 use crate::{
     components::AvatarData,
@@ -35,7 +29,7 @@ use crate::{
     spawn, spawn_tokio,
     utils::{
         matrix::{self, ClientSetupError},
-        oauth, TokioDrop,
+        TokioDrop,
     },
     Application,
 };
@@ -44,7 +38,13 @@ use crate::{
 const SESSION_PROFILE_KEY: &str = "session_profile";
 /// The number of consecutive missed synchronizations before the session is
 /// marked as offline.
-const MISSED_SYNC_MAX_COUNT: u8 = 3;
+///
+/// Note that this is set to `2`, but the count begins at `0` so this would
+/// match the third missed synchronization.
+const MISSED_SYNC_OFFLINE_COUNT: usize = 2;
+/// The delays in seconds to wait for when a sync fails, depending on the number
+/// of missed attempts.
+const MISSED_SYNC_DELAYS: &[u64] = &[1, 5, 10, 20, 30];
 
 /// The state of the session.
 #[derive(Debug, Default, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, glib::Enum)]
@@ -65,8 +65,6 @@ pub struct BoxedClient(Client);
 mod imp {
     use std::cell::{Cell, OnceCell, RefCell};
 
-    use async_once_cell::OnceCell as AsyncOnceCell;
-
     use super::*;
 
     #[derive(Debug, Default, glib::Properties)]
@@ -105,17 +103,17 @@ mod imp {
         /// Information about security for this session.
         #[property(get)]
         security: SessionSecurity,
+        /// The cache for remote data.
+        remote_cache: OnceCell<RemoteCache>,
+        session_changes_handle: RefCell<Option<AbortHandle>>,
         sync_handle: RefCell<Option<AbortHandle>>,
-        network_monitor_handler_id: RefCell<Option<SignalHandlerId>>,
+        network_monitor_handler_id: RefCell<Option<glib::SignalHandlerId>>,
+        homeserver_reachable_lock: Mutex<()>,
+        homeserver_reachable_source: RefCell<Option<glib::SourceId>>,
         /// The number of missed synchonizations in a row.
         ///
-        /// Capped at `MISSED_SYNC_MAX_COUNT - 1`.
-        missed_sync_count: Cell<u8>,
-        /// The OIDC authentication issuer, if any.
-        auth_issuer: AsyncOnceCell<Url>,
-        /// The account management URL of the OIDC authentication issuer, if
-        /// any.
-        account_management_url: AsyncOnceCell<Url>,
+        /// Capped at `MISSED_SYNC_DELAYS.len() - 1`.
+        missed_sync_count: Cell<usize>,
     }
 
     #[glib::object_subclass]
@@ -154,6 +152,14 @@ mod imp {
                 gio::NetworkMonitor::default().disconnect(handler_id);
             }
 
+            if let Some(source) = self.homeserver_reachable_source.take() {
+                source.remove();
+            }
+
+            if let Some(handle) = self.session_changes_handle.take() {
+                handle.abort();
+            }
+
             if let Some(handle) = self.sync_handle.take() {
                 handle.abort();
             }
@@ -218,20 +224,45 @@ mod imp {
             self.obj().notify_state();
         }
 
-        /// Update whether the homeserver is reachable.
-        pub(super) async fn update_homeserver_reachable(&self) {
+        /// The homeserver URL as a `GNetworkAddress`.
+        fn homeserver_address(&self) -> gio::NetworkAddress {
             let obj = self.obj();
-            let monitor = gio::NetworkMonitor::default();
+            let homeserver = obj.homeserver();
+            let default_port = if homeserver.scheme() == "http" {
+                80
+            } else {
+                443
+            };
 
-            let is_homeserver_reachable = if monitor.is_network_available() {
-                let homeserver = obj.homeserver();
-                let address = gio::NetworkAddress::parse_uri(homeserver.as_ref(), 80)
-                    .expect("url is parsed successfully");
+            gio::NetworkAddress::parse_uri(homeserver.as_str(), default_port)
+                .expect("url is parsed successfully")
+        }
+
+        /// Check whether the homeserver is reachable.
+        pub(super) async fn update_homeserver_reachable(&self) {
+            // If there is a timeout, remove it, we will add a new one later if needed.
+            if let Some(source) = self.homeserver_reachable_source.take() {
+                source.remove();
+            }
+            let Some(_guard) = self.homeserver_reachable_lock.try_lock() else {
+                // There is an ongoing check.
+                return;
+            };
+
+            let monitor = gio::NetworkMonitor::default();
+            let is_network_available = monitor.is_network_available();
+
+            let is_homeserver_reachable = if is_network_available {
+                // Check if we can reach the homeserver.
+                let address = self.homeserver_address();
 
                 match monitor.can_reach_future(&address).await {
                     Ok(()) => true,
                     Err(error) => {
-                        error!("Homeserver {homeserver} is not reachable: {error}");
+                        error!(
+                            session = self.obj().session_id(),
+                            "Homeserver is not reachable: {error}"
+                        );
                         false
                     }
                 }
@@ -239,17 +270,44 @@ mod imp {
                 false
             };
 
-            if self.is_homeserver_reachable.get() == is_homeserver_reachable {
+            self.set_is_homeserver_reachable(is_homeserver_reachable);
+
+            if is_network_available && !is_homeserver_reachable {
+                // Check again later if the homeserver is reachable.
+                let source = glib::timeout_add_seconds_local_once(
+                    10,
+                    clone!(
+                        #[weak(rename_to = imp)]
+                        self,
+                        move || {
+                            imp.homeserver_reachable_source.take();
+
+                            spawn!(async move {
+                                imp.update_homeserver_reachable().await;
+                            });
+                        }
+                    ),
+                );
+                self.homeserver_reachable_source.replace(Some(source));
+            }
+        }
+
+        /// Set whether the homeserver is reachable.
+        fn set_is_homeserver_reachable(&self, is_reachable: bool) {
+            if self.is_homeserver_reachable.get() == is_reachable {
                 return;
             }
+            let obj = self.obj();
 
-            self.is_homeserver_reachable.set(is_homeserver_reachable);
+            self.is_homeserver_reachable.set(is_reachable);
 
             if let Some(handle) = self.sync_handle.take() {
                 handle.abort();
             }
 
-            if is_homeserver_reachable {
+            if is_reachable {
+                info!(session = obj.session_id(), "Homeserver is reachable");
+
                 // Restart the sync loop.
                 self.sync();
             } else {
@@ -277,6 +335,12 @@ mod imp {
             self.obj().notify_is_offline();
         }
 
+        /// The cache for remote data.
+        pub(crate) fn remote_cache(&self) -> &RemoteCache {
+            self.remote_cache
+                .get_or_init(|| RemoteCache::new(self.obj().clone()))
+        }
+
         /// Finish initialization of this session.
         pub(super) async fn prepare(&self) {
             spawn!(
@@ -292,6 +356,7 @@ mod imp {
                     }
                 )
             );
+            self.watch_session_changes();
             self.update_homeserver_reachable().await;
 
             self.room_list().load().await;
@@ -309,7 +374,50 @@ mod imp {
             self.set_state(SessionState::InitialSync);
             self.sync();
 
-            debug!("A new session was prepared");
+            debug!(
+                session = self.obj().session_id(),
+                "A new session was prepared"
+            );
+        }
+
+        /// Watch the changes of the session, like being logged out or the
+        /// tokens being refreshed.
+        fn watch_session_changes(&self) {
+            let receiver = self.client().subscribe_to_session_changes();
+            let stream = BroadcastStream::new(receiver);
+
+            let obj_weak = glib::SendWeakRef::from(self.obj().downgrade());
+            let fut = stream.for_each(move |change| {
+                let obj_weak = obj_weak.clone();
+                async move {
+                    let Ok(change) = change else {
+                        return;
+                    };
+
+                    let ctx = glib::MainContext::default();
+                    ctx.spawn(async move {
+                        spawn!(async move {
+                            if let Some(obj) = obj_weak.upgrade() {
+                                match change {
+                                    SessionChange::UnknownToken { .. } => {
+                                        info!(
+                                            session = obj.session_id(),
+                                            "The access token is invalid, cleaning up the session…"
+                                        );
+                                        obj.imp().clean_up().await;
+                                    }
+                                    SessionChange::TokensRefreshed => {
+                                        obj.imp().store_tokens().await;
+                                    }
+                                }
+                            }
+                        });
+                    });
+                }
+            });
+
+            let handle = spawn_tokio!(fut).abort_handle();
+            self.session_changes_handle.replace(Some(handle));
         }
 
         /// Start syncing the Matrix client.
@@ -337,13 +445,24 @@ mod imp {
                 while let Some(response) = sync_stream.next().await {
                     let obj_weak = obj_weak.clone();
                     let ctx = glib::MainContext::default();
-                    ctx.spawn(async move {
-                        spawn!(async move {
-                            if let Some(obj) = obj_weak.upgrade() {
-                                obj.imp().handle_sync_response(response).await;
-                            }
-                        });
-                    });
+                    let delay = ctx
+                        .spawn(async move {
+                            spawn!(async move {
+                                if let Some(obj) = obj_weak.upgrade() {
+                                    obj.imp().handle_sync_response(response)
+                                } else {
+                                    None
+                                }
+                            })
+                            .await
+                            .expect("task was not aborted")
+                        })
+                        .await
+                        .expect("task was not aborted");
+
+                    if let Some(delay) = delay {
+                        sleep(delay).await;
+                    }
                 }
             })
             .abort_handle();
@@ -352,8 +471,16 @@ mod imp {
         }
 
         /// Handle the response received via sync.
-        async fn handle_sync_response(&self, response: Result<SyncResponse, matrix_sdk::Error>) {
-            debug!("Received sync response");
+        ///
+        /// Returns the delay to wait for before making the next sync, if
+        /// necessary.
+        fn handle_sync_response(
+            &self,
+            response: Result<SyncResponse, matrix_sdk::Error>,
+        ) -> Option<Duration> {
+            let obj = self.obj();
+            let session_id = obj.session_id();
+            debug!(session = session_id, "Received sync response");
 
             match response {
                 Ok(response) => {
@@ -366,20 +493,27 @@ mod imp {
 
                     self.set_offline(false);
                     self.missed_sync_count.set(0);
+
+                    None
                 }
                 Err(error) => {
-                    if let Some(ErrorKind::UnknownToken { .. }) = error.client_api_error_kind() {
-                        self.clean_up().await;
-                    } else {
-                        let missed_sync_count = self.missed_sync_count.get() + 1;
+                    let missed_sync_count = self.missed_sync_count.get();
 
-                        if missed_sync_count >= MISSED_SYNC_MAX_COUNT {
-                            self.set_offline(true);
-                        } else {
-                            self.missed_sync_count.set(missed_sync_count);
-                        }
+                    // If there are too many failed attempts, mark the session as offline.
+                    if missed_sync_count == MISSED_SYNC_OFFLINE_COUNT {
+                        self.set_offline(true);
                     }
-                    error!("Could not perform sync: {error}");
+
+                    // Increase the count of missed syncs, if we have not reached the maximum value.
+                    if missed_sync_count < 4 {
+                        self.missed_sync_count.set(missed_sync_count + 1);
+                    }
+
+                    error!(session = session_id, "Could not perform sync: {error}");
+
+                    // Sleep a little between attempts.
+                    let delay = MISSED_SYNC_DELAYS[missed_sync_count];
+                    Some(Duration::from_secs(delay))
                 }
             }
         }
@@ -389,7 +523,7 @@ mod imp {
             let client = self.client().clone();
             let handle = spawn_tokio!(async move {
                 client
-                    .store()
+                    .state_store()
                     .get_custom_value(SESSION_PROFILE_KEY.as_bytes())
                     .await
             });
@@ -398,13 +532,19 @@ mod imp {
                 Ok(Some(bytes)) => match serde_json::from_slice::<UserProfile>(&bytes) {
                     Ok(profile) => profile,
                     Err(error) => {
-                        error!("Failed to deserialize session profile: {error}");
+                        error!(
+                            session = self.obj().session_id(),
+                            "Failed to deserialize session profile: {error}"
+                        );
                         return;
                     }
                 },
                 Ok(None) => return,
                 Err(error) => {
-                    error!("Could not load cached session profile: {error}");
+                    error!(
+                        session = self.obj().session_id(),
+                        "Could not load cached session profile: {error}"
+                    );
                     return;
                 }
             };
@@ -432,7 +572,10 @@ mod imp {
                     profile
                 }
                 Err(error) => {
-                    error!("Could not fetch session profile: {error}");
+                    error!(
+                        session = self.obj().session_id(),
+                        "Could not fetch session profile: {error}"
+                    );
                     return;
                 }
             };
@@ -460,21 +603,27 @@ mod imp {
             let value = match value {
                 Ok(value) => value,
                 Err(error) => {
-                    error!("Failed to serialize session profile: {error}");
+                    error!(
+                        session = self.obj().session_id(),
+                        "Failed to serialize session profile: {error}"
+                    );
                     return;
                 }
             };
 
             let handle = spawn_tokio!(async move {
                 client
-                    .store()
+                    .state_store()
                     .set_custom_value(SESSION_PROFILE_KEY.as_bytes(), value)
                     .await
             });
 
             if let Err(error) = handle.await.expect("task was not aborted") {
-                error!("Could not cache session profile: {error}");
-            };
+                error!(
+                    session = self.obj().session_id(),
+                    "Could not cache session profile: {error}"
+                );
+            }
         }
 
         /// Start listening to notifications.
@@ -501,11 +650,25 @@ mod imp {
             });
         }
 
+        /// Update the stored session tokens.
+        async fn store_tokens(&self) {
+            let Some(session_tokens) = self.client().session_tokens() else {
+                return;
+            };
+
+            debug!(
+                session = self.obj().session_id(),
+                "Storing updated session tokens…"
+            );
+            self.obj().info().store_tokens(session_tokens).await;
+        }
+
         /// Clean up this session after it was logged out.
         ///
         /// This should only be called if the session has been logged out
         /// without calling `Session::log_out`.
         pub(super) async fn clean_up(&self) {
+            let obj = self.obj();
             self.set_state(SessionState::LoggedOut);
 
             if let Some(handle) = self.sync_handle.take() {
@@ -516,57 +679,14 @@ mod imp {
                 settings.delete();
             }
 
-            self.obj().info().clone().delete().await;
+            obj.info().clone().delete().await;
 
             self.notifications.clear();
 
-            debug!("The logged out session was cleaned up");
-        }
-
-        /// The OAuth 2.0 authorization provider, if any.
-        async fn auth_issuer(&self) -> Option<&Url> {
-            self.auth_issuer
-                .get_or_try_init(clone!(
-                    #[strong(rename_to = imp)]
-                    self,
-                    async move {
-                        let client = imp.client().clone();
-
-                        spawn_tokio!(
-                            async move { oauth::fetch_auth_issuer(&client).await.ok_or(()) }
-                        )
-                        .await
-                        .expect("task was not aborted")
-                    }
-                ))
-                .await
-                .ok()
-        }
-
-        /// The account management URL of the OAuth 2.0 authorization provider,
-        /// if any.
-        pub(super) async fn account_management_url(&self) -> Option<&Url> {
-            self.account_management_url
-                .get_or_try_init(clone!(
-                    #[strong(rename_to = imp)]
-                    self,
-                    async move {
-                        let auth_issuer = imp.auth_issuer().await.ok_or(())?.clone();
-
-                        let client = imp.client().clone();
-                        spawn_tokio!(async move {
-                            oauth::discover_account_management_url(&client, auth_issuer)
-                                .await
-                                .map_err(|error| {
-                                    warn!("Could not discover account management URL: {error}");
-                                })
-                        })
-                        .await
-                        .expect("task was not aborted")
-                    }
-                ))
-                .await
-                .ok()
+            debug!(
+                session = obj.session_id(),
+                "The logged out session was cleaned up"
+            );
         }
     }
 }
@@ -578,35 +698,52 @@ glib::wrapper! {
 }
 
 impl Session {
-    /// Create a new session.
-    pub async fn new(homeserver: Url, data: MatrixSession) -> Result<Self, ClientSetupError> {
-        let stored_session = StoredSession::with_login_data(homeserver, data)?;
+    /// Construct an existing session.
+    pub(crate) async fn new(
+        stored_session: StoredSession,
+        settings: SessionSettings,
+    ) -> Result<Self, ClientSetupError> {
+        let tokens = stored_session
+            .load_tokens()
+            .await
+            .ok_or(ClientSetupError::NoSessionTokens)?;
+
+        let stored_session_clone = stored_session.clone();
+        let client = spawn_tokio!(async move {
+            let client = matrix::client_with_stored_session(stored_session_clone, tokens).await?;
+
+            // Make sure that we use the proper retention policy.
+            let media = client.media();
+            let used_media_retention_policy = media.media_retention_policy().await?;
+            let wanted_media_retention_policy = MediaRetentionPolicy::default();
+
+            if used_media_retention_policy != wanted_media_retention_policy {
+                media
+                    .set_media_retention_policy(wanted_media_retention_policy)
+                    .await?;
+            }
+
+            Ok::<_, ClientSetupError>(client)
+        })
+        .await
+        .expect("task was not aborted")?;
+
+        Ok(glib::Object::builder()
+            .property("info", stored_session)
+            .property("settings", settings)
+            .property("client", BoxedClient(client))
+            .build())
+    }
+
+    /// Create a new session from the session of the given Matrix client.
+    pub(crate) async fn create(client: &Client) -> Result<Self, ClientSetupError> {
+        let stored_session = StoredSession::new(client).await?;
         let settings = Application::default()
             .session_list()
             .settings()
             .get_or_create(&stored_session.id);
 
-        Self::restore(stored_session, settings).await
-    }
-
-    /// Restore a stored session.
-    pub(crate) async fn restore(
-        stored_session: StoredSession,
-        settings: SessionSettings,
-    ) -> Result<Self, ClientSetupError> {
-        let stored_session_clone = stored_session.clone();
-        let client =
-            spawn_tokio!(
-                async move { matrix::client_with_stored_session(stored_session_clone).await }
-            )
-            .await
-            .expect("task was not aborted")?;
-
-        Ok(glib::Object::builder()
-            .property("info", &stored_session)
-            .property("settings", settings)
-            .property("client", BoxedClient(client))
-            .build())
+        Self::new(stored_session, settings).await
     }
 
     /// Finish initialization of this session.
@@ -629,25 +766,31 @@ impl Session {
         self.imp().client().clone()
     }
 
-    /// The account management URL of the OIDC authentication issuer, if any.
-    pub(crate) async fn account_management_url(&self) -> Option<&Url> {
-        self.imp().account_management_url().await
+    /// The cache for remote data.
+    pub(crate) fn remote_cache(&self) -> &RemoteCache {
+        self.imp().remote_cache()
     }
 
     /// Log out of this session.
     pub(crate) async fn log_out(&self) -> Result<(), String> {
-        debug!("The session is about to be logged out");
+        debug!(
+            session = self.session_id(),
+            "The session is about to be logged out"
+        );
 
         let client = self.client();
-        let handle = spawn_tokio!(async move { client.matrix_auth().logout().await });
+        let handle = spawn_tokio!(async move { client.logout().await });
 
         match handle.await.expect("task was not aborted") {
-            Ok(_) => {
+            Ok(()) => {
                 self.imp().clean_up().await;
                 Ok(())
             }
             Err(error) => {
-                error!("Could not log the session out: {error}");
+                error!(
+                    session = self.session_id(),
+                    "Could not log the session out: {error}"
+                );
                 Err(gettext("Could not log the session out"))
             }
         }
