@@ -1,9 +1,7 @@
 use adw::{prelude::*, subclass::prelude::*};
 use gettextrs::gettext;
 use gtk::{gdk, glib, glib::clone};
-use matrix_sdk_ui::timeline::{
-    Message, RepliedToInfo, ReplyContent, TimelineDetails, TimelineItemContent,
-};
+use matrix_sdk_ui::timeline::{MsgLikeKind, TimelineDetails, TimelineItemContent};
 use ruma::{events::room::message::MessageType, OwnedEventId, OwnedTransactionId};
 use tracing::{error, warn};
 
@@ -13,7 +11,10 @@ use super::{
 };
 use crate::{
     prelude::*,
-    session::model::{Event, Member, Room, Session},
+    session::{
+        model::{Event, Member, Room},
+        view::content::room_history::message_toolbar::MessageEventSource,
+    },
     spawn,
     utils::matrix::MediaMessage,
 };
@@ -40,7 +41,7 @@ pub enum ContentFormat {
 }
 
 mod imp {
-    use std::cell::Cell;
+    use std::{cell::Cell, marker::PhantomData};
 
     use super::*;
 
@@ -50,6 +51,12 @@ mod imp {
         /// The displayed format of the message.
         #[property(get, set = Self::set_format, explicit_notify, builder(ContentFormat::default()))]
         format: Cell<ContentFormat>,
+        /// The texture of the image preview displayed by the descendant of this
+        /// widget, if any.
+        #[property(get = Self::texture)]
+        texture: PhantomData<Option<gdk::Texture>>,
+        /// The widget with the visual media content of the event, if any.
+        pub(super) visual_media_widget: glib::WeakRef<MessageVisualMedia>,
     }
 
     #[glib::object_subclass]
@@ -75,6 +82,62 @@ mod imp {
             self.format.set(format);
             self.obj().notify_format();
         }
+
+        /// The texture of the image preview displayed by the descendant of this
+        /// widget, if any.
+        fn texture(&self) -> Option<gdk::Texture> {
+            self.visual_media_widget.upgrade()?.texture()
+        }
+
+        /// Update the current visual media widget if necessary.
+        pub(super) fn update_visual_media_widget(&self) {
+            let prev_widget = self.visual_media_widget.upgrade();
+            let current_widget = self.visual_media_widget();
+
+            if prev_widget == current_widget {
+                return;
+            }
+
+            let obj = self.obj();
+
+            if let Some(visual_media) = &current_widget {
+                visual_media.connect_texture_notify(clone!(
+                    #[weak]
+                    obj,
+                    move |_| {
+                        obj.notify_texture();
+                    }
+                ));
+            }
+            self.visual_media_widget.set(current_widget.as_ref());
+
+            let prev_texture = prev_widget.and_then(|visual_media| visual_media.texture());
+            let current_texture = current_widget.and_then(|visual_media| visual_media.texture());
+            if prev_texture != current_texture {
+                obj.notify_texture();
+            }
+        }
+
+        /// The widget with the visual media content of the event, if any.
+        ///
+        /// This allows to access the descendant content while discarding the
+        /// content of a related message, like a replied-to event, or the
+        /// caption of the event.
+        fn visual_media_widget(&self) -> Option<MessageVisualMedia> {
+            let mut child = self.obj().child()?;
+
+            // If it is a reply, the media is in the main content.
+            if let Some(reply) = child.downcast_ref::<MessageReply>() {
+                child = reply.content().child()?;
+            }
+
+            // If it is a caption, the media is the child of the caption.
+            if let Some(caption) = child.downcast_ref::<MessageCaption>() {
+                child = caption.child()?;
+            }
+
+            child.downcast::<MessageVisualMedia>().ok()
+        }
     }
 }
 
@@ -89,25 +152,9 @@ impl MessageContent {
         glib::Object::new()
     }
 
-    /// Access the widget with the visual media content of the event, if any.
-    ///
-    /// This allows to access the descendant content while discarding the
-    /// content of a related message, like a replied-to event, or the caption of
-    /// the event.
+    /// The widget with the visual media content of the event, if any.
     pub(crate) fn visual_media_widget(&self) -> Option<MessageVisualMedia> {
-        let mut child = BinExt::child(self)?;
-
-        // If it is a reply, the media is in the main content.
-        if let Some(reply) = child.downcast_ref::<MessageReply>() {
-            child = BinExt::child(reply.content())?;
-        }
-
-        // If it is a caption, the media is the child of the caption.
-        if let Some(caption) = child.downcast_ref::<MessageCaption>() {
-            child = caption.child()?;
-        }
-
-        child.downcast::<MessageVisualMedia>().ok()
+        self.imp().visual_media_widget.upgrade()
     }
 
     /// Update this widget to present the given `Event`.
@@ -135,7 +182,9 @@ impl MessageContent {
                     TimelineDetails::Error(error) => {
                         error!(
                             "Could not fetch replied to event '{}': {error}",
-                            event.reply_to_id().unwrap()
+                            event
+                                .reply_to_id()
+                                .expect("reply event should have replied-to event ID")
                         );
                     }
                     TimelineDetails::Ready(replied_to_event) => {
@@ -168,7 +217,9 @@ impl MessageContent {
                             event.transaction_id(),
                             event.event_id(),
                         );
-                        BinExt::set_child(self, Some(&reply));
+                        self.set_child(Some(&reply));
+
+                        self.imp().update_visual_media_widget();
 
                         return;
                     }
@@ -185,54 +236,37 @@ impl MessageContent {
             event.transaction_id(),
             event.event_id(),
         );
+
+        self.imp().update_visual_media_widget();
     }
 
     /// Update this widget to present the given related event.
-    pub(crate) fn update_for_related_event(&self, info: &RepliedToInfo, sender: &Member) {
-        let ReplyContent::Message(message) = info.content() else {
-            return;
-        };
-
-        let detect_at_room = message.can_contain_at_room() && sender.can_notify_room();
+    pub(crate) fn update_for_related_event(
+        &self,
+        msgtype: &MessageType,
+        message_event: &MessageEventSource,
+        sender: &Member,
+    ) {
+        let detect_at_room = message_event.can_contain_at_room() && sender.can_notify_room();
 
         self.build_message_content(
-            message,
+            msgtype,
             self.format(),
             sender,
             detect_at_room,
-            None,
-            Some(info.event_id().to_owned()),
+            MessageCacheKey {
+                transaction_id: None,
+                event_id: Some(message_event.event_id()),
+                is_edited: message_event.is_edited(),
+            },
         );
     }
-
-    /// Get the texture displayed by this widget, if any.
-    pub(crate) fn texture(&self) -> Option<gdk::Texture> {
-        self.visual_media_widget()?.texture()
-    }
 }
 
+impl IsABin for MessageContent {}
+
 /// Helper trait for types used to build a message's content.
-trait MessageContentContainer: IsA<gtk::Widget> {
-    /// Get the child of this widget
-    fn child(&self) -> Option<gtk::Widget>;
-
-    /// Set the child of this widget
-    fn set_child(&self, child: Option<gtk::Widget>);
-
-    /// Reuse the child of this widget if it is of the correct type `W`, or
-    /// replace it with a new `W` constructed with its `Default` implementation.
-    ///
-    /// Returns the reused or new widget.
-    fn reuse_child_or_default<W: IsA<gtk::Widget> + Clone + Default>(&self) -> W {
-        if let Some(child) = self.child().and_downcast::<W>() {
-            child
-        } else {
-            let child = W::default();
-            self.set_child(Some(child.clone().upcast()));
-            child
-        }
-    }
-
+trait MessageContentContainer: ChildPropertyExt {
     /// Build the content widget of `event` as a child of this widget.
     fn build_content(
         &self,
@@ -246,75 +280,80 @@ trait MessageContentContainer: IsA<gtk::Widget> {
         let room = sender.room();
 
         match content {
-            TimelineItemContent::Message(message) => {
-                self.build_message_content(
-                    &message,
-                    format,
-                    sender,
-                    detect_at_room,
-                    transaction_id,
-                    event_id,
-                );
-            }
-            TimelineItemContent::Sticker(sticker) => {
-                self.build_media_message_content(
-                    sticker.content().clone().into(),
-                    format,
-                    &room,
-                    detect_at_room,
-                    MessageCacheKey {
-                        transaction_id,
-                        event_id,
-                        is_edited: false,
-                    },
-                );
-            }
-            TimelineItemContent::UnableToDecrypt(_) => {
-                let child = self.reuse_child_or_default::<MessageText>();
-                child.with_plain_text(gettext("Could not decrypt this message, decryption will be retried once the keys are available."), format);
-            }
-            TimelineItemContent::RedactedMessage => {
-                let child = self.reuse_child_or_default::<MessageText>();
-                child.with_plain_text(gettext("This message was removed."), format);
-            }
+            TimelineItemContent::MsgLike(msg_like) => match msg_like.kind {
+                MsgLikeKind::Message(message) => {
+                    self.build_message_content(
+                        message.msgtype(),
+                        format,
+                        sender,
+                        detect_at_room,
+                        MessageCacheKey {
+                            transaction_id,
+                            event_id,
+                            is_edited: message.is_edited(),
+                        },
+                    );
+                }
+                MsgLikeKind::Sticker(sticker) => {
+                    self.build_media_message_content(
+                        sticker.content().clone().into(),
+                        format,
+                        &room,
+                        detect_at_room,
+                        MessageCacheKey {
+                            transaction_id,
+                            event_id,
+                            is_edited: false,
+                        },
+                    );
+                }
+                MsgLikeKind::UnableToDecrypt(_) => {
+                    let child = self.child_or_default::<MessageText>();
+                    child.with_plain_text(gettext("Could not decrypt this message, decryption will be retried once the keys are available."), format);
+                }
+                msg_like_kind => {
+                    warn!("Unsupported message-like event content: {msg_like_kind:?}");
+                    self.build_unsupported_content(format);
+                }
+            },
             content => {
                 warn!("Unsupported event content: {content:?}");
-                let child = self.reuse_child_or_default::<MessageText>();
-                child.with_plain_text(gettext("Unsupported event"), format);
+                self.build_unsupported_content(format);
             }
         }
     }
 
+    /// Build the widget of an unsupported content as a child of this widget.
+    fn build_unsupported_content(&self, format: ContentFormat) {
+        let child = self.child_or_default::<MessageText>();
+        child.with_plain_text(gettext("Unsupported event"), format);
+    }
+
     /// Build the content widget of the given message as a child of this widget.
     fn build_message_content(
         &self,
-        message: &Message,
+        msgtype: &MessageType,
         format: ContentFormat,
         sender: &Member,
         detect_at_room: bool,
-        transaction_id: Option<OwnedTransactionId>,
-        event_id: Option<OwnedEventId>,
+        cache_key: MessageCacheKey,
     ) {
         let room = sender.room();
 
-        if let Some(media_message) = MediaMessage::from_message(message.msgtype()) {
+        if let Some(media_message) = MediaMessage::from_message(msgtype) {
             self.build_media_message_content(
                 media_message,
                 format,
                 &room,
                 detect_at_room,
-                MessageCacheKey {
-                    transaction_id,
-                    event_id,
-                    is_edited: message.is_edited(),
-                },
+                cache_key,
             );
             return;
         }
 
-        match message.msgtype() {
+        match msgtype {
             MessageType::Emote(message) => {
-                let child = self.reuse_child_or_default::<MessageText>();
+                let child = self.child_or_default::<MessageText>();
                 child.with_emote(
                     message.formatted.clone(),
                     message.body.clone(),
@@ -325,11 +364,11 @@ trait MessageContentContainer: IsA<gtk::Widget> {
                 );
             }
             MessageType::Location(message) => {
-                let child = self.reuse_child_or_default::<MessageLocation>();
+                let child = self.child_or_default::<MessageLocation>();
                 child.set_geo_uri(&message.geo_uri, format);
             }
             MessageType::Notice(message) => {
-                let child = self.reuse_child_or_default::<MessageText>();
+                let child = self.child_or_default::<MessageText>();
                 child.with_markup(
                     message.formatted.clone(),
                     message.body.clone(),
@@ -339,11 +378,11 @@ trait MessageContentContainer: IsA<gtk::Widget> {
                 );
             }
             MessageType::ServerNotice(message) => {
-                let child = self.reuse_child_or_default::<MessageText>();
+                let child = self.child_or_default::<MessageText>();
                 child.with_plain_text(message.body.clone(), format);
             }
             MessageType::Text(message) => {
-                let child = self.reuse_child_or_default::<MessageText>();
+                let child = self.child_or_default::<MessageText>();
                 child.with_markup(
                     message.formatted.clone(),
                     message.body.clone(),
@@ -354,7 +393,7 @@ trait MessageContentContainer: IsA<gtk::Widget> {
             }
             msgtype => {
                 warn!("Event not supported: {msgtype:?}");
-                let child = self.reuse_child_or_default::<MessageText>();
+                let child = self.child_or_default::<MessageText>();
                 child.with_plain_text(gettext("Unsupported event"), format);
             }
         }
@@ -370,12 +409,8 @@ trait MessageContentContainer: IsA<gtk::Widget> {
         detect_at_room: bool,
         cache_key: MessageCacheKey,
     ) {
-        let Some(session) = room.session() else {
-            return;
-        };
-
         if let Some((caption, formatted_caption)) = media_message.caption() {
-            let caption_widget = self.reuse_child_or_default::<MessageCaption>();
+            let caption_widget = self.child_or_default::<MessageCaption>();
 
             caption_widget.set_caption(
                 caption.to_owned(),
@@ -385,9 +420,9 @@ trait MessageContentContainer: IsA<gtk::Widget> {
                 detect_at_room,
             );
 
-            caption_widget.build_media_content(media_message, format, &session, cache_key);
+            caption_widget.build_media_content(media_message, format, room, cache_key);
         } else {
-            self.build_media_content(media_message, format, &session, cache_key);
+            self.build_media_content(media_message, format, room, cache_key);
         }
     }
 
@@ -399,59 +434,43 @@ trait MessageContentContainer: IsA<gtk::Widget> {
         &self,
         media_message: MediaMessage,
         format: ContentFormat,
-        session: &Session,
+        room: &Room,
         cache_key: MessageCacheKey,
     ) {
         match media_message {
             MediaMessage::Audio(audio) => {
-                let widget = self.reuse_child_or_default::<MessageAudio>();
-                widget.audio(audio.into(), session, format, cache_key);
+                let Some(session) = room.session() else {
+                    return;
+                };
+                let widget = self.child_or_default::<MessageAudio>();
+                widget.audio(audio.into(), &session, format, cache_key);
             }
             MediaMessage::File(file) => {
-                let widget = self.reuse_child_or_default::<MessageFile>();
+                let widget = self.child_or_default::<MessageFile>();
 
                 let media_message = MediaMessage::from(file);
                 widget.set_filename(Some(media_message.filename()));
                 widget.set_format(format);
             }
             MediaMessage::Image(image) => {
-                let widget = self.reuse_child_or_default::<MessageVisualMedia>();
-                widget.set_media_message(image.into(), session, format, cache_key);
+                let widget = self.child_or_default::<MessageVisualMedia>();
+                widget.set_media_message(image.into(), room, format, cache_key);
             }
             MediaMessage::Video(video) => {
-                let widget = self.reuse_child_or_default::<MessageVisualMedia>();
-                widget.set_media_message(video.into(), session, format, cache_key);
+                let widget = self.child_or_default::<MessageVisualMedia>();
+                widget.set_media_message(video.into(), room, format, cache_key);
             }
             MediaMessage::Sticker(sticker) => {
-                let widget = self.reuse_child_or_default::<MessageVisualMedia>();
-                widget.set_media_message(sticker.into(), session, format, cache_key);
+                let widget = self.child_or_default::<MessageVisualMedia>();
+                widget.set_media_message(sticker.into(), room, format, cache_key);
             }
         }
     }
 }
 
-impl<W> MessageContentContainer for W
-where
-    W: IsA<adw::Bin> + IsA<gtk::Widget>,
-{
-    fn child(&self) -> Option<gtk::Widget> {
-        BinExt::child(self)
-    }
+impl<W> MessageContentContainer for W where W: IsABin {}
 
-    fn set_child(&self, child: Option<gtk::Widget>) {
-        BinExt::set_child(self, child.as_ref());
-    }
-}
-
-impl MessageContentContainer for MessageCaption {
-    fn child(&self) -> Option<gtk::Widget> {
-        self.child()
-    }
-
-    fn set_child(&self, child: Option<gtk::Widget>) {
-        self.set_child(child);
-    }
-}
+impl MessageContentContainer for MessageCaption {}
 
 /// The data used as a cache key for messages.
 ///
@@ -480,7 +499,7 @@ pub(crate) struct MessageCacheKey {
 
 impl MessageCacheKey {
     /// Whether the given new `MessageCacheKey` should trigger a reload of the
-    /// mmessage compared to this one.
+    /// message compared to this one.
     pub(super) fn should_reload(&self, new: &MessageCacheKey) -> bool {
         if new.is_edited {
             return true;
