@@ -1,11 +1,6 @@
 use adw::{prelude::*, subclass::prelude::*};
 use gettextrs::gettext;
-use gtk::{
-    gdk,
-    glib::{self, clone},
-    CompositeTemplate,
-};
-use matrix_sdk::Client;
+use gtk::{gdk, glib, glib::clone, CompositeTemplate};
 use ruma::api::client::media::get_content_thumbnail::v3::Method;
 use tracing::{error, warn};
 
@@ -13,15 +8,16 @@ use super::{content::MessageCacheKey, ContentFormat};
 use crate::{
     components::{AnimatedImagePaintable, VideoPlayer},
     gettext_f,
-    session::model::Session,
+    session::model::Room,
     spawn,
     utils::{
-        matrix::VisualMediaMessage,
+        key_bindings,
+        matrix::{VisualMediaMessage, VisualMediaType},
         media::{
             image::{ImageRequestPriority, ThumbnailSettings, THUMBNAIL_MAX_DIMENSIONS},
             FrameDimensions,
         },
-        CountedRef, File, LoadingState,
+        CountedRef, File, LoadingState, TemplateCallbacks,
     },
 };
 
@@ -35,11 +31,18 @@ const MAX_COMPACT_DIMENSIONS: FrameDimensions = FrameDimensions {
     width: 75,
     height: 50,
 };
+/// The name of the empty stack page.
+const EMPTY_PAGE: &str = "empty";
+/// The name of the placeholder stack page.
+const PLACEHOLDER_PAGE: &str = "placeholder";
 /// The name of the media stack page.
 const MEDIA_PAGE: &str = "media";
 
 mod imp {
-    use std::cell::{Cell, RefCell};
+    use std::{
+        cell::{Cell, RefCell},
+        marker::PhantomData,
+    };
 
     use glib::subclass::InitializingObject;
 
@@ -56,11 +59,26 @@ mod imp {
         #[template_child]
         stack: TemplateChild<gtk::Stack>,
         #[template_child]
+        preview_instructions: TemplateChild<gtk::Box>,
+        #[template_child]
+        preview_instructions_icon: TemplateChild<gtk::Image>,
+        #[template_child]
         spinner: TemplateChild<adw::Spinner>,
         #[template_child]
+        hide_preview_button: TemplateChild<gtk::Button>,
+        #[template_child]
         error: TemplateChild<gtk::Image>,
-        /// The supposed dimensions of the media.
-        dimensions: Cell<Option<FrameDimensions>>,
+        /// The room where the message was sent.
+        room: glib::WeakRef<Room>,
+        join_rule_handler: RefCell<Option<glib::SignalHandlerId>>,
+        session_settings_handler: RefCell<Option<glib::SignalHandlerId>>,
+        /// The visual media message to display.
+        media_message: RefCell<Option<VisualMediaMessage>>,
+        /// The cache key for the current media message.
+        ///
+        /// We only try to reload the media if the key changes. This is to avoid
+        /// reloading the media when a local echo changes to a remote echo.
+        cache_key: RefCell<MessageCacheKey>,
         /// The loading state of the media.
         #[property(get, builder(LoadingState::default()))]
         state: Cell<LoadingState>,
@@ -74,14 +92,17 @@ mod imp {
         #[property(get)]
         activatable: Cell<bool>,
         gesture_click: glib::WeakRef<gtk::GestureClick>,
-        /// The cache key for the current media message.
+        /// The current placeholder, if any.
         ///
-        /// We only try to reload the media if the key changes. This is to avoid
-        /// reloading the media when a local echo changes to a remote echo.
-        cache_key: RefCell<MessageCacheKey>,
+        /// This is the low-quality image shown while the content is loading or
+        /// when the preview is hidden.
+        placeholder: RefCell<Option<gtk::Picture>>,
         /// The current video file, if any.
         file: RefCell<Option<File>>,
         paintable_animation_ref: RefCell<Option<CountedRef>>,
+        /// The texture of the current image preview, if any.
+        #[property(get = Self::texture)]
+        texture: PhantomData<Option<gdk::Texture>>,
     }
 
     #[glib::object_subclass]
@@ -92,9 +113,16 @@ mod imp {
 
         fn class_init(klass: &mut Self::Class) {
             Self::bind_template(klass);
+            Self::bind_template_callbacks(klass);
+            TemplateCallbacks::bind_template_callbacks(klass);
 
             klass.set_css_name("message-visual-media");
             klass.set_accessible_role(gtk::AccessibleRole::Group);
+
+            klass.install_action("message-visual-media.activate", None, |obj, _, _| {
+                obj.imp().activate();
+            });
+            key_bindings::add_activate_bindings(klass, "message-visual-media.activate");
         }
 
         fn instance_init(obj: &InitializingObject<Self>) {
@@ -105,12 +133,16 @@ mod imp {
     #[glib::derived_properties]
     impl ObjectImpl for MessageVisualMedia {
         fn dispose(&self) {
+            self.clear();
             self.overlay.unparent();
         }
     }
 
     impl WidgetImpl for MessageVisualMedia {
         fn measure(&self, orientation: gtk::Orientation, for_size: i32) -> (i32, i32, i32, i32) {
+            // Get the minimum size wanted by the overlay.
+            let (overlay_min, ..) = self.overlay.measure(orientation, for_size);
+
             // Get the maximum size for the current state.
             let max_size = if self.compact.get() {
                 MAX_COMPACT_DIMENSIONS
@@ -144,14 +176,17 @@ mod imp {
                     };
                     let (_, intrinsic_for_size, ..) = child.measure(other_orientation, -1);
 
-                    let (min, nat, ..) =
+                    let (child_min, child_nat, ..) =
                         child.measure(orientation, for_size.min(intrinsic_for_size));
 
-                    if nat != 0 {
+                    if child_nat != 0 {
                         // Limit the returned size to the max.
                         let max = max.try_into().unwrap_or(i32::MAX);
 
-                        return (min.min(max), nat.min(max), -1, -1);
+                        let min = child_min.max(overlay_min).min(max);
+                        let nat = child_nat.max(overlay_min).min(max);
+
+                        return (min, nat, -1, -1);
                     }
                 }
             }
@@ -171,14 +206,23 @@ mod imp {
             };
 
             // Use the size from the info or the fallback size.
-            let media_size = self.dimensions.get().unwrap_or(FALLBACK_DIMENSIONS);
-            let nat = media_size
+            let media_size = self
+                .media_message
+                .borrow()
+                .as_ref()
+                .and_then(VisualMediaMessage::dimensions)
+                .unwrap_or(FALLBACK_DIMENSIONS);
+            let child_nat = media_size
                 .scale_to_fit(wanted_size, gtk::ContentFit::ScaleDown)
                 .dimension_for_orientation(orientation)
                 .try_into()
                 .unwrap_or(i32::MAX);
 
-            (0, nat, -1, -1)
+            let max = max.try_into().unwrap_or(i32::MAX);
+            let min = overlay_min.min(max);
+            let nat = child_nat.max(overlay_min).min(max);
+
+            (min, nat, -1, -1)
         }
 
         fn request_mode(&self) -> gtk::SizeRequestMode {
@@ -200,6 +244,7 @@ mod imp {
         }
     }
 
+    #[gtk::template_callbacks]
     impl MessageVisualMedia {
         /// The media child of the given type, if any.
         pub(super) fn media_child<T: IsA<gtk::Widget>>(&self) -> Option<T> {
@@ -209,12 +254,20 @@ mod imp {
         /// Set the media child.
         ///
         /// Removes the previous media child if one was set.
-        fn set_media_child(&self, child: &impl IsA<gtk::Widget>) {
+        fn set_media_child(&self, child: Option<&impl IsA<gtk::Widget>>) {
+            let prev_texture = self.texture();
+
             if let Some(prev_child) = self.stack.child_by_name(MEDIA_PAGE) {
                 self.stack.remove(&prev_child);
             }
 
-            self.stack.add_named(child, Some(MEDIA_PAGE));
+            if let Some(child) = child {
+                self.stack.add_named(child, Some(MEDIA_PAGE));
+            }
+
+            if self.texture() != prev_texture {
+                self.obj().notify_texture();
+            }
         }
 
         /// Set the state of the media.
@@ -223,27 +276,48 @@ mod imp {
                 return;
             }
 
-            match state {
-                LoadingState::Loading | LoadingState::Initial => {
-                    self.stack.set_visible_child_name("placeholder");
-                    self.spinner.set_visible(true);
-                    self.error.set_visible(false);
-                }
-                LoadingState::Ready => {
-                    self.stack.set_visible_child_name(MEDIA_PAGE);
-                    self.spinner.set_visible(false);
-                    self.error.set_visible(false);
-                }
-                LoadingState::Error => {
-                    self.spinner.set_visible(false);
-                    self.error.set_visible(true);
-                }
-            }
-
             self.state.set(state);
+
+            self.update_visible_page();
             self.obj().notify_state();
         }
 
+        /// Update the visible page for the current state.
+        fn update_visible_page(&self) {
+            let Some(room) = self.room.upgrade() else {
+                return;
+            };
+            let Some(session) = room.session() else {
+                return;
+            };
+
+            let state = self.state.get();
+
+            self.preview_instructions
+                .set_visible(state == LoadingState::Initial);
+            self.spinner.set_visible(state == LoadingState::Loading);
+            self.hide_preview_button.set_visible(
+                state == LoadingState::Ready
+                    && !session.settings().should_room_show_media_previews(&room),
+            );
+            self.error.set_visible(state == LoadingState::Error);
+
+            let visible_page = match state {
+                LoadingState::Initial | LoadingState::Loading => {
+                    if self.placeholder.borrow().is_some() {
+                        Some(PLACEHOLDER_PAGE)
+                    } else {
+                        Some(EMPTY_PAGE)
+                    }
+                }
+                LoadingState::Ready => Some(MEDIA_PAGE),
+                LoadingState::Error => None,
+            };
+            if let Some(visible_page) = visible_page {
+                self.stack.set_visible_child_name(visible_page);
+            }
+        }
+
         /// Update the state of the animated paintable, if any.
         fn update_animated_paintable_state(&self) {
             self.paintable_animation_ref.take();
@@ -276,7 +350,14 @@ mod imp {
                 self.overlay.remove_css_class("compact");
             }
 
-            self.update_gesture_click();
+            let icon_size = if compact {
+                gtk::IconSize::Normal
+            } else {
+                gtk::IconSize::Large
+            };
+            self.preview_instructions_icon.set_icon_size(icon_size);
+
+            self.update_activation();
             self.obj().notify_compact();
         }
 
@@ -288,29 +369,23 @@ mod imp {
 
             self.activatable.set(activatable);
 
-            self.update_gesture_click();
+            self.update_activation();
             self.obj().notify_activatable();
         }
 
-        /// Add or remove the click controller given the current state.
-        fn update_gesture_click(&self) {
-            let needs_controller = self.activatable.get() && !self.compact.get();
+        /// Enable or disable the activation of this row for the current state.
+        fn update_activation(&self) {
+            let is_activatable = self.activatable.get() && !self.compact.get();
             let gesture_click = self.gesture_click.upgrade();
 
-            if needs_controller && gesture_click.is_none() {
+            if is_activatable && gesture_click.is_none() {
                 let gesture_click = gtk::GestureClick::new();
 
                 gesture_click.connect_released(clone!(
                     #[weak(rename_to = imp)]
                     self,
                     move |_, _, _, _| {
-                        if imp
-                            .obj()
-                            .activate_action("message-row.show-media", None)
-                            .is_err()
-                        {
-                            error!("Could not activate `message-row.show-media` action");
-                        }
+                        imp.activate();
                     }
                 ));
 
@@ -320,6 +395,9 @@ mod imp {
                 self.gesture_click.set(None);
                 self.overlay.remove_controller(&gesture_click);
             }
+
+            self.obj()
+                .action_set_enabled("message-visual-media.activate", is_activatable);
         }
 
         /// Set the cache key with the given value.
@@ -333,47 +411,156 @@ mod imp {
             should_reload
         }
 
-        /// Build the content for the given media message.
-        pub(super) fn build(
+        /// Set the texture to use as a placeholder.
+        fn set_placeholder(&self, texture: Option<gdk::Texture>) {
+            if let Some(texture) = texture {
+                let placeholder = self.placeholder.borrow().clone();
+                let placeholder = if let Some(placeholder) = placeholder {
+                    placeholder
+                } else {
+                    let placeholder = gtk::Picture::new();
+                    self.placeholder.replace(Some(placeholder.clone()));
+                    self.stack.add_named(&placeholder, Some(PLACEHOLDER_PAGE));
+                    self.overlay.add_css_class("has-placeholder");
+                    placeholder
+                };
+
+                placeholder.set_paintable(Some(&texture));
+            } else if let Some(placeholder) = self.placeholder.take() {
+                self.stack.remove(&placeholder);
+                self.overlay.remove_css_class("has-placeholder");
+            }
+
+            self.update_visible_page();
+        }
+
+        /// The texture of the current image preview, if any.
+        fn texture(&self) -> Option<gdk::Texture> {
+            let paintable = self
+                .media_child::<gtk::Picture>()
+                .and_then(|p| p.paintable())?;
+
+            if let Some(paintable) = paintable.downcast_ref::<AnimatedImagePaintable>() {
+                paintable.current_texture()
+            } else {
+                paintable.downcast().ok()
+            }
+        }
+
+        /// Set the visual media message to display.
+        pub(super) fn set_media_message(
             &self,
             media_message: VisualMediaMessage,
-            session: &Session,
+            room: &Room,
             format: ContentFormat,
             cache_key: MessageCacheKey,
         ) {
+            self.media_message.replace(Some(media_message));
+
             if !self.set_cache_key(cache_key) {
                 // We do not need to reload the media.
                 return;
             }
 
-            self.file.take();
-            self.dimensions.set(media_message.dimensions());
+            // Reset the widget.
+            self.clear();
+            self.set_state(LoadingState::Initial);
 
             let compact = matches!(format, ContentFormat::Compact | ContentFormat::Ellipsized);
             self.set_compact(compact);
 
-            let activatable = matches!(
-                media_message,
-                VisualMediaMessage::Image(_) | VisualMediaMessage::Video(_)
-            );
-            self.set_activatable(activatable);
+            let Some(session) = room.session() else {
+                return;
+            };
+
+            let join_rule_handler = room.join_rule().connect_anyone_can_join_notify(clone!(
+                #[weak(rename_to = imp)]
+                self,
+                move |_| {
+                    imp.update_media();
+                }
+            ));
+            self.join_rule_handler.replace(Some(join_rule_handler));
+
+            let session_settings_handler = session
+                .settings()
+                .connect_media_previews_enabled_changed(clone!(
+                    #[weak(rename_to = imp)]
+                    self,
+                    move |_| {
+                        imp.update_media();
+                    }
+                ));
+            self.session_settings_handler
+                .replace(Some(session_settings_handler));
+
+            self.room.set(Some(room));
+
+            self.load_placeholder();
+            self.update_accessible_label();
+            self.update_preview_instructions_icon();
+            self.update_media();
+        }
+
+        /// Load the placeholder.
+        fn load_placeholder(&self) {
+            let Some((original_dimensions, blurhash)) = self
+                .media_message
+                .borrow()
+                .as_ref()
+                .and_then(|media_message| media_message.dimensions().zip(media_message.blurhash()))
+            else {
+                // Nothing to load.
+                self.set_placeholder(None);
+                return;
+            };
+
+            let max_dimensions = FrameDimensions::thumbnail_max_dimensions(1);
+            let dimensions =
+                original_dimensions.scale_to_fit(max_dimensions, gtk::ContentFit::ScaleDown);
+
+            let cache_key = self.cache_key.borrow().clone();
+            spawn!(clone!(
+                #[weak(rename_to = imp)]
+                self,
+                async move {
+                    let placeholder_texture = blurhash.into_texture(dimensions).await;
+
+                    if imp.cache_key.borrow().should_reload(&cache_key) {
+                        // The media has changed while this was loading, drop the placeholder.
+                        return;
+                    }
+
+                    imp.set_placeholder(placeholder_texture);
+                }
+            ));
+        }
+
+        /// Update the accessible label for the current state.
+        fn update_accessible_label(&self) {
+            let Some((filename, visual_media_type)) =
+                self.media_message.borrow().as_ref().map(|media_message| {
+                    (media_message.filename(), media_message.visual_media_type())
+                })
+            else {
+                return;
+            };
 
-            let filename = media_message.filename();
             let accessible_label = if filename.is_empty() {
-                match &media_message {
-                    VisualMediaMessage::Image(_) => gettext("Image"),
-                    VisualMediaMessage::Sticker(_) => gettext("Sticker"),
-                    VisualMediaMessage::Video(_) => gettext("Video"),
+                match visual_media_type {
+                    VisualMediaType::Image => gettext("Image"),
+                    VisualMediaType::Sticker => gettext("Sticker"),
+                    VisualMediaType::Video => gettext("Video"),
                 }
             } else {
-                match &media_message {
-                    VisualMediaMessage::Image(_) => {
+                match visual_media_type {
+                    VisualMediaType::Image => {
                         gettext_f("Image: {filename}", &[("filename", &filename)])
                     }
-                    VisualMediaMessage::Sticker(_) => {
+                    VisualMediaType::Sticker => {
                         gettext_f("Sticker: {filename}", &[("filename", &filename)])
                     }
-                    VisualMediaMessage::Video(_) => {
+                    VisualMediaType::Video => {
                         gettext_f("Video: {filename}", &[("filename", &filename)])
                     }
                 }
@@ -381,9 +568,66 @@ mod imp {
             self.obj()
                 .update_property(&[gtk::accessible::Property::Label(&accessible_label)]);
 
+            self.overlay.set_tooltip_text(Some(&filename));
+        }
+
+        /// Update the preview instructions icon for the current state.
+        fn update_preview_instructions_icon(&self) {
+            let Some(content_type) = self
+                .media_message
+                .borrow()
+                .as_ref()
+                .map(VisualMediaMessage::content_type)
+            else {
+                return;
+            };
+
+            self.preview_instructions_icon
+                .set_icon_name(Some(content_type.icon_name()));
+        }
+
+        /// Update the media for the current state.
+        fn update_media(&self) {
+            let Some(room) = self.room.upgrade() else {
+                return;
+            };
+            let Some(session) = room.session() else {
+                return;
+            };
+
+            if session.settings().should_room_show_media_previews(&room) {
+                // Only load the media if it was not loaded before.
+                if self.state.get() == LoadingState::Initial {
+                    self.show_media();
+                }
+            } else {
+                self.hide_media();
+            }
+        }
+
+        /// Hide the media.
+        #[template_callback]
+        fn hide_media(&self) {
+            self.set_state(LoadingState::Initial);
+            self.set_media_child(None::<&gtk::Widget>);
+            self.file.take();
+            self.set_activatable(true);
+        }
+
+        /// Show the media.
+        fn show_media(&self) {
+            let Some(media_message) = self.media_message.borrow().clone() else {
+                return;
+            };
+
             self.set_state(LoadingState::Loading);
 
-            let client = session.client();
+            let activatable = matches!(
+                media_message,
+                VisualMediaMessage::Image(_) | VisualMediaMessage::Video(_)
+            );
+            self.set_activatable(activatable);
+
             spawn!(
                 glib::Priority::LOW,
                 clone!(
@@ -392,10 +636,10 @@ mod imp {
                     async move {
                         match &media_message {
                             VisualMediaMessage::Image(_) | VisualMediaMessage::Sticker(_) => {
-                                imp.build_image(&media_message, client).await;
+                                imp.build_image(&media_message).await;
                             }
                             VisualMediaMessage::Video(_) => {
-                                imp.build_video(media_message, &client).await;
+                                imp.build_video(media_message).await;
                             }
                         }
 
@@ -406,14 +650,22 @@ mod imp {
         }
 
         /// Build the content for the image in the given media message.
-        async fn build_image(&self, media_message: &VisualMediaMessage, client: Client) {
-            // Disable the copy-image action while the image is loading.
-            if matches!(media_message, VisualMediaMessage::Image(_)) {
-                self.enable_copy_image_action(false);
+        async fn build_image(&self, media_message: &VisualMediaMessage) {
+            let Some(client) = self
+                .room
+                .upgrade()
+                .and_then(|room| room.session())
+                .map(|session| session.client())
+            else {
+                return;
+            };
+
+            if self.state.get() != LoadingState::Loading {
+                // Something occurred after the task was spawned, cancel the task.
+                return;
             }
 
             let scale_factor = self.obj().scale_factor();
-
             let settings = ThumbnailSettings {
                 dimensions: FrameDimensions::thumbnail_max_dimensions(scale_factor),
                 method: Method::Scale,
@@ -433,18 +685,24 @@ mod imp {
                 }
             };
 
-            let child = if let Some(child) = self.media_child::<gtk::Picture>() {
-                child
+            if self.state.get() != LoadingState::Loading {
+                // Something occurred while the image was loading, cancel the task.
+                return;
+            }
+
+            let paintable = gdk::Paintable::from(image);
+
+            if let Some(child) = self.media_child::<gtk::Picture>() {
+                child.set_paintable(Some(&paintable));
+                self.obj().notify_texture();
             } else {
                 let child = gtk::Picture::builder()
                     .content_fit(gtk::ContentFit::ScaleDown)
                     .build();
-                self.set_media_child(&child);
-                child
-            };
-            child.set_paintable(Some(&gdk::Paintable::from(image)));
+                child.set_paintable(Some(&paintable));
+                self.set_media_child(Some(&child));
+            }
 
-            child.set_tooltip_text(Some(&media_message.filename()));
             if matches!(&media_message, VisualMediaMessage::Sticker(_)) {
                 self.overlay.remove_css_class("opaque-bg");
             } else {
@@ -452,35 +710,25 @@ mod imp {
             }
 
             self.set_state(LoadingState::Ready);
-
-            // Enable the copy-image action now that the image is loaded.
-            if matches!(media_message, VisualMediaMessage::Image(_)) {
-                self.enable_copy_image_action(true);
-            }
-        }
-
-        /// Enable or disable the context menu action to copy the image.
-        fn enable_copy_image_action(&self, enable: bool) {
-            if self.compact.get() {
-                // In its compact form the message does not have actions.
-                return;
-            }
-
-            if self
-                .obj()
-                .activate_action(
-                    "room-history-row.enable-copy-image",
-                    Some(&enable.to_variant()),
-                )
-                .is_err()
-            {
-                error!("Could not change state of copy-image action: `room-history-row.enable-copy-image` action not found");
-            }
         }
 
         /// Build the content for the video in the given media message.
-        async fn build_video(&self, media_message: VisualMediaMessage, client: &Client) {
-            let file = match media_message.into_tmp_file(client).await {
+        async fn build_video(&self, media_message: VisualMediaMessage) {
+            let Some(client) = self
+                .room
+                .upgrade()
+                .and_then(|room| room.session())
+                .map(|session| session.client())
+            else {
+                return;
+            };
+
+            if self.state.get() != LoadingState::Loading {
+                // Something occurred after the task was spawned, cancel the task.
+                return;
+            }
+
+            let file = match media_message.into_tmp_file(&client).await {
                 Ok(file) => file,
                 Err(error) => {
                     warn!("Could not retrieve video: {error}");
@@ -489,6 +737,11 @@ mod imp {
                 }
             };
 
+            if self.state.get() != LoadingState::Loading {
+                // Something occurred while the video was loading, cancel the task.
+                return;
+            }
+
             let child = if let Some(child) = self.media_child::<VideoPlayer>() {
                 child
             } else {
@@ -500,7 +753,7 @@ mod imp {
                         imp.video_state_changed(player);
                     }
                 ));
-                self.set_media_child(&child);
+                self.set_media_child(Some(&child));
                 child
             };
 
@@ -533,6 +786,36 @@ mod imp {
                 }
             }
         }
+
+        /// Reset the state of this widget.
+        fn clear(&self) {
+            self.file.take();
+
+            if let Some(room) = self.room.upgrade() {
+                if let Some(handler) = self.join_rule_handler.take() {
+                    room.join_rule().disconnect(handler);
+                }
+
+                if let Some(handler) = self.session_settings_handler.take() {
+                    if let Some(session) = room.session() {
+                        session.settings().disconnect(handler);
+                    }
+                }
+            }
+        }
+
+        /// Handle when the widget is activated.
+        fn activate(&self) {
+            if self.state.get() == LoadingState::Initial {
+                self.show_media();
+            } else if self
+                .obj()
+                .activate_action("message-row.show-media", None)
+                .is_err()
+            {
+                error!("Could not activate `message-row.show-media` action");
+            }
+        }
     }
 }
 
@@ -548,29 +831,16 @@ impl MessageVisualMedia {
         glib::Object::new()
     }
 
-    /// Display the given visual media message.
+    /// Set the visual media message to display.
     pub(crate) fn set_media_message(
         &self,
         media_message: VisualMediaMessage,
-        session: &Session,
+        room: &Room,
         format: ContentFormat,
         cache_key: MessageCacheKey,
     ) {
-        self.imp().build(media_message, session, format, cache_key);
-    }
-
-    /// Get the texture displayed by this widget, if any.
-    pub(crate) fn texture(&self) -> Option<gdk::Texture> {
-        let paintable = self
-            .imp()
-            .media_child::<gtk::Picture>()
-            .and_then(|p| p.paintable())?;
-
-        if let Some(paintable) = paintable.downcast_ref::<AnimatedImagePaintable>() {
-            paintable.current_texture()
-        } else {
-            paintable.downcast().ok()
-        }
+        self.imp()
+            .set_media_message(media_message, room, format, cache_key);
     }
 }
 
