@@ -4,11 +4,14 @@ use gtk::{
     prelude::*,
     subclass::prelude::*,
 };
-use matrix_sdk::{ComposerDraft, ComposerDraftType};
-use matrix_sdk_ui::timeline::{Message, RepliedToInfo, TimelineEventItemId};
+use matrix_sdk::{deserialized_responses::TimelineEvent, ComposerDraft, ComposerDraftType};
+use matrix_sdk_ui::timeline::Message;
 use ruma::{
-    events::room::message::{MessageFormat, MessageType},
-    OwnedEventId, RoomOrAliasId, UserId,
+    events::{
+        room::message::{MessageFormat, MessageType, OriginalSyncRoomMessageEvent},
+        AnySyncMessageLikeEvent, AnySyncTimelineEvent, SyncMessageLikeEvent,
+    },
+    OwnedEventId, OwnedUserId, RoomOrAliasId, UserId,
 };
 use sourceview::prelude::*;
 use tracing::{error, warn};
@@ -17,7 +20,7 @@ use super::ComposerParser;
 use crate::{
     components::{Pill, PillSource},
     prelude::*,
-    session::model::{Member, Room},
+    session::model::{Event, Member, Room, Timeline},
     spawn, spawn_tokio,
     utils::matrix::{find_at_room, find_html_mentions, AT_ROOM},
 };
@@ -233,18 +236,19 @@ mod imp {
 
         /// Add the given widget at the position of the given iter to this
         /// state.
-        pub(super) fn add_widget(&self, widget: impl IsA<gtk::Widget>, iter: &mut gtk::TextIter) {
-            let widget = widget.upcast();
-
-            let anchor = match iter.child_anchor() {
-                Some(anchor) => anchor,
-                None => self.buffer.create_child_anchor(iter),
+        pub(super) fn add_widget(&self, widget: gtk::Widget, iter: &mut gtk::TextIter) {
+            let Some(view) = self.view.upgrade() else {
+                return;
             };
 
-            if let Some(view) = self.view.upgrade() {
-                view.add_child_at_anchor(&widget, &anchor);
-            }
+            // Reuse the child anchor at the iter if it does not have a child widget,
+            // otherwise create a new one.
+            let anchor = iter
+                .child_anchor()
+                .filter(|anchor| self.widget_at_anchor(anchor).is_none())
+                .unwrap_or_else(|| self.buffer.create_child_anchor(iter));
 
+            view.add_child_at_anchor(&widget, &anchor);
             self.widgets.borrow_mut().push((widget, anchor));
         }
 
@@ -261,16 +265,12 @@ mod imp {
         }
 
         /// Restore the state from the persisted draft.
-        pub(super) async fn restore_draft(&self) {
-            let Some(room) = self.room.upgrade() else {
-                return;
-            };
-
-            let matrix_room = room.matrix_room().clone();
+        pub(super) async fn restore_draft(&self, timeline: &Timeline) {
+            let matrix_room = timeline.room().matrix_room().clone();
             let handle = spawn_tokio!(async move { matrix_room.load_composer_draft().await });
 
-            match handle.await.unwrap() {
-                Ok(Some(draft)) => self.restore_from_draft(draft).await,
+            match handle.await.expect("task was not aborted") {
+                Ok(Some(draft)) => self.restore_from_draft(timeline, draft).await,
                 Ok(None) => {}
                 Err(error) => {
                     error!("Could not restore draft: {error}");
@@ -279,13 +279,11 @@ mod imp {
         }
 
         /// Restore the state from the given draft.
-        async fn restore_from_draft(&self, draft: ComposerDraft) {
-            let Some(room) = self.room.upgrade() else {
-                return;
-            };
+        async fn restore_from_draft(&self, timeline: &Timeline, draft: ComposerDraft) {
+            let room = timeline.room();
 
             // Restore the relation.
-            self.restore_related_to_from_draft(draft.draft_type.clone())
+            self.restore_related_to_from_draft(&room, draft.draft_type.clone())
                 .await;
 
             // Make sure we start from an empty state.
@@ -314,7 +312,7 @@ mod imp {
 
                 match DraftMention::new(&room, &text[content_start..content_end]) {
                     DraftMention::Source(source) => {
-                        self.add_widget(source.to_pill(), &mut end_iter);
+                        self.add_widget(source.to_pill().upcast(), &mut end_iter);
                     }
                     DraftMention::Text(s) => {
                         self.buffer.insert(&mut end_iter, s);
@@ -332,33 +330,8 @@ mod imp {
         }
 
         /// Restore the relation from the given draft content.
-        async fn restore_related_to_from_draft(&self, draft_type: ComposerDraftType) {
-            let Some(room) = self.room.upgrade() else {
-                return;
-            };
-
-            let related_to = match draft_type {
-                ComposerDraftType::NewMessage => None,
-                ComposerDraftType::Reply { event_id } => {
-                    let matrix_timeline = room.timeline().matrix_timeline();
-
-                    let handle = spawn_tokio!(async move {
-                        matrix_timeline
-                            .replied_to_info_from_event_id(&event_id)
-                            .await
-                    });
-
-                    match handle.await.unwrap() {
-                        Ok(info) => Some(RelationInfo::Reply(info)),
-                        Err(error) => {
-                            warn!("Could not fetch replied-to event content of draft: {error}");
-                            None
-                        }
-                    }
-                }
-                ComposerDraftType::Edit { event_id } => Some(RelationInfo::Edit(event_id)),
-            };
-
+        async fn restore_related_to_from_draft(&self, room: &Room, draft_type: ComposerDraftType) {
+            let related_to = RelationInfo::from_draft(room, draft_type).await;
             self.related_to.replace(related_to);
 
             let obj = self.obj();
@@ -411,7 +384,7 @@ mod imp {
             };
 
             // Try to detect `@room` mentions.
-            let can_contain_at_room = message.mentions().map_or(true, |m| m.room);
+            let can_contain_at_room = message.mentions().is_none_or(|m| m.room);
             if room.permissions().can_notify_room() && can_contain_at_room {
                 if let Some(start) = find_at_room(&text) {
                     let pill = room.at_room().to_pill();
@@ -438,7 +411,7 @@ mod imp {
                         self.buffer.insert(&mut iter, &text[pos..start]);
                     }
 
-                    self.add_widget(pill, &mut iter);
+                    self.add_widget(pill.upcast(), &mut iter);
 
                     pos = end;
                 }
@@ -470,20 +443,22 @@ glib::wrapper! {
 }
 
 impl ComposerState {
-    /// Create a new empty `ComposerState` for the given room.
-    pub fn new(room: Option<&Room>) -> Self {
+    /// Create a new empty `ComposerState` for the room of the given timeline.
+    pub fn new(timeline: Option<Timeline>) -> Self {
         let obj = glib::Object::builder::<Self>()
-            .property("room", room)
+            .property("room", timeline.as_ref().map(Timeline::room))
             .build();
 
-        let imp = obj.imp();
-        spawn!(clone!(
-            #[weak]
-            imp,
-            async move {
-                imp.restore_draft().await;
-            }
-        ));
+        if let Some(timeline) = timeline {
+            let imp = obj.imp();
+            spawn!(clone!(
+                #[weak]
+                imp,
+                async move {
+                    imp.restore_draft(&timeline).await;
+                }
+            ));
+        }
 
         obj
     }
@@ -515,7 +490,7 @@ impl ComposerState {
 
     /// Add the given widget at the position of the given iter to this state.
     pub(crate) fn add_widget(&self, widget: impl IsA<gtk::Widget>, iter: &mut gtk::TextIter) {
-        self.imp().add_widget(widget, iter);
+        self.imp().add_widget(widget.upcast(), iter);
     }
 
     /// Get the widget at the given anchor, if any.
@@ -541,27 +516,60 @@ impl ComposerState {
 /// The possible relations to send with a message.
 #[derive(Debug, Clone)]
 pub(crate) enum RelationInfo {
-    /// Send a reply with the given replied to info.
-    Reply(RepliedToInfo),
+    /// Send a reply to the given event.
+    Reply(MessageEventSource),
 
     /// Send an edit to the event with the given ID.
     Edit(OwnedEventId),
 }
 
 impl RelationInfo {
+    /// Construct a relation info from a composer draft, if possible.
+    pub(crate) async fn from_draft(room: &Room, draft_type: ComposerDraftType) -> Option<Self> {
+        match draft_type {
+            ComposerDraftType::NewMessage => None,
+            ComposerDraftType::Reply { event_id } => {
+                // We need to fetch the event and extract its content, so we can display it.
+                let matrix_room = room.matrix_room().clone();
+                let event_id_clone = event_id.clone();
+
+                let handle = spawn_tokio!(async move {
+                    matrix_room.load_or_fetch_event(&event_id_clone, None).await
+                });
+
+                let event = match handle.await.expect("task was not aborted") {
+                    Ok(event) => event,
+                    Err(error) => {
+                        warn!("Could not fetch replied-to event content of draft: {error}");
+                        return None;
+                    }
+                };
+
+                // We only reply to messages.
+                let Some(message_event) = MessageEventSource::from_original_event(&event) else {
+                    warn!("Could not fetch replied-to event content of draft: unsupported event");
+                    return None;
+                };
+
+                Some(RelationInfo::Reply(message_event))
+            }
+            ComposerDraftType::Edit { event_id } => Some(RelationInfo::Edit(event_id)),
+        }
+    }
+
     /// The unique global identifier of the related event.
-    pub(crate) fn identifier(&self) -> TimelineEventItemId {
+    pub(crate) fn event_id(&self) -> OwnedEventId {
         match self {
-            RelationInfo::Reply(info) => TimelineEventItemId::EventId(info.event_id().to_owned()),
-            RelationInfo::Edit(event_id) => TimelineEventItemId::EventId(event_id.clone()),
+            RelationInfo::Reply(message_event) => message_event.event_id(),
+            RelationInfo::Edit(event_id) => event_id.clone(),
         }
     }
 
     /// Get this `RelationInfo` as a draft type.
     pub(crate) fn as_draft_type(&self) -> ComposerDraftType {
         match self {
-            Self::Reply(info) => ComposerDraftType::Reply {
-                event_id: info.event_id().to_owned(),
+            Self::Reply(message_event) => ComposerDraftType::Reply {
+                event_id: message_event.event_id(),
             },
             Self::Edit(event_id) => ComposerDraftType::Edit {
                 event_id: event_id.clone(),
@@ -635,3 +643,84 @@ struct DetectedMention {
     /// The end of the mention in the text.
     end: usize,
 }
+
+/// The possible sources of a message event.
+#[derive(Debug, Clone)]
+pub(crate) enum MessageEventSource {
+    /// An original event.
+    OriginalEvent(OriginalSyncRoomMessageEvent),
+    /// An [`Event`].
+    Event(Event),
+}
+
+impl MessageEventSource {
+    /// Try to construct a `MessageEventSource` from the given
+    /// [`TimelineEvent`].
+    ///
+    /// Returns `None` if the event is not a message.
+    fn from_original_event(event: &TimelineEvent) -> Option<Self> {
+        let event = event.raw().deserialize().ok()?;
+        match event {
+            AnySyncTimelineEvent::MessageLike(AnySyncMessageLikeEvent::RoomMessage(
+                SyncMessageLikeEvent::Original(message_event),
+            )) => Some(Self::OriginalEvent(message_event)),
+            _ => None,
+        }
+    }
+
+    /// Try to construct a `MessageEventSource` from the given [`Event`].
+    ///
+    /// Returns `None` if the event is not a message.
+    pub(crate) fn from_event(event: Event) -> Option<Self> {
+        (event.can_be_replied_to()).then_some(Self::Event(event))
+    }
+
+    /// The ID of the underlying event.
+    pub(crate) fn event_id(&self) -> OwnedEventId {
+        match self {
+            Self::OriginalEvent(event) => event.event_id.clone(),
+            Self::Event(event) => event
+                .event_id()
+                .expect("replied-to event should always have an event ID"),
+        }
+    }
+
+    /// The ID of the sender of the event.
+    pub(crate) fn sender(&self) -> OwnedUserId {
+        match self {
+            Self::OriginalEvent(event) => event.sender.clone(),
+            Self::Event(event) => event.sender_id(),
+        }
+    }
+
+    /// The message content of the event.
+    ///
+    /// Returns `None` if the event was redacted after being selected to be
+    /// replied to.
+    pub(crate) fn msgtype(&self) -> Option<MessageType> {
+        match self {
+            Self::OriginalEvent(event) => Some(event.content.msgtype.clone()),
+            Self::Event(event) => Some(event.message()?.msgtype().clone()),
+        }
+    }
+
+    /// Whether the message was edited.
+    pub(crate) fn is_edited(&self) -> bool {
+        match self {
+            Self::OriginalEvent(event) => event.unsigned.relations.has_replacement(),
+            Self::Event(event) => event.is_edited(),
+        }
+    }
+
+    /// Whether this message might contain an `@room` mention.
+    pub(crate) fn can_contain_at_room(&self) -> bool {
+        match self {
+            Self::OriginalEvent(event) => event
+                .content
+                .mentions
+                .as_ref()
+                .is_none_or(|mentions| mentions.room),
+            Self::Event(event) => event.can_contain_at_room(),
+        }
+    }
+}
