@@ -10,14 +10,15 @@ use gtk::{
 };
 use matrix_sdk::{
     attachment::{AttachmentConfig, AttachmentInfo, BaseFileInfo, Thumbnail},
-    room::edit::EditedContent,
+    room::{
+        edit::EditedContent,
+        reply::{EnforceThread, Reply},
+    },
 };
-use matrix_sdk_ui::timeline::{AttachmentSource, RepliedToInfo, TimelineItemContent};
+use matrix_sdk_ui::timeline::{AttachmentSource, TimelineEventItemId, TimelineItemContent};
 use ruma::{
     events::{
-        room::message::{
-            ForwardThread, LocationMessageEventContent, MessageType, RoomMessageEventContent,
-        },
+        room::message::{LocationMessageEventContent, MessageType, RoomMessageEventContent},
         Mentions,
     },
     OwnedRoomId,
@@ -29,7 +30,7 @@ mod completion;
 mod composer_parser;
 mod composer_state;
 
-pub(crate) use self::composer_state::{ComposerState, RelationInfo};
+pub(crate) use self::composer_state::{ComposerState, MessageEventSource, RelationInfo};
 use self::{
     attachment_dialog::AttachmentDialog, completion::CompletionPopover,
     composer_parser::ComposerParser,
@@ -39,15 +40,14 @@ use crate::{
     components::{CustomEntry, LabelWithWidgets},
     gettext_f,
     prelude::*,
-    session::model::{Event, Member, Room},
+    session::model::{Event, Member, Room, Timeline},
     spawn, spawn_tokio, toast,
     utils::{
         media::{
             filename_for_mime, image::ImageInfoLoader, load_audio_info, video::load_video_info,
             FileInfo,
         },
-        template_callbacks::TemplateCallbacks,
-        Location, LocationError, TokioDrop,
+        Location, LocationError, TemplateCallbacks, TokioDrop,
     },
 };
 
@@ -74,16 +74,16 @@ mod imp {
         #[template_child]
         main_stack: TemplateChild<gtk::Stack>,
         #[template_child]
-        message_entry: TemplateChild<sourceview::View>,
+        pub(super) message_entry: TemplateChild<sourceview::View>,
         #[template_child]
         send_button: TemplateChild<gtk::Button>,
         #[template_child]
         related_event_header: TemplateChild<LabelWithWidgets>,
         #[template_child]
         related_event_content: TemplateChild<MessageContent>,
-        /// The room to send messages in.
-        #[property(get, set = Self::set_room, explicit_notify, nullable)]
-        room: glib::WeakRef<Room>,
+        /// The timeline used to send messages.
+        #[property(get, set = Self::set_timeline, explicit_notify, nullable)]
+        timeline: glib::WeakRef<Timeline>,
         send_message_permission_handler: RefCell<Option<glib::SignalHandlerId>>,
         /// Whether outgoing messages should be interpreted as markdown.
         #[property(get, set)]
@@ -146,9 +146,6 @@ mod imp {
 
             // Tab auto-completion.
             self.completion.set_parent(&*self.message_entry);
-            obj.bind_property("room", &self.completion, "room")
-                .sync_create()
-                .build();
 
             // Location.
             let location = Location::new();
@@ -158,36 +155,51 @@ mod imp {
         fn dispose(&self) {
             self.completion.unparent();
 
-            if let Some(room) = self.room.upgrade() {
+            if let Some(timeline) = self.timeline.upgrade() {
                 if let Some(handler) = self.send_message_permission_handler.take() {
-                    room.permissions().disconnect(handler);
+                    timeline.room().permissions().disconnect(handler);
                 }
             }
         }
     }
 
-    impl WidgetImpl for MessageToolbar {}
+    impl WidgetImpl for MessageToolbar {
+        fn grab_focus(&self) -> bool {
+            if self
+                .main_stack
+                .visible_child_name()
+                .is_none_or(|name| name == "disabled")
+            {
+                return false;
+            }
+
+            self.message_entry.grab_focus()
+        }
+    }
+
     impl BinImpl for MessageToolbar {}
 
     #[gtk::template_callbacks]
     impl MessageToolbar {
-        /// Set the room currently displayed.
-        fn set_room(&self, room: Option<&Room>) {
-            let old_room = self.room.upgrade();
-            if old_room.as_ref() == room {
+        /// Set the timeline used to send messages.
+        fn set_timeline(&self, timeline: Option<&Timeline>) {
+            let old_timeline = self.timeline.upgrade();
+            if old_timeline.as_ref() == timeline {
                 return;
             }
             let obj = self.obj();
 
-            if let Some(room) = &old_room {
+            if let Some(timeline) = &old_timeline {
                 if let Some(handler) = self.send_message_permission_handler.take() {
-                    room.permissions().disconnect(handler);
+                    timeline.room().permissions().disconnect(handler);
                 }
             }
 
-            if let Some(room) = room {
-                let send_message_permission_handler =
-                    room.permissions().connect_can_send_message_notify(clone!(
+            if let Some(timeline) = timeline {
+                let send_message_permission_handler = timeline
+                    .room()
+                    .permissions()
+                    .connect_can_send_message_notify(clone!(
                         #[weak(rename_to = imp)]
                         self,
                         move |_| {
@@ -198,13 +210,13 @@ mod imp {
                     .replace(Some(send_message_permission_handler));
             }
 
-            self.room.set(room);
+            self.completion.set_room(timeline.map(Timeline::room));
+            self.timeline.set(timeline);
 
             self.send_message_permission_updated();
-            self.message_entry.grab_focus();
 
-            obj.notify_room();
-            self.update_current_composer_state(old_room.as_ref());
+            obj.notify_timeline();
+            self.update_current_composer_state(old_timeline);
         }
 
         /// Whether the user can compose a message.
@@ -212,9 +224,9 @@ mod imp {
         /// It depends on whether our own user has the permission to send a
         /// message in the current room.
         pub(super) fn can_compose_message(&self) -> bool {
-            self.room
+            self.timeline
                 .upgrade()
-                .is_some_and(|r| r.permissions().can_send_message())
+                .is_some_and(|timeline| timeline.room().permissions().can_send_message())
         }
 
         /// Handle an update of the permission to send a message in the current
@@ -228,31 +240,58 @@ mod imp {
             self.main_stack.set_visible_child_name(page);
         }
 
-        /// Get the current composer state.
-        fn current_composer_state(&self) -> ComposerState {
-            let room = self.room.upgrade();
-            self.composer_state(room.as_ref())
+        /// Whether the buffer of the composer is empty.
+        ///
+        /// Returns `true` if the buffer is empty or contains only whitespace.
+        fn is_buffer_empty(&self) -> bool {
+            let mut iter = self.message_entry.buffer().start_iter();
+
+            if iter.is_end() {
+                return true;
+            }
+
+            loop {
+                if !iter.char().is_whitespace() {
+                    // The buffer is not empty.
+                    return false;
+                }
+
+                if !iter.forward_cursor_position() {
+                    // We are at the end and we did not encounter a non-whitespace character, the
+                    // buffer is empty.
+                    return true;
+                }
+            }
         }
 
-        /// Get the composer state for the given room.
+        /// The current composer state.
+        fn current_composer_state(&self) -> ComposerState {
+            let timeline = self.timeline.upgrade();
+            self.composer_state(timeline)
+        }
+
+        /// The composer state for the given room.
         ///
-        /// If the composer state doesn't exist, it is created.
-        fn composer_state(&self, room: Option<&Room>) -> ComposerState {
+        /// If the composer state does not exist, it is created.
+        fn composer_state(&self, timeline: Option<Timeline>) -> ComposerState {
+            let room = timeline.as_ref().map(Timeline::room);
+
             self.composer_states
                 .borrow_mut()
                 .entry(
-                    room.and_then(Room::session)
+                    room.as_ref()
+                        .and_then(Room::session)
                         .map(|s| s.session_id().to_owned()),
                 )
                 .or_default()
-                .entry(room.map(|r| r.room_id().to_owned()))
-                .or_insert_with(|| ComposerState::new(room))
+                .entry(room.map(|room| room.room_id().to_owned()))
+                .or_insert_with(|| ComposerState::new(timeline))
                 .clone()
         }
 
         /// Update the current composer state.
-        fn update_current_composer_state(&self, old_room: Option<&Room>) {
-            let old_composer_state = self.composer_state(old_room);
+        fn update_current_composer_state(&self, old_timeline: Option<Timeline>) {
+            let old_composer_state = self.composer_state(old_timeline);
             old_composer_state.attach_to_view(None);
 
             if let Some(handler) = self.composer_state_handler.take() {
@@ -275,16 +314,14 @@ mod imp {
             let text_notify_handler = buffer.connect_text_notify(clone!(
                 #[weak(rename_to = imp)]
                 self,
-                move |buffer| {
-                    let (start_iter, end_iter) = buffer.bounds();
-                    let is_empty = start_iter == end_iter;
+                move |_| {
+                    let is_empty = imp.is_buffer_empty();
                     imp.send_button.set_sensitive(!is_empty);
                     imp.send_typing_notification(!is_empty);
                 }
             ));
 
-            let (start_iter, end_iter) = buffer.bounds();
-            let is_empty = start_iter == end_iter;
+            let is_empty = self.is_buffer_empty();
             self.send_button.set_sensitive(!is_empty);
 
             // Markdown highlighting.
@@ -316,8 +353,8 @@ mod imp {
             let composer_state = self.current_composer_state();
 
             match composer_state.related_to() {
-                Some(RelationInfo::Reply(info)) => {
-                    self.update_for_reply(&info);
+                Some(RelationInfo::Reply(event)) => {
+                    self.update_for_reply(&event);
                 }
                 Some(RelationInfo::Edit(_)) => {
                     self.update_for_edit();
@@ -326,15 +363,21 @@ mod imp {
             }
         }
 
-        /// Update the displayed related event for the given reply.
-        fn update_for_reply(&self, info: &RepliedToInfo) {
-            let Some(room) = self.room.upgrade() else {
+        /// Update the displayed related event for the given replied-to event.
+        fn update_for_reply(&self, message_event: &MessageEventSource) {
+            let Some(msgtype) = message_event.msgtype() else {
+                // The event was probably redacted, we cannot reply to it anymore.
+                self.clear_related_event();
+                return;
+            };
+            let Some(timeline) = self.timeline.upgrade() else {
                 return;
             };
 
+            let room = timeline.room();
             let sender = room
                 .get_or_create_members()
-                .get_or_create(info.sender().to_owned());
+                .get_or_create(message_event.sender());
 
             let label = gettext_f(
                 // Translators: Do NOT translate the content between '{' and '}',
@@ -348,7 +391,7 @@ mod imp {
                 .set_label_and_widgets(label, vec![pill]);
 
             self.related_event_content
-                .update_for_related_event(info, &sender);
+                .update_for_related_event(&msgtype, message_event, &sender);
             self.related_event_content.set_visible(true);
         }
 
@@ -384,18 +427,22 @@ mod imp {
         }
 
         /// Set the event to reply to.
-        pub(super) fn set_reply_to(&self, event: &Event) {
+        pub(super) fn set_reply_to(&self, event: Event) {
             if !self.can_compose_message() {
                 return;
             }
 
-            let Ok(info) = event.item().replied_to_info() else {
+            if event.event_id().is_none() {
+                warn!("Cannot send reply for event that is not sent yet");
+                return;
+            }
+            let Some(message_event) = MessageEventSource::from_event(event) else {
                 warn!("Unsupported event type for reply");
                 return;
             };
 
             self.current_composer_state()
-                .set_related_to(Some(RelationInfo::Reply(info)));
+                .set_related_to(Some(RelationInfo::Reply(message_event)));
 
             self.message_entry.grab_focus();
         }
@@ -412,13 +459,17 @@ mod imp {
                 warn!("Cannot send edit for event that is not sent yet");
                 return;
             };
-            let TimelineItemContent::Message(message) = item.content() else {
+            let TimelineItemContent::MsgLike(msg_like) = item.content() else {
+                warn!("Unsupported event type for edit");
+                return;
+            };
+            let Some(message) = msg_like.as_message() else {
                 warn!("Unsupported event type for edit");
                 return;
             };
 
             self.current_composer_state()
-                .set_edit_source(event_id.to_owned(), message);
+                .set_edit_source(event_id.to_owned(), &message);
 
             self.message_entry.grab_focus();
         }
@@ -431,7 +482,18 @@ mod imp {
             _keycode: u32,
             modifier: gdk::ModifierType,
         ) -> glib::Propagation {
-            if modifier.is_empty() && (key == gdk::Key::Return || key == gdk::Key::KP_Enter) {
+            // Do not capture key press if there is a mask other than CapsLock.
+            if modifier != gdk::ModifierType::NO_MODIFIER_MASK
+                && modifier != gdk::ModifierType::LOCK_MASK
+            {
+                return glib::Propagation::Proceed;
+            }
+
+            // Send message on enter.
+            if matches!(
+                key,
+                gdk::Key::Return | gdk::Key::KP_Enter | gdk::Key::ISO_Enter,
+            ) {
                 spawn!(clone!(
                     #[weak(rename_to = imp)]
                     self,
@@ -439,16 +501,32 @@ mod imp {
                         imp.send_text_message().await;
                     }
                 ));
-                glib::Propagation::Stop
-            } else if modifier.is_empty()
-                && key == gdk::Key::Escape
-                && self.current_composer_state().has_relation()
-            {
-                self.clear_related_event();
-                glib::Propagation::Stop
-            } else {
-                glib::Propagation::Proceed
+                return glib::Propagation::Stop;
             }
+
+            // Clear related event on escape.
+            if key == gdk::Key::Escape && self.current_composer_state().has_relation() {
+                self.clear_related_event();
+                return glib::Propagation::Stop;
+            }
+
+            // Edit the last message on key up, if the composer is empty and the completion
+            // popover is not open.
+            if matches!(key, gdk::Key::Up | gdk::Key::KP_Up)
+                && !self.completion.is_visible()
+                && self.is_buffer_empty()
+            {
+                if self
+                    .obj()
+                    .activate_action("room-history.edit-latest-message", None)
+                    .is_err()
+                {
+                    error!("Could not activate `room-history.edit-latest-message` action");
+                }
+                return glib::Propagation::Stop;
+            }
+
+            glib::Propagation::Proceed
         }
 
         /// Send the text message that is currently in the message entry.
@@ -460,7 +538,7 @@ mod imp {
             if !self.can_compose_message() {
                 return;
             }
-            let Some(room) = self.room.upgrade() else {
+            let Some(timeline) = self.timeline.upgrade() else {
                 return;
             };
 
@@ -474,24 +552,28 @@ mod imp {
                 return;
             };
 
-            let matrix_timeline = room.timeline().matrix_timeline();
+            let matrix_timeline = timeline.matrix_timeline();
 
             // Send event depending on relation.
             match composer_state.related_to() {
-                Some(RelationInfo::Reply(replied_to_info)) => {
-                    let handle = spawn_tokio!(async move {
-                        matrix_timeline
-                            .send_reply(content, replied_to_info, ForwardThread::Yes)
-                            .await
-                    });
-                    if let Err(error) = handle.await.unwrap() {
+                Some(RelationInfo::Reply(message_event)) => {
+                    let reply = Reply {
+                        event_id: message_event.event_id(),
+                        enforce_thread: EnforceThread::MaybeThreaded,
+                    };
+
+                    let handle =
+                        spawn_tokio!(
+                            async move { matrix_timeline.send_reply(content, reply).await }
+                        );
+
+                    if let Err(error) = handle.await.expect("task was not aborted") {
                         error!("Could not send reply: {error}");
-                        let obj = self.obj();
-                        toast!(obj, gettext("Could not send reply"));
+                        toast!(self.obj(), gettext("Could not send reply"));
                     }
                 }
                 Some(RelationInfo::Edit(event_id)) => {
-                    let matrix_room = room.matrix_room().clone();
+                    let matrix_room = timeline.room().matrix_room().clone();
                     let handle = spawn_tokio!(async move {
                         let full_content = matrix_room
                             .make_edit_event(&event_id, EditedContent::RoomMessage(content))
@@ -503,8 +585,7 @@ mod imp {
                     });
                     if let Err(error) = handle.await.unwrap() {
                         error!("Could not send edit: {error}");
-                        let obj = self.obj();
-                        toast!(obj, gettext("Could not send edit"));
+                        toast!(self.obj(), gettext("Could not send edit"));
                     }
                 }
                 _ => {
@@ -515,8 +596,7 @@ mod imp {
                     });
                     if let Err(error) = handle.await.unwrap() {
                         error!("Could not send message: {error}");
-                        let obj = self.obj();
-                        toast!(obj, gettext("Could not send message"));
+                        toast!(self.obj(), gettext("Could not send message"));
                     }
                 }
             }
@@ -545,7 +625,7 @@ mod imp {
             if !self.can_compose_message() {
                 return;
             }
-            let Some(room) = self.room.upgrade() else {
+            let Some(timeline) = self.timeline.upgrade() else {
                 return;
             };
 
@@ -610,7 +690,7 @@ mod imp {
 
                         return;
                     }
-                };
+                }
             }
 
             let Some(geo_uri) = geo_uri_wrapper else {
@@ -640,13 +720,12 @@ mod imp {
             // even if they are empty.
             .add_mentions(Mentions::default());
 
-            let matrix_timeline = room.timeline().matrix_timeline();
+            let matrix_timeline = timeline.matrix_timeline();
             let handle = spawn_tokio!(async move { matrix_timeline.send(content.into()).await });
 
             if let Err(error) = handle.await.unwrap() {
                 error!("Could not send location: {error}");
-                let obj = self.obj();
-                toast!(obj, gettext("Could not send location"));
+                toast!(self.obj(), gettext("Could not send location"));
             }
         }
 
@@ -658,8 +737,7 @@ mod imp {
                 LocationError::Other => gettext("Could not retrieve current location"),
             };
 
-            let obj = self.obj();
-            toast!(obj, msg);
+            toast!(self.obj(), msg);
         }
 
         /// Send the attachment with the given data.
@@ -670,13 +748,13 @@ mod imp {
             info: AttachmentInfo,
             thumbnail: Option<Thumbnail>,
         ) {
-            let Some(room) = self.room.upgrade() else {
+            let Some(timeline) = self.timeline.upgrade() else {
                 return;
             };
 
             let config = AttachmentConfig::new().thumbnail(thumbnail).info(info);
 
-            let matrix_timeline = room.timeline().matrix_timeline();
+            let matrix_timeline = timeline.matrix_timeline();
 
             let handle = spawn_tokio!(async move {
                 matrix_timeline
@@ -687,8 +765,7 @@ mod imp {
 
             if let Err(error) = handle.await.unwrap() {
                 error!("Could not send file: {error}");
-                let obj = self.obj();
-                toast!(obj, gettext("Could not send file"));
+                toast!(self.obj(), gettext("Could not send file"));
             }
         }
 
@@ -881,12 +958,16 @@ mod imp {
         #[template_callback]
         fn handle_related_event_click(&self) {
             if let Some(related_to) = self.current_composer_state().related_to() {
-                self.obj()
+                if self
+                    .obj()
                     .activate_action(
                         "room-history.scroll-to-event",
-                        Some(&related_to.identifier().to_variant()),
+                        Some(&TimelineEventItemId::EventId(related_to.event_id()).to_variant()),
                     )
-                    .expect("action exists");
+                    .is_err()
+                {
+                    error!("Could not activate `room-history.scroll-to-event` action");
+                }
             }
         }
 
@@ -948,9 +1029,11 @@ mod imp {
 
         /// Send a typing notification for the given typing state.
         fn send_typing_notification(&self, typing: bool) {
-            let Some(room) = self.room.upgrade() else {
+            let Some(timeline) = self.timeline.upgrade() else {
                 return;
             };
+            let room = timeline.room();
+
             let Some(session) = room.session() else {
                 return;
             };
@@ -981,7 +1064,7 @@ impl MessageToolbar {
     }
 
     /// Set the event to reply to.
-    pub(crate) fn set_reply_to(&self, event: &Event) {
+    pub(crate) fn set_reply_to(&self, event: Event) {
         self.imp().set_reply_to(event);
     }
 
