@@ -5,27 +5,37 @@ use gettextrs::gettext;
 use gtk::{gdk, gio, glib, glib::clone, graphene::Point, CompositeTemplate};
 use matrix_sdk::ruma::EventId;
 use matrix_sdk_ui::timeline::TimelineEventItemId;
-use ruma::{api::client::receipt::create_receipt::v3::ReceiptType, OwnedEventId};
+use ruma::{
+    api::client::receipt::create_receipt::v3::ReceiptType, events::room::message::MessageType,
+    OwnedEventId,
+};
 use tracing::{error, warn};
 
 mod divider_row;
-mod item_row;
-mod item_row_context_menu;
+mod event_actions;
+mod event_row;
 mod member_timestamp;
 mod message_row;
 mod message_toolbar;
 mod read_receipts_list;
 mod sender_avatar;
-mod state_row;
+mod state;
 mod title;
 mod typing_row;
 mod verification_info_bar;
 
 use self::{
-    divider_row::DividerRow, item_row::ItemRow, item_row_context_menu::ItemRowContextMenu,
-    message_row::MessageRow, message_toolbar::MessageToolbar, read_receipts_list::ReadReceiptsList,
-    sender_avatar::SenderAvatar, state_row::StateRow, title::RoomHistoryTitle,
-    typing_row::TypingRow, verification_info_bar::VerificationInfoBar,
+    divider_row::DividerRow,
+    event_actions::*,
+    event_row::EventRow,
+    message_row::MessageRow,
+    message_toolbar::MessageToolbar,
+    read_receipts_list::ReadReceiptsList,
+    sender_avatar::SenderAvatar,
+    state::{StateGroupRow, StateRow},
+    title::RoomHistoryTitle,
+    typing_row::TypingRow,
+    verification_info_bar::VerificationInfoBar,
 };
 use super::{room_details, RoomDetails};
 use crate::{
@@ -33,10 +43,10 @@ use crate::{
     prelude::*,
     session::model::{
         Event, MemberList, Membership, ReceiptPosition, Room, TargetRoomCategory, Timeline,
-        TimelineState,
+        VirtualItem, VirtualItemKind,
     },
     spawn, toast,
-    utils::{template_callbacks::TemplateCallbacks, BoundObject},
+    utils::{BoundObject, GroupingListGroup, GroupingListModel, LoadingState, TemplateCallbacks},
     Window,
 };
 
@@ -48,7 +58,6 @@ const READ_TIMEOUT: Duration = Duration::from_secs(5);
 mod imp {
     use std::{
         cell::{Cell, OnceCell, RefCell},
-        marker::PhantomData,
         ops::ControlFlow,
     };
 
@@ -65,6 +74,8 @@ mod imp {
         #[template_child]
         pub(super) header_bar: TemplateChild<adw::HeaderBar>,
         #[template_child]
+        room_title: TemplateChild<RoomHistoryTitle>,
+        #[template_child]
         room_menu: TemplateChild<gtk::MenuButton>,
         #[template_child]
         listview: TemplateChild<gtk::ListView>,
@@ -88,34 +99,31 @@ mod imp {
         tombstoned_banner: TemplateChild<adw::Banner>,
         #[template_child]
         drag_overlay: TemplateChild<DragOverlay>,
-        item_context_menu: OnceCell<ItemRowContextMenu>,
+        /// The context menu for rows presenting an [`Event`].
+        event_context_menu: OnceCell<EventActionsContextMenu>,
         sender_context_menu: OnceCell<gtk::PopoverMenu>,
-        /// The room currently displayed.
-        #[property(get, set = Self::set_room, explicit_notify, nullable)]
-        room: BoundObject<Room>,
+        /// The timeline currently displayed.
+        #[property(get, set = Self::set_timeline, explicit_notify, nullable)]
+        timeline: BoundObject<Timeline>,
         /// Whether this is the only view visible, i.e. there is no sidebar.
         #[property(get, set)]
         is_only_view: Cell<bool>,
-        /// Whether this `RoomHistory` is empty, aka no room is currently
-        /// displayed.
-        #[property(get = Self::is_empty)]
-        is_empty: PhantomData<bool>,
         /// The members of the current room.
         ///
         /// We hold a strong reference here to keep the list in memory as long
         /// as the room is opened.
         room_members: RefCell<Option<MemberList>>,
-        timeline_handlers: RefCell<Vec<glib::SignalHandlerId>>,
         /// Whether the current room history scrolling is automatic.
         is_auto_scrolling: Cell<bool>,
         /// Whether the room history should stick to the newest message in the
         /// timeline.
         #[property(get)]
         is_sticky: Cell<bool>,
-        /// The `GtkSelectionModel` used in the list view.
-        selection_model: OnceCell<gtk::NoSelection>,
+        /// The `GroupingListModel` used in the list view.
+        grouping_model: OnceCell<GroupingListModel>,
         scroll_timeout: RefCell<Option<glib::SourceId>>,
         read_timeout: RefCell<Option<glib::SourceId>>,
+        room_handlers: RefCell<Vec<glib::SignalHandlerId>>,
         can_invite_handler: RefCell<Option<glib::SignalHandlerId>>,
         membership_handler: RefCell<Option<glib::SignalHandlerId>>,
         join_rule_handler: RefCell<Option<glib::SignalHandlerId>>,
@@ -128,10 +136,7 @@ mod imp {
         type ParentType = adw::Bin;
 
         fn class_init(klass: &mut Self::Class) {
-            RoomHistoryTitle::ensure_type();
-            ItemRow::ensure_type();
             VerificationInfoBar::ensure_type();
-            Timeline::ensure_type();
 
             Self::bind_template(klass);
             Self::bind_template_callbacks(klass);
@@ -181,15 +186,14 @@ mod imp {
                         return;
                     };
 
-                    let Some(event) = obj.room().and_then(|room| {
-                        room.timeline()
-                            .event_by_identifier(&TimelineEventItemId::EventId(event_id))
+                    let Some(event) = obj.timeline().and_then(|timeline| {
+                        timeline.event_by_identifier(&TimelineEventItemId::EventId(event_id))
                     }) else {
                         warn!("Could not find event to reply to");
                         return;
                     };
 
-                    obj.imp().message_toolbar.set_reply_to(&event);
+                    obj.imp().message_toolbar.set_reply_to(event);
                 },
             );
 
@@ -205,9 +209,8 @@ mod imp {
                         return;
                     };
 
-                    let Some(event) = obj.room().and_then(|room| {
-                        room.timeline()
-                            .event_by_identifier(&TimelineEventItemId::EventId(event_id))
+                    let Some(event) = obj.timeline().and_then(|timeline| {
+                        timeline.event_by_identifier(&TimelineEventItemId::EventId(event_id))
                     }) else {
                         warn!("Could not find event to edit");
                         return;
@@ -216,6 +219,39 @@ mod imp {
                     obj.imp().message_toolbar.set_edit(&event);
                 },
             );
+
+            klass.install_action("room-history.edit-latest-message", None, |obj, _, _| {
+                let Some(timeline) = obj.timeline() else {
+                    return;
+                };
+
+                let own_member = timeline.room().own_member();
+                let own_user_id = own_member.user_id();
+
+                // Find the latest editable message that was sent by our user.
+                let Some(event) = timeline
+                    .items()
+                    .iter::<glib::Object>()
+                    .rev()
+                    .find_map(|item| {
+                        item.ok().and_downcast::<Event>().filter(|event| {
+                            event.sender_id() == *own_user_id
+                                && event.event_id().is_some()
+                                && event.message().is_some_and(|message| {
+                                    matches!(
+                                        message.msgtype(),
+                                        MessageType::Text(_) | MessageType::Emote(_)
+                                    )
+                                })
+                        })
+                    })
+                else {
+                    warn!("Could not find latest event to edit");
+                    return;
+                };
+
+                obj.imp().message_toolbar.set_edit(&event);
+            });
         }
 
         fn instance_init(obj: &InitializingObject<Self>) {
@@ -246,29 +282,42 @@ mod imp {
         }
     }
 
-    impl WidgetImpl for RoomHistory {}
+    impl WidgetImpl for RoomHistory {
+        fn grab_focus(&self) -> bool {
+            if self.message_toolbar.grab_focus() {
+                true
+            } else {
+                self.room_title.grab_focus()
+            }
+        }
+    }
+
     impl BinImpl for RoomHistory {}
 
     #[gtk::template_callbacks]
     impl RoomHistory {
         /// Initialize the list view.
         fn init_listview(&self) {
-            let obj = self.obj();
-
             let factory = gtk::SignalListItemFactory::new();
-            factory.connect_setup(clone!(
-                #[weak]
-                obj,
-                move |_, item| {
-                    let Some(item) = item.downcast_ref::<gtk::ListItem>() else {
-                        error!("List item factory did not receive a list item: {item:?}");
+            factory.connect_setup(move |_, list_item| {
+                let Some(list_item) = list_item.downcast_ref::<gtk::ListItem>() else {
+                    error!("List item factory did not receive a list item: {list_item:?}");
+                    return;
+                };
+
+                list_item.set_activatable(false);
+                list_item.set_selectable(false);
+            });
+            factory.connect_bind(clone!(
+                #[weak(rename_to = imp)]
+                self,
+                move |_, list_item| {
+                    let Some(list_item) = list_item.downcast_ref::<gtk::ListItem>() else {
+                        error!("List item factory did not receive a list item: {list_item:?}");
                         return;
                     };
-                    let row = ItemRow::new(&obj);
-                    item.set_child(Some(&row));
-                    item.bind_property("item", &row, "item").build();
-                    item.set_activatable(false);
-                    item.set_selectable(false);
+
+                    imp.bind_list_item_to_item(list_item);
                 }
             ));
             self.listview.set_factory(Some(&factory));
@@ -277,7 +326,8 @@ mod imp {
             self.listview
                 .set_vscroll_policy(gtk::ScrollablePolicy::Natural);
 
-            self.listview.set_model(Some(self.selection_model()));
+            let selection_model = gtk::NoSelection::new(Some(self.grouping_model().clone()));
+            self.listview.set_model(Some(&selection_model));
 
             self.set_sticky(true);
             let adj = self.listview.vadjustment().unwrap();
@@ -340,9 +390,9 @@ mod imp {
 
         /// Disconnect all the signals.
         fn disconnect_all(&self) {
-            if let Some(room) = self.room.obj() {
-                for handler in self.timeline_handlers.take() {
-                    room.timeline().disconnect(handler);
+            if let Some(room) = self.room() {
+                for handler in self.room_handlers.take() {
+                    room.disconnect(handler);
                 }
 
                 if let Some(handler) = self.can_invite_handler.take() {
@@ -356,13 +406,13 @@ mod imp {
                 }
             }
 
-            self.room.disconnect_signals();
+            self.timeline.disconnect_signals();
         }
 
-        /// Set the room currently displayed.
+        /// Set the timeline currently displayed.
         #[allow(clippy::too_many_lines)]
-        fn set_room(&self, room: Option<Room>) {
-            if self.room.obj() == room {
+        fn set_timeline(&self, timeline: Option<Timeline>) {
+            if self.timeline.obj() == timeline {
                 return;
             }
 
@@ -374,8 +424,8 @@ mod imp {
                 source_id.remove();
             }
 
-            if let Some(room) = room {
-                let timeline = room.timeline();
+            if let Some(timeline) = timeline {
+                let room = timeline.room();
 
                 // Keep a strong reference to the members list before changing the model, so all
                 // events use the same list.
@@ -438,15 +488,12 @@ mod imp {
                     }
                 ));
 
-                self.room.set(
-                    room,
-                    vec![
-                        is_direct_handler,
-                        tombstoned_handler,
-                        successor_id_handler,
-                        successor_handler,
-                    ],
-                );
+                self.room_handlers.replace(vec![
+                    is_direct_handler,
+                    tombstoned_handler,
+                    successor_id_handler,
+                    successor_handler,
+                ]);
 
                 let empty_handler = timeline.connect_is_empty_notify(clone!(
                     #[weak(rename_to = imp)]
@@ -465,22 +512,22 @@ mod imp {
                         // Always test if we need to load more when the timeline is ready.
                         // This is mostly to make sure that we load events if the timeline was not
                         // initialized when the room was opened.
-                        if timeline.state() == TimelineState::Ready {
+                        if timeline.state() == LoadingState::Ready {
                             imp.load_more_events_if_needed();
                         }
                     }
                 ));
 
-                self.timeline_handlers
-                    .replace(vec![empty_handler, state_handler]);
+                self.timeline
+                    .set(timeline.clone(), vec![empty_handler, state_handler]);
 
                 timeline.remove_empty_typing_row();
-                self.selection_model().set_model(Some(&timeline.items()));
+                self.grouping_model().set_model(Some(timeline.items()));
 
                 self.trigger_read_receipts_update();
                 self.scroll_down();
             } else {
-                self.selection_model().set_model(None::<&gio::ListModel>);
+                self.grouping_model().set_model(None::<gio::ListModel>);
             }
 
             self.update_view();
@@ -489,21 +536,46 @@ mod imp {
             self.update_tombstoned_banner();
             self.update_invite_action();
 
-            let obj = self.obj();
-            obj.notify_room();
-            obj.notify_is_empty();
+            self.obj().notify_timeline();
         }
 
-        /// The `GtkSelectionModel` used in the list view.
-        fn selection_model(&self) -> &gtk::NoSelection {
-            self.selection_model
-                .get_or_init(|| gtk::NoSelection::new(gio::ListModel::NONE.cloned()))
+        /// The room of the current timeline, if any.
+        pub(super) fn room(&self) -> Option<Room> {
+            self.timeline.obj().map(|timeline| timeline.room())
         }
 
-        /// Whether this `RoomHistory` is empty, aka no room is currently
-        /// displayed.
-        fn is_empty(&self) -> bool {
-            self.room.obj().is_none()
+        /// The `GroupingListModel` used in the list view.
+        fn grouping_model(&self) -> &GroupingListModel {
+            self.grouping_model.get_or_init(|| {
+                GroupingListModel::new(|lhs, rhs| {
+                    lhs.downcast_ref::<Event>()
+                        .is_some_and(Event::is_state_group_event)
+                        && rhs
+                            .downcast_ref::<Event>()
+                            .is_some_and(Event::is_state_group_event)
+                })
+            })
+        }
+
+        /// Bind the given `GtkListItem` to its item.
+        fn bind_list_item_to_item(&self, list_item: &gtk::ListItem) {
+            let Some(item) = list_item.item() else {
+                error!("List item does not have an item",);
+                list_item.set_child(None::<&gtk::Widget>);
+                return;
+            };
+
+            if let Some(event) = item.downcast_ref::<Event>() {
+                let child = list_item.child_or_else::<EventRow>(|| EventRow::new(&self.obj()));
+                child.set_event(Some(event.clone()));
+            } else if let Some(virtual_item) = item.downcast_ref::<VirtualItem>() {
+                set_virtual_item_child(list_item, virtual_item);
+            } else if let Some(group) = item.downcast_ref::<GroupingListGroup>() {
+                let child = list_item.child_or_default::<StateGroupRow>();
+                child.set_group(Some(group.clone()));
+            } else {
+                error!("Could not build widget for unsupported room history item: {item:?}");
+            }
         }
 
         /// Handle when the scroll value changed.
@@ -521,8 +593,8 @@ mod imp {
 
                 // Remove the typing row if the user scrolls up.
                 if !is_at_bottom {
-                    if let Some(room) = self.room.obj() {
-                        room.timeline().remove_empty_typing_row();
+                    if let Some(timeline) = self.timeline.obj() {
+                        timeline.remove_empty_typing_row();
                     }
                 }
 
@@ -577,7 +649,7 @@ mod imp {
 
             self.set_is_auto_scrolling(true);
 
-            let n_items = self.selection_model().n_items();
+            let n_items = self.grouping_model().n_items();
 
             if n_items > 0 {
                 // Wait until the next tick, to make sure that the GtkListView has created the
@@ -616,7 +688,7 @@ mod imp {
 
         /// Update the room menu for the current state.
         fn update_room_menu(&self) {
-            let Some(room) = self.room.obj() else {
+            let Some(room) = self.room() else {
                 self.room_menu.set_visible(false);
                 return;
             };
@@ -638,12 +710,12 @@ mod imp {
 
         /// Update the view for the current state.
         fn update_view(&self) {
-            let Some(room) = self.room.obj() else {
+            let Some(timeline) = self.timeline.obj() else {
                 return;
             };
 
-            let visible_child_name = if room.timeline().is_empty() {
-                if room.timeline().state() == TimelineState::Error {
+            let visible_child_name = if timeline.is_empty() {
+                if timeline.state() == LoadingState::Error {
                     "error"
                 } else {
                     "loading"
@@ -654,12 +726,12 @@ mod imp {
             self.stack.set_visible_child_name(visible_child_name);
         }
 
-        /// Whether we need to load more events.
-        fn needs_more_events(&self) -> bool {
-            if self.selection_model().n_items() == 0 {
+        /// Whether we need to load more events at the start of the timeline.
+        fn needs_more_events_at_the_start(&self) -> bool {
+            if self.grouping_model().n_items() == 0 {
                 // We definitely want events if the history is empty.
                 return true;
-            };
+            }
 
             // Load more messages when the user gets close to the top of the known room
             // history. Use the page size twice to detect if the user gets close to
@@ -671,19 +743,16 @@ mod imp {
             adj.value() < adj.page_size() * 2.0
         }
 
-        /// Load more events at the beginning of the history if needed.
+        /// Load more events in the history if needed.
         fn load_more_events_if_needed(&self) {
-            if !self.needs_more_events() {
-                return;
+            if self.needs_more_events_at_the_start() {
+                self.load_more_events_at_the_start();
             }
-
-            self.load_more_events();
         }
 
         /// Load more events at the beginning of the history.
-        #[template_callback]
-        fn load_more_events(&self) {
-            let Some(room) = self.room.obj() else {
+        fn load_more_events_at_the_start(&self) {
+            let Some(timeline) = self.timeline.obj() else {
                 return;
             };
 
@@ -691,14 +760,14 @@ mod imp {
                 #[weak(rename_to = imp)]
                 self,
                 async move {
-                    room.timeline()
-                        .load(clone!(
+                    timeline
+                        .paginate_backwards(clone!(
                             #[weak]
                             imp,
                             #[upgrade_or]
                             ControlFlow::Break(()),
                             move || {
-                                if imp.needs_more_events() {
+                                if imp.needs_more_events_at_the_start() {
                                     ControlFlow::Continue(())
                                 } else {
                                     ControlFlow::Break(())
@@ -710,13 +779,21 @@ mod imp {
             ));
         }
 
+        /// Load more events in the history, regardless of if we need them.
+        ///
+        /// This should only be used to try to fix timeline loading errors.
+        #[template_callback]
+        fn load_more_events(&self) {
+            self.load_more_events_at_the_start();
+        }
+
         /// Scroll to the event with the given identifier.
         fn scroll_to_event(&self, key: &TimelineEventItemId) {
-            let Some(room) = self.room.obj() else {
+            let Some(timeline) = self.timeline.obj() else {
                 return;
             };
 
-            if let Some(pos) = room.timeline().find_event_position(key) {
+            if let Some(pos) = timeline.find_event_position(key) {
                 let pos = pos as u32;
                 self.listview
                     .scroll_to(pos, gtk::ListScrollFlags::FOCUS, None);
@@ -725,11 +802,10 @@ mod imp {
 
         /// Trigger the process to update read receipts.
         fn trigger_read_receipts_update(&self) {
-            let Some(room) = self.room.obj() else {
+            let Some(timeline) = self.timeline.obj() else {
                 return;
             };
 
-            let timeline = room.timeline();
             if !timeline.is_empty() {
                 if let Some(source_id) = self.scroll_timeout.take() {
                     source_id.remove();
@@ -780,7 +856,7 @@ mod imp {
                 #[weak(rename_to = imp)]
                 self,
                 async move {
-                    let Some(room) = imp.room.obj() else { return };
+                    let Some(room) = imp.room() else { return };
                     room.send_receipt(ReceiptType::Read, position).await;
                 }
             ));
@@ -798,7 +874,7 @@ mod imp {
                 #[weak(rename_to = imp)]
                 self,
                 async move {
-                    let Some(room) = imp.room.obj() else { return };
+                    let Some(room) = imp.room() else { return };
                     room.send_receipt(ReceiptType::FullyRead, position).await;
                 }
             ));
@@ -841,9 +917,8 @@ mod imp {
                 if top_in_view || bottom_in_view || content_in_view {
                     if let Some(event_id) = item
                         .first_child()
-                        .and_downcast::<ItemRow>()
-                        .and_then(|row| row.item())
-                        .and_downcast::<Event>()
+                        .and_downcast::<EventRow>()
+                        .and_then(|row| row.event())
                         .and_then(|event| event.event_id())
                     {
                         return Some(event_id);
@@ -861,7 +936,7 @@ mod imp {
         fn update_tombstoned_banner(&self) {
             let banner = &self.tombstoned_banner;
 
-            let Some(room) = self.room.obj() else {
+            let Some(room) = self.room() else {
                 banner.set_revealed(false);
                 return;
             };
@@ -888,7 +963,7 @@ mod imp {
 
         /// Leave the room.
         async fn leave(&self) {
-            let Some(room) = self.room.obj() else {
+            let Some(room) = self.room() else {
                 return;
             };
 
@@ -917,7 +992,7 @@ mod imp {
 
         /// Join the room.
         async fn join(&self) {
-            let Some(room) = self.room.obj() else {
+            let Some(room) = self.room() else {
                 return;
             };
 
@@ -940,7 +1015,7 @@ mod imp {
 
         /// Forget the room.
         async fn forget(&self) {
-            let Some(room) = self.room.obj() else {
+            let Some(room) = self.room() else {
                 return;
             };
 
@@ -956,7 +1031,7 @@ mod imp {
 
         // Update the invite action according to the current state.
         fn update_invite_action(&self) {
-            let Some(room) = self.room.obj() else {
+            let Some(room) = self.room() else {
                 return;
             };
 
@@ -970,7 +1045,7 @@ mod imp {
         /// Join or view the successor of the room, if possible.
         #[template_callback]
         async fn join_or_view_successor(&self) {
-            let Some(room) = self.room.obj() else {
+            let Some(room) = self.room() else {
                 return;
             };
             let Some(session) = room.session() else {
@@ -987,7 +1062,7 @@ mod imp {
                     return;
                 };
 
-                window.show_room(session.session_id(), successor.room_id());
+                window.session_view().select_room(successor);
             } else if let Some(successor_id) = room.successor_id().map(ToOwned::to_owned) {
                 let via = successor_id
                     .server_name()
@@ -1005,9 +1080,9 @@ mod imp {
             }
         }
 
-        /// The context menu for the item rows.
-        pub(super) fn item_context_menu(&self) -> &ItemRowContextMenu {
-            self.item_context_menu.get_or_init(Default::default)
+        /// The context menu for rows presenting an [`Event`].
+        pub(super) fn event_context_menu(&self) -> &EventActionsContextMenu {
+            self.event_context_menu.get_or_init(Default::default)
         }
 
         /// The context menu for the sender avatars.
@@ -1044,7 +1119,7 @@ impl RoomHistory {
     }
 
     /// The message toolbar of the room history.
-    fn message_toolbar(&self) -> &MessageToolbar {
+    pub(super) fn message_toolbar(&self) -> &MessageToolbar {
         &self.imp().message_toolbar
     }
 
@@ -1053,7 +1128,7 @@ impl RoomHistory {
     /// If `subpage_name` is set, the room details will be opened on the given
     /// subpage.
     pub(crate) fn open_room_details(&self, subpage_name: Option<room_details::SubpageName>) {
-        let Some(room) = self.room() else {
+        let Some(room) = self.imp().room() else {
             return;
         };
 
@@ -1080,9 +1155,9 @@ impl RoomHistory {
         self.imp().message_toolbar.handle_paste_action();
     }
 
-    /// The context menu for the item rows.
-    fn item_context_menu(&self) -> &ItemRowContextMenu {
-        self.imp().item_context_menu()
+    /// The context menu for rows presenting an [`Event`].
+    fn event_context_menu(&self) -> &EventActionsContextMenu {
+        self.imp().event_context_menu()
     }
 
     /// The context menu for the sender avatars.
@@ -1090,3 +1165,41 @@ impl RoomHistory {
         self.imp().sender_context_menu()
     }
 }
+
+/// Set the proper child of the given `GtkListItem` for the given
+/// [`VirtualItem`].
+///
+/// Constructs or reuses the child widget as necessary.
+fn set_virtual_item_child(list_item: &gtk::ListItem, virtual_item: &VirtualItem) {
+    let kind = &virtual_item.kind();
+
+    match kind {
+        VirtualItemKind::Spinner => {
+            if !list_item
+                .child()
+                .is_some_and(|widget| widget.is::<adw::Spinner>())
+            {
+                let spinner = adw::Spinner::builder()
+                    .margin_top(12)
+                    .margin_bottom(12)
+                    .height_request(24)
+                    .width_request(24)
+                    .build();
+                spinner.add_css_class("room-history-row");
+                spinner.set_accessible_role(gtk::AccessibleRole::ListItem);
+                list_item.set_child(Some(&spinner));
+            }
+        }
+        VirtualItemKind::Typing => {
+            let child = list_item.child_or_default::<TypingRow>();
+            let typing_list = virtual_item.room().typing_list();
+            child.set_list(Some(typing_list));
+        }
+        VirtualItemKind::TimelineStart
+        | VirtualItemKind::DayDivider(_)
+        | VirtualItemKind::NewMessages => {
+            let divider = list_item.child_or_default::<DividerRow>();
+            divider.set_virtual_item(Some(virtual_item));
+        }
+    }
+}
