@@ -172,24 +172,14 @@ mod imp {
 
             if let Some(item) = item {
                 if let Some(section) = item.downcast_ref::<SidebarSection>() {
-                    let child = if let Some(child) = obj.child().and_downcast::<SidebarSectionRow>()
-                    {
-                        child
-                    } else {
+                    let child = obj.child_or_else::<SidebarSectionRow>(|| {
                         let child = SidebarSectionRow::new();
-                        obj.set_child(Some(&child));
                         obj.update_relation(&[Relation::LabelledBy(&[child.labelled_by()])]);
                         child
-                    };
+                    });
                     child.set_section(Some(section.clone()));
                 } else if let Some(room) = item.downcast_ref::<Room>() {
-                    let child = if let Some(child) = obj.child().and_downcast::<SidebarRoomRow>() {
-                        child
-                    } else {
-                        let child = SidebarRoomRow::new();
-                        obj.set_child(Some(&child));
-                        child
-                    };
+                    let child = obj.child_or_default::<SidebarRoomRow>();
 
                     let room_is_direct_handler = room.connect_is_direct_notify(clone!(
                         #[weak(rename_to = imp)]
@@ -222,26 +212,10 @@ mod imp {
 
                     child.set_room(Some(room.clone()));
                 } else if let Some(icon_item) = item.downcast_ref::<SidebarIconItem>() {
-                    let child =
-                        if let Some(child) = obj.child().and_downcast::<SidebarIconItemRow>() {
-                            child
-                        } else {
-                            let child = SidebarIconItemRow::new();
-                            obj.set_child(Some(&child));
-                            child
-                        };
-
+                    let child = obj.child_or_default::<SidebarIconItemRow>();
                     child.set_icon_item(Some(icon_item.clone()));
                 } else if let Some(verification) = item.downcast_ref::<IdentityVerification>() {
-                    let child =
-                        if let Some(child) = obj.child().and_downcast::<SidebarVerificationRow>() {
-                            child
-                        } else {
-                            let child = SidebarVerificationRow::new();
-                            obj.set_child(Some(&child));
-                            child
-                        };
-
+                    let child = obj.child_or_default::<SidebarVerificationRow>();
                     child.set_identity_verification(Some(verification.clone()));
                 } else {
                     panic!("Wrong row item: {item:?}");
@@ -577,7 +551,7 @@ mod imp {
                 obj.remove_css_class("drop-disabled");
                 obj.remove_css_class("drop-empty");
                 obj.remove_css_class("drop-active");
-            };
+            }
 
             if let Some(section_row) = obj.child().and_downcast::<SidebarSectionRow>() {
                 section_row.set_show_label_for_room_category(source_category);
@@ -766,9 +740,8 @@ mod imp {
         /// Forget the given room.
         async fn forget_room(&self, room: &Room) {
             if room.forget().await.is_err() {
-                let obj = self.obj();
                 toast!(
-                    obj,
+                    self.obj(),
                     // Translators: Do NOT translate the content between '{' and '}', this is a variable name.
                     gettext("Could not forget {room}"),
                     @room,
@@ -788,12 +761,12 @@ mod imp {
                     error!("Could not mark room as direct chat: {error}");
                     // Translators: Do NOT translate the content between '{' and '}', this is a
                     // variable name.
-                    toast!(obj, gettext("Could not mark {room} as direct chat"), @room,);
+                    toast!(obj, gettext("Could not mark {room} as direct chat"), @room);
                 } else {
                     error!("Could not unmark room as direct chat: {error}");
                     // Translators: Do NOT translate the content between '{' and '}', this is a
                     // variable name.
-                    toast!(obj, gettext("Could not unmark {room} as direct chat"), @room,);
+                    toast!(obj, gettext("Could not unmark {room} as direct chat"), @room);
                 }
             }
         }
@@ -811,3 +784,13 @@ impl SidebarRow {
         glib::Object::builder().property("sidebar", sidebar).build()
     }
 }
+
+impl ChildPropertyExt for SidebarRow {
+    fn child_property(&self) -> Option<gtk::Widget> {
+        self.child()
+    }
+
+    fn set_child_property(&self, child: Option<&impl IsA<gtk::Widget>>) {
+        self.set_child(child);
+    }
+}
