@@ -1,15 +1,18 @@
 //! Collection of methods related to the Matrix specification.
 
-use std::{borrow::Cow, str::FromStr};
+use std::{borrow::Cow, fmt, str::FromStr};
 
 use gettextrs::gettext;
 use gtk::{glib, prelude::*};
 use matrix_sdk::{
-    authentication::matrix::{MatrixSession, MatrixSessionTokens},
+    authentication::{
+        matrix::MatrixSession,
+        oauth::{OAuthSession, UserSession},
+    },
     config::RequestConfig,
     deserialized_responses::RawAnySyncOrStrippedTimelineEvent,
     encryption::{BackupDownloadStrategy, EncryptionSettings},
-    Client, ClientBuildError, SessionMeta,
+    AuthSession, Client, ClientBuildError, SessionMeta, SessionTokens,
 };
 use ruma::{
     events::{
@@ -28,37 +31,32 @@ use ruma::{
     RoomId, RoomOrAliasId, UserId,
 };
 use thiserror::Error;
+use tracing::error;
 
-pub mod ext_traits;
+pub(crate) mod ext_traits;
 mod media_message;
 
-pub use self::media_message::{MediaMessage, VisualMediaMessage};
-use crate::{
-    components::Pill,
-    gettext_f,
-    prelude::*,
-    secret::{Secret, StoredSession},
-    session::model::{RemoteRoom, Room},
-};
+pub(crate) use self::media_message::*;
+use crate::{components::Pill, gettext_f, prelude::*, secret::StoredSession, session::model::Room};
 
 /// The result of a password validation.
 #[derive(Debug, Default, Clone, Copy)]
 #[allow(clippy::struct_excessive_bools)]
-pub struct PasswordValidity {
+pub(crate) struct PasswordValidity {
     /// Whether the password includes at least one lowercase letter.
-    pub has_lowercase: bool,
+    pub(crate) has_lowercase: bool,
     /// Whether the password includes at least one uppercase letter.
-    pub has_uppercase: bool,
+    pub(crate) has_uppercase: bool,
     /// Whether the password includes at least one number.
-    pub has_number: bool,
+    pub(crate) has_number: bool,
     /// Whether the password includes at least one symbol.
-    pub has_symbol: bool,
+    pub(crate) has_symbol: bool,
     /// Whether the password is at least 8 characters long.
-    pub has_length: bool,
+    pub(crate) has_length: bool,
     /// The percentage of checks passed for the password, between 0 and 100.
     ///
     /// If progress is 100, the password is valid.
-    pub progress: u32,
+    pub(crate) progress: u32,
 }
 
 impl PasswordValidity {
@@ -73,7 +71,7 @@ impl PasswordValidity {
 /// number and a symbol and be at a minimum 8 characters in length.
 ///
 /// See: <https://spec.matrix.org/v1.1/client-server-api/#notes-on-password-management>
-pub fn validate_password(password: &str) -> PasswordValidity {
+pub(crate) fn validate_password(password: &str) -> PasswordValidity {
     let mut validity = PasswordValidity::new();
 
     for char in password.chars() {
@@ -113,7 +111,7 @@ pub fn validate_password(password: &str) -> PasswordValidity {
 
 /// An deserialized event received in a sync response.
 #[derive(Debug, Clone)]
-pub enum AnySyncOrStrippedTimelineEvent {
+pub(crate) enum AnySyncOrStrippedTimelineEvent {
     /// An event from a joined or left room.
     Sync(AnySyncTimelineEvent),
     /// An event from an invited room.
@@ -122,7 +120,9 @@ pub enum AnySyncOrStrippedTimelineEvent {
 
 impl AnySyncOrStrippedTimelineEvent {
     /// Deserialize the given raw event.
-    pub fn from_raw(raw: &RawAnySyncOrStrippedTimelineEvent) -> Result<Self, serde_json::Error> {
+    pub(crate) fn from_raw(
+        raw: &RawAnySyncOrStrippedTimelineEvent,
+    ) -> Result<Self, serde_json::Error> {
         let ev = match raw {
             RawAnySyncOrStrippedTimelineEvent::Sync(ev) => Self::Sync(ev.deserialize()?),
             RawAnySyncOrStrippedTimelineEvent::Stripped(ev) => Self::Stripped(ev.deserialize()?),
@@ -132,7 +132,7 @@ impl AnySyncOrStrippedTimelineEvent {
     }
 
     /// The sender of the event.
-    pub fn sender(&self) -> &UserId {
+    pub(crate) fn sender(&self) -> &UserId {
         match self {
             AnySyncOrStrippedTimelineEvent::Sync(ev) => ev.sender(),
             AnySyncOrStrippedTimelineEvent::Stripped(ev) => ev.sender(),
@@ -140,7 +140,7 @@ impl AnySyncOrStrippedTimelineEvent {
     }
 
     /// The ID of the event, if it's not a stripped state event.
-    pub fn event_id(&self) -> Option<&EventId> {
+    pub(crate) fn event_id(&self) -> Option<&EventId> {
         match self {
             AnySyncOrStrippedTimelineEvent::Sync(ev) => Some(ev.event_id()),
             AnySyncOrStrippedTimelineEvent::Stripped(_) => None,
@@ -154,7 +154,7 @@ impl AnySyncOrStrippedTimelineEvent {
 /// localized string.
 ///
 /// Returns `None` if the event type is not supported.
-pub fn get_event_body(
+pub(crate) fn get_event_body(
     event: &AnySyncOrStrippedTimelineEvent,
     sender_name: &str,
     own_user: &UserId,
@@ -176,7 +176,7 @@ pub fn get_event_body(
 /// If it's a media message, this will return a localized body.
 ///
 /// Returns `None` if the message type is not supported.
-pub fn get_message_event_body(
+pub(crate) fn get_message_event_body(
     event: &AnySyncMessageLikeEvent,
     sender_name: &str,
     show_sender: bool,
@@ -237,7 +237,7 @@ fn text_event_body(message: String, sender_name: &str, show_sender: bool) -> Str
 /// This will return a localized body.
 ///
 /// Returns `None` if the state event type is not supported.
-pub fn get_stripped_state_event_body(
+pub(crate) fn get_stripped_state_event_body(
     event: &AnyStrippedStateEvent,
     sender_name: &str,
     own_user: &UserId,
@@ -257,7 +257,7 @@ pub fn get_stripped_state_event_body(
 
 /// All errors that can occur when setting up the Matrix client.
 #[derive(Error, Debug)]
-pub enum ClientSetupError {
+pub(crate) enum ClientSetupError {
     /// An error when building the client.
     #[error(transparent)]
     Client(#[from] ClientBuildError),
@@ -267,6 +267,9 @@ pub enum ClientSetupError {
     /// An error creating the unique local ID of the session.
     #[error("Could not generate unique session ID")]
     NoSessionId,
+    /// An error accessing the session tokens.
+    #[error("Could not access session tokens")]
+    NoSessionTokens,
 }
 
 impl UserFacingError for ClientSetupError {
@@ -275,14 +278,17 @@ impl UserFacingError for ClientSetupError {
             Self::Client(err) => err.to_user_facing(),
             Self::Sdk(err) => err.to_user_facing(),
             Self::NoSessionId => gettext("Could not generate unique session ID"),
+            Self::NoSessionTokens => gettext("Could not access the session tokens"),
         }
     }
 }
 
 /// Create a [`Client`] with the given stored session.
-pub async fn client_with_stored_session(
+pub(crate) async fn client_with_stored_session(
     session: StoredSession,
+    tokens: SessionTokens,
 ) -> Result<Client, ClientSetupError> {
+    let has_refresh_token = tokens.refresh_token.is_some();
     let data_path = session.data_path();
     let cache_path = session.cache_path();
 
@@ -290,19 +296,20 @@ pub async fn client_with_stored_session(
         homeserver,
         user_id,
         device_id,
-        id: _,
-        secret: Secret {
-            access_token,
-            passphrase,
-        },
+        passphrase,
+        client_id,
+        ..
     } = session;
 
-    let session_data = MatrixSession {
-        meta: SessionMeta { user_id, device_id },
-        tokens: MatrixSessionTokens {
-            access_token,
-            refresh_token: None,
-        },
+    let meta = SessionMeta { user_id, device_id };
+    let session_data: AuthSession = if let Some(client_id) = client_id {
+        OAuthSession {
+            user: UserSession { meta, tokens },
+            client_id,
+        }
+        .into()
+    } else {
+        MatrixSession { meta, tokens }.into()
     };
 
     let encryption_settings = EncryptionSettings {
@@ -311,16 +318,20 @@ pub async fn client_with_stored_session(
         auto_enable_backups: true,
     };
 
-    let client = Client::builder()
+    let mut client_builder = Client::builder()
         .homeserver_url(homeserver)
         .sqlite_store_with_cache_path(data_path, cache_path, Some(&passphrase))
         // force_auth option to solve an issue with some servers configuration to require
         // auth for profiles:
         // https://gitlab.gnome.org/World/fractal/-/issues/934
         .request_config(RequestConfig::new().retry_limit(2).force_auth())
-        .with_encryption_settings(encryption_settings)
-        .build()
-        .await?;
+        .with_encryption_settings(encryption_settings);
+
+    if has_refresh_token {
+        client_builder = client_builder.handle_refresh_tokens();
+    }
+
+    let client = client_builder.build().await?;
 
     client.restore_session(session_data).await?;
 
@@ -330,7 +341,7 @@ pub async fn client_with_stored_session(
 /// Find mentions in the given HTML string.
 ///
 /// Returns a list of `(pill, mention_content)` tuples.
-pub fn find_html_mentions(html: &str, room: &Room) -> Vec<(Pill, StrTendril)> {
+pub(crate) fn find_html_mentions(html: &str, room: &Room) -> Vec<(Pill, StrTendril)> {
     let mut mentions = Vec::new();
     let html = Html::parse(html);
 
@@ -381,7 +392,7 @@ fn node_as_mention(node: &NodeRef, room: &Room) -> Option<(Pill, StrTendril)> {
 }
 
 /// The textual representation of a room mention.
-pub const AT_ROOM: &str = "@room";
+pub(crate) const AT_ROOM: &str = "@room";
 
 /// Find `@room` in the given string.
 ///
@@ -389,7 +400,7 @@ pub const AT_ROOM: &str = "@room";
 /// it in the `body`.
 ///
 /// Returns the position of the first match.
-pub fn find_at_room(s: &str) -> Option<usize> {
+pub(crate) fn find_at_room(s: &str) -> Option<usize> {
     for (pos, _) in s.match_indices(AT_ROOM) {
         let is_at_word_start = pos == 0 || s[..pos].ends_with(char_is_ascii_word_boundary);
         if !is_at_word_start {
@@ -416,7 +427,7 @@ fn char_is_ascii_word_boundary(c: char) -> bool {
 }
 
 /// Compare two raw JSON sources.
-pub fn raw_eq<T, U>(lhs: Option<&Raw<T>>, rhs: Option<&Raw<U>>) -> bool {
+pub(crate) fn raw_eq<T, U>(lhs: Option<&Raw<T>>, rhs: Option<&Raw<U>>) -> bool {
     let Some(lhs) = lhs else {
         // They are equal only if both are `None`.
         return rhs.is_none();
@@ -431,7 +442,7 @@ pub fn raw_eq<T, U>(lhs: Option<&Raw<T>>, rhs: Option<&Raw<U>>) -> bool {
 
 /// A URI for a Matrix ID.
 #[derive(Debug, Clone, PartialEq, Eq)]
-pub enum MatrixIdUri {
+pub(crate) enum MatrixIdUri {
     /// A room.
     Room(MatrixRoomIdUri),
     /// A user.
@@ -467,7 +478,7 @@ impl MatrixIdUri {
     }
 
     /// Try parsing a `&str` into a `MatrixIdUri`.
-    pub fn parse(s: &str) -> Result<Self, MatrixIdUriParseError> {
+    pub(crate) fn parse(s: &str) -> Result<Self, MatrixIdUriParseError> {
         if let Ok(uri) = MatrixToUri::parse(s) {
             return uri.try_into();
         }
@@ -476,7 +487,7 @@ impl MatrixIdUri {
     }
 
     /// Try to construct a [`Pill`] from this ID in the given room.
-    pub fn into_pill(self, room: &Room) -> Option<Pill> {
+    pub(crate) fn into_pill(self, room: &Room) -> Option<Pill> {
         match self {
             Self::Room(room_uri) => {
                 let session = room.session()?;
@@ -485,7 +496,7 @@ impl MatrixIdUri {
                     .get_by_identifier(&room_uri.id)
                     .as_ref()
                     .map(Pill::new)
-                    .or_else(|| Some(Pill::new(&RemoteRoom::new(&session, room_uri))))
+                    .or_else(|| Some(Pill::new(&session.remote_cache().room(room_uri))))
             }
             Self::User(user_id) => {
                 // We should have a strong reference to the list wherever we show a user pill,
@@ -496,6 +507,32 @@ impl MatrixIdUri {
             Self::Event(_) => None,
         }
     }
+
+    /// Get this ID as a `matrix:` URI.
+    pub(crate) fn as_matrix_uri(&self) -> MatrixUri {
+        match self {
+            MatrixIdUri::Room(room_uri) => match <&RoomId>::try_from(&*room_uri.id) {
+                Ok(room_id) => room_id.matrix_uri_via(room_uri.via.clone(), false),
+                Err(room_alias) => room_alias.matrix_uri(false),
+            },
+            MatrixIdUri::User(user_id) => user_id.matrix_uri(false),
+            MatrixIdUri::Event(event_uri) => {
+                let room_id = <&RoomId>::try_from(&*event_uri.room_uri.id)
+                    .expect("room alias should not be used to construct event URI");
+
+                room_id.matrix_event_uri_via(
+                    event_uri.event_id.clone(),
+                    event_uri.room_uri.via.clone(),
+                )
+            }
+        }
+    }
+}
+
+impl fmt::Display for MatrixIdUri {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        self.as_matrix_uri().fmt(f)
+    }
 }
 
 impl TryFrom<&MatrixUri> for MatrixIdUri {
@@ -576,6 +613,12 @@ impl StaticVariantType for MatrixIdUri {
     }
 }
 
+impl ToVariant for MatrixIdUri {
+    fn to_variant(&self) -> glib::Variant {
+        self.to_string().to_variant()
+    }
+}
+
 impl FromVariant for MatrixIdUri {
     fn from_variant(variant: &glib::Variant) -> Option<Self> {
         Self::parse(&variant.get::<String>()?).ok()
@@ -584,16 +627,16 @@ impl FromVariant for MatrixIdUri {
 
 /// A URI for a Matrix room ID.
 #[derive(Debug, Clone, PartialEq, Eq)]
-pub struct MatrixRoomIdUri {
+pub(crate) struct MatrixRoomIdUri {
     /// The room ID.
-    pub id: OwnedRoomOrAliasId,
+    pub(crate) id: OwnedRoomOrAliasId,
     /// Matrix servers usable to route a `RoomId`.
-    pub via: Vec<OwnedServerName>,
+    pub(crate) via: Vec<OwnedServerName>,
 }
 
 impl MatrixRoomIdUri {
     /// Try parsing a `&str` into a `MatrixRoomIdUri`.
-    pub fn parse(s: &str) -> Option<MatrixRoomIdUri> {
+    pub(crate) fn parse(s: &str) -> Option<MatrixRoomIdUri> {
         MatrixIdUri::parse(s)
             .ok()
             .and_then(|uri| match uri {
@@ -636,7 +679,7 @@ impl From<&MatrixRoomIdUri> for MatrixUri {
 
 /// A URI for a Matrix event ID.
 #[derive(Debug, Clone, PartialEq, Eq)]
-pub struct MatrixEventIdUri {
+pub(crate) struct MatrixEventIdUri {
     /// The event ID.
     pub event_id: OwnedEventId,
     /// The event's room ID URI.
@@ -645,7 +688,7 @@ pub struct MatrixEventIdUri {
 
 /// Errors encountered when parsing a Matrix ID URI.
 #[derive(Debug, Clone, Error)]
-pub enum MatrixIdUriParseError {
+pub(crate) enum MatrixIdUriParseError {
     /// Not a valid Matrix URI.
     #[error(transparent)]
     InvalidUri(#[from] IdParseError),
