@@ -1,9 +1,9 @@
 //! Collection of methods for images.
 
-use std::{error::Error, fmt, str::FromStr, sync::Arc};
+use std::{cmp::Ordering, error::Error, fmt, str::FromStr, sync::Arc};
 
 use gettextrs::gettext;
-use gtk::{gdk, gio, graphene, gsk, prelude::*};
+use gtk::{gdk, gio, glib, graphene, gsk, prelude::*};
 use matrix_sdk::{
     attachment::{BaseImageInfo, Thumbnail},
     media::{MediaFormat, MediaRequestParameters, MediaThumbnailSettings},
@@ -27,10 +27,12 @@ mod queue;
 pub(crate) use queue::{ImageRequestPriority, IMAGE_QUEUE};
 
 use super::{FrameDimensions, MediaFileError};
-use crate::{components::AnimatedImagePaintable, spawn_tokio, utils::File, DISABLE_GLYCIN_SANDBOX};
+use crate::{
+    components::AnimatedImagePaintable, spawn_tokio, utils::File, DISABLE_GLYCIN_SANDBOX, RUNTIME,
+};
 
 /// The maximum dimensions of a thumbnail in the timeline.
-pub const THUMBNAIL_MAX_DIMENSIONS: FrameDimensions = FrameDimensions {
+pub(crate) const THUMBNAIL_MAX_DIMENSIONS: FrameDimensions = FrameDimensions {
     width: 600,
     height: 400,
 };
@@ -114,7 +116,7 @@ async fn load_image(
 
 /// An image that was just loaded.
 #[derive(Clone)]
-pub struct Image {
+pub(crate) struct Image {
     /// The file of the image.
     file: File,
     /// The image loader.
@@ -140,7 +142,7 @@ impl From<Image> for gdk::Paintable {
 }
 
 /// An API to load image information.
-pub enum ImageInfoLoader {
+pub(crate) enum ImageInfoLoader {
     /// An image file.
     File(gio::File),
     /// A texture in memory.
@@ -164,7 +166,7 @@ impl ImageInfoLoader {
     }
 
     /// Load the information for this image.
-    pub async fn load_info(self) -> BaseImageInfo {
+    pub(crate) async fn load_info(self) -> BaseImageInfo {
         self.into_first_frame()
             .await
             .map(|f| f.info())
@@ -173,7 +175,7 @@ impl ImageInfoLoader {
 
     /// Load the information for this image and try to generate a thumbnail
     /// given the filesize of the original image.
-    pub async fn load_info_and_thumbnail(
+    pub(crate) async fn load_info_and_thumbnail(
         self,
         filesize: Option<u32>,
         widget: &impl IsA<gtk::Widget>,
@@ -182,7 +184,7 @@ impl ImageInfoLoader {
             return (BaseImageInfo::default(), None);
         };
 
-        let info = frame.info();
+        let mut info = frame.info();
 
         // Generate the same thumbnail dimensions as we will need in the timeline.
         let scale_factor = widget.scale_factor();
@@ -195,6 +197,8 @@ impl ImageInfoLoader {
                 .is_some_and(|d| d.needs_thumbnail(max_thumbnail_dimensions))
         {
             // It is not worth it to generate a thumbnail.
+            info.blurhash = frame.generate_blurhash().map(|blurhash| blurhash.0);
+
             return (info, None);
         }
 
@@ -208,7 +212,10 @@ impl ImageInfoLoader {
             return (info, None);
         };
 
-        let thumbnail = frame.generate_thumbnail(scale_factor, &renderer);
+        let (thumbnail, blurhash) = frame
+            .generate_thumbnail_and_blurhash(scale_factor, &renderer)
+            .unzip();
+        info.blurhash = blurhash.map(|blurhash| blurhash.0);
 
         (info, thumbnail)
     }
@@ -266,27 +273,51 @@ impl Frame {
         }
     }
 
-    /// Generate a thumbnail of this frame.
-    fn generate_thumbnail(self, scale_factor: i32, renderer: &gsk::Renderer) -> Option<Thumbnail> {
+    /// Generate a Blurhash of this frame.
+    fn generate_blurhash(self) -> Option<Blurhash> {
         let texture = match self {
             Self::Glycin(frame) => frame.texture(),
             Self::Texture(texture) => texture,
         };
 
-        let thumbnail = TextureThumbnailer(texture).generate_thumbnail(scale_factor, renderer);
+        let blurhash = Blurhash::with_texture(&texture);
 
-        if thumbnail.is_none() {
-            warn!("Could not generate thumbnail from GdkTexture");
+        if blurhash.is_none() {
+            warn!("Could not generate Blurhash from GdkTexture");
         }
 
-        thumbnail
+        blurhash
+    }
+
+    /// Generate a thumbnail and a Blurhash of this frame.
+    ///
+    /// We use the thumbnail to compute the blurhash, which should be less
+    /// expensive than using the original frame.
+    fn generate_thumbnail_and_blurhash(
+        self,
+        scale_factor: i32,
+        renderer: &gsk::Renderer,
+    ) -> Option<(Thumbnail, Blurhash)> {
+        let texture = match self {
+            Self::Glycin(frame) => frame.texture(),
+            Self::Texture(texture) => texture,
+        };
+
+        let thumbnail_blurhash =
+            TextureThumbnailer(texture).generate_thumbnail_and_blurhash(scale_factor, renderer);
+
+        if thumbnail_blurhash.is_none() {
+            warn!("Could not generate thumbnail and Blurhash from GdkTexture");
+        }
+
+        thumbnail_blurhash
     }
 }
 
 /// Extensions to `FrameDimensions` for computing thumbnail dimensions.
 impl FrameDimensions {
     /// Get the maximum dimensions for a thumbnail with the given scale factor.
-    pub fn thumbnail_max_dimensions(scale_factor: i32) -> Self {
+    pub(crate) fn thumbnail_max_dimensions(scale_factor: i32) -> Self {
         let scale_factor = scale_factor.try_into().unwrap_or(1);
         THUMBNAIL_MAX_DIMENSIONS.scale(scale_factor)
     }
@@ -413,16 +444,21 @@ impl TextureThumbnailer {
     }
 
     /// Generate the thumbnail for the given scale factor, with the given
-    /// `GskRenderer`.
-    pub(super) fn generate_thumbnail(
+    /// `GskRenderer`, and a Blurhash.
+    ///
+    /// We use the thumbnail to compute the blurhash, which should be less
+    /// expensive than using the original texture.
+    pub(super) fn generate_thumbnail_and_blurhash(
         self,
         scale_factor: i32,
         renderer: &gsk::Renderer,
-    ) -> Option<Thumbnail> {
+    ) -> Option<(Thumbnail, Blurhash)> {
         let max_thumbnail_dimensions = FrameDimensions::thumbnail_max_dimensions(scale_factor);
         let thumbnail = self.downscale_texture_if_needed(max_thumbnail_dimensions, renderer)?;
         let dimensions = FrameDimensions::with_texture(&thumbnail)?;
 
+        let blurhash = Blurhash::with_texture(&thumbnail)?;
+
         let (downloader_format, webp_layout) =
             Self::texture_format_to_thumbnail_format(thumbnail.format())?;
 
@@ -437,27 +473,93 @@ impl TextureThumbnailer {
         let content_type =
             mime::Mime::from_str(WEBP_CONTENT_TYPE).expect("content type should be valid");
 
-        Some(Thumbnail {
+        let thumbnail = Thumbnail {
             data,
             content_type,
             width: dimensions.width.into(),
             height: dimensions.height.into(),
             size,
+        };
+
+        Some((thumbnail, blurhash))
+    }
+}
+
+/// A [Blurhash].
+///
+/// [Blurhash]: https://blurha.sh/
+#[derive(Debug, Clone)]
+pub(crate) struct Blurhash(pub(crate) String);
+
+impl Blurhash {
+    /// Try to compute the Blurhash for the given `GdkTexture`.
+    pub(super) fn with_texture(texture: &gdk::Texture) -> Option<Self> {
+        let dimensions = FrameDimensions::with_texture(texture)?;
+
+        let mut downloader = gdk::TextureDownloader::new(texture);
+        downloader.set_format(gdk::MemoryFormat::R8g8b8a8);
+        let (data, _) = downloader.download_bytes();
+
+        let (components_x, components_y) = match dimensions.width.cmp(&dimensions.height) {
+            Ordering::Less => (3, 4),
+            Ordering::Equal => (3, 3),
+            Ordering::Greater => (4, 3),
+        };
+
+        let hash = blurhash::encode(
+            components_x,
+            components_y,
+            dimensions.width,
+            dimensions.height,
+            &data,
+        )
+        .inspect_err(|error| {
+            warn!("Could not encode Blurhash: {error}");
         })
+        .ok()?;
+
+        Some(Self(hash))
+    }
+
+    /// Try to convert this Blurhash to a `GdkTexture` with the given
+    /// dimensions.
+    pub(crate) async fn into_texture(self, dimensions: FrameDimensions) -> Option<gdk::Texture> {
+        // Because it can take some time, spawn on a separate thread.
+        RUNTIME
+            .spawn_blocking(move || {
+                let data = blurhash::decode(&self.0, dimensions.width, dimensions.height, 1.0)
+                    .inspect_err(|error| {
+                        warn!("Could not decode Blurhash: {error}");
+                    })
+                    .ok()?;
+
+                Some(
+                    gdk::MemoryTexture::new(
+                        dimensions.width.try_into().ok()?,
+                        dimensions.height.try_into().ok()?,
+                        gdk::MemoryFormat::R8g8b8a8,
+                        &glib::Bytes::from_owned(data),
+                        4 * dimensions.width as usize,
+                    )
+                    .upcast(),
+                )
+            })
+            .await
+            .expect("task was not aborted")
     }
 }
 
 /// An API to download a thumbnail for a media.
 #[derive(Debug, Clone, Copy)]
-pub struct ThumbnailDownloader<'a> {
+pub(crate) struct ThumbnailDownloader<'a> {
     /// The main source of the image.
     ///
     /// This should be the source with the best quality.
-    pub main: ImageSource<'a>,
+    pub(crate) main: ImageSource<'a>,
     /// An alternative source for the image.
     ///
     /// This should be a source with a lower quality.
-    pub alt: Option<ImageSource<'a>>,
+    pub(crate) alt: Option<ImageSource<'a>>,
 }
 
 impl ThumbnailDownloader<'_> {
@@ -465,7 +567,7 @@ impl ThumbnailDownloader<'_> {
     ///
     /// This might not return a thumbnail at the requested dimensions, depending
     /// on the sources and the homeserver.
-    pub async fn download(
+    pub(crate) async fn download(
         self,
         client: Client,
         settings: ThumbnailSettings,
@@ -475,7 +577,10 @@ impl ThumbnailDownloader<'_> {
 
         // First, select which source we are going to download from.
         let source = if let Some(alt) = self.alt {
-            if !self.main.can_be_thumbnailed()
+            let is_animated = settings.animated && self.main.is_animated();
+
+            if !is_animated
+                && !self.main.can_be_thumbnailed()
                 && (filesize_is_too_big(self.main.filesize())
                     || alt.dimensions().is_some_and(|s| s.ge(settings.dimensions)))
             {
@@ -488,7 +593,11 @@ impl ThumbnailDownloader<'_> {
             self.main
         };
 
-        if source.should_thumbnail(settings.prefer_thumbnail, settings.dimensions) {
+        if source.should_thumbnail(
+            settings.prefer_thumbnail,
+            settings.animated,
+            settings.dimensions,
+        ) {
             // Try to get a thumbnail.
             let request = MediaRequestParameters {
                 source: source.source.to_common_media_source(),
@@ -518,11 +627,11 @@ impl ThumbnailDownloader<'_> {
 
 /// The source of an image.
 #[derive(Debug, Clone, Copy)]
-pub struct ImageSource<'a> {
+pub(crate) struct ImageSource<'a> {
     /// The source of the image.
-    pub source: MediaSource<'a>,
+    pub(crate) source: MediaSource<'a>,
     /// Information about the image.
-    pub info: Option<ImageSourceInfo<'a>>,
+    pub(crate) info: Option<ImageSourceInfo<'a>>,
 }
 
 impl ImageSource<'_> {
@@ -531,6 +640,7 @@ impl ImageSource<'_> {
     fn should_thumbnail(
         &self,
         prefer_thumbnail: bool,
+        prefer_animated: bool,
         thumbnail_dimensions: FrameDimensions,
     ) -> bool {
         if !self.can_be_thumbnailed() {
@@ -540,7 +650,7 @@ impl ImageSource<'_> {
         // Even if we request animated thumbnails, not a lot of media repositories
         // support scaling animated images. So we just download the original to be able
         // to play it.
-        if self.is_animated() {
+        if prefer_animated && self.is_animated() {
             return false;
         }
 
@@ -607,7 +717,7 @@ fn filesize_is_too_big(filesize: Option<u32>) -> bool {
 
 /// The source of a media file.
 #[derive(Debug, Clone, Copy)]
-pub enum MediaSource<'a> {
+pub(crate) enum MediaSource<'a> {
     /// A common media source.
     Common(&'a CommonMediaSource),
     /// The media source of a sticker.
@@ -656,7 +766,7 @@ impl<'a> From<&'a OwnedMxcUri> for MediaSource<'a> {
 
 /// Information about the source of an image.
 #[derive(Debug, Clone, Copy, Default)]
-pub struct ImageSourceInfo<'a> {
+pub(crate) struct ImageSourceInfo<'a> {
     /// The dimensions of the image.
     dimensions: Option<FrameDimensions>,
     /// The MIME type of the image.
@@ -702,13 +812,13 @@ impl<'a> From<&'a AvatarImageInfo> for ImageSourceInfo<'a> {
 
 /// The settings for downloading a thumbnail.
 #[derive(Debug, Clone)]
-pub struct ThumbnailSettings {
+pub(crate) struct ThumbnailSettings {
     /// The requested dimensions of the thumbnail.
-    pub dimensions: FrameDimensions,
+    pub(crate) dimensions: FrameDimensions,
     /// The method to use to resize the thumbnail.
-    pub method: Method,
+    pub(crate) method: Method,
     /// Whether to request an animated thumbnail.
-    pub animated: bool,
+    pub(crate) animated: bool,
     /// Whether we should prefer to get a thumbnail if dimensions are unknown.
     ///
     /// This is particularly useful for avatars where we will prefer to save
@@ -716,7 +826,7 @@ pub struct ThumbnailSettings {
     /// appear several times on the screen. For media messages, we will on the
     /// contrary prefer to download the original content to reduce the space
     /// taken in the media cache.
-    pub prefer_thumbnail: bool,
+    pub(crate) prefer_thumbnail: bool,
 }
 
 impl From<ThumbnailSettings> for MediaThumbnailSettings {
@@ -739,7 +849,7 @@ impl From<ThumbnailSettings> for MediaThumbnailSettings {
 
 /// An error encountered when loading an image.
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
-pub enum ImageError {
+pub(crate) enum ImageError {
     /// Could not download the image.
     Download,
     /// Could not save the image to a temporary file.
